/* $Id: fa.c,v 1.266 2001/09/29 07:52:24 jm Exp $
 * Foreign Agent
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <string.h>
#include <arpa/inet.h>
#include <errno.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <time.h>
#include <assert.h>
#include <getopt.h>
#include <sys/uio.h>
#include <fcntl.h>
#include <netinet/ip_icmp.h>

#include "fa.h"
#include "fa_hash.h"
#include "tunnel.h"
#include "list.h"
#include "hashtable.h"
#include "agentadv.h"
#include "rsa_dyn.h"
#include "debug.h"
#include "fixed_fd_zero.h"
#include "dyn_ip.h"
#include "agentapi.h"
#include "util.h"
#include "dyn_api.h"
#include "agent_utils.h"
#include "fa_mn_addr.h"
#ifdef INCLUDE_IPAY
#include "fa_ipay.h"
#endif

struct sol_reply_list {
	struct timeval send;
	struct interface_entry *iface;
	struct sockaddr_ll to;
	struct in_addr addr;
	struct sol_reply_list *next;
};

struct dest_unreachable_data {
	char code;
	struct in_addr addr;
	int port;
};

/* sockets */
#ifdef INCLUDE_IPAY
static int ipay_sock = -1;
#endif
static int api_rw_sock = -1;
static int api_ro_sock = -1;
int fa_info_sock = -1;
struct sockaddr_un fa_info_addr;

/* dynamic data */
static struct dynamics_fa_status stats;

struct hashtable *tunnels_hash;
struct bindingtable *bindings_hash;

struct sockaddr_in upper_fa_addr;
struct interface_entry *up_interface = NULL;
struct msg_key *fa_public_key = NULL;
struct fa_nai_ext *fa_nai = NULL;
struct fa_config *config;
struct binding_counters bcounters;
struct challenge_ext **last_challenges;
int last_challenges_pos;
static char *program_name; /* the name this program was run with */
static struct sol_reply_list *solrep_list = NULL;
extern time_t fa_reg_next_try; /* fa_hash */

static void init_sockets(void);
static void init_config_data(void);
static int check_delayed_solrep(struct timeval *now, struct timeval *next);
static int
fa_send_agent_adv(struct interface_entry *iface, struct sockaddr_ll *to,
		  struct in_addr *dest);
static void handle_dest_unreach(struct dest_unreachable_data *data);

/* util.c - common command line arguments */
extern int opt_foreground;
extern char *opt_config;

static void
close_sockets(void)
{
#ifdef INCLUDE_IPAY
	close_socket(ipay_sock);
	ipay_sock = -1;
#endif
	if (api_rw_sock > -1)
		close_socket(api_rw_sock);
	api_rw_sock = -1;
	if (api_ro_sock > -1)
		close_socket(api_ro_sock);
	api_ro_sock = -1;
	if (fa_info_sock > -1) {
		info_send(FA_INFO_CLOSE, FA_INFO_OK, NULL, 0);
		close_socket(fa_info_sock);
		fa_info_sock = -1;
	}

	/* remove unix sockets */
	if (config != NULL && config->fa_api_read_socket_path != NULL)
		unlink(config->fa_api_read_socket_path);
	if (config != NULL && config->fa_api_admin_socket_path != NULL)
		unlink(config->fa_api_admin_socket_path);
}


static void
flush_solrep_list(void)
{
	struct sol_reply_list *prev, *tmp = solrep_list;
	while (tmp != NULL) {
		DEBUG(DEBUG_FLAG, "Flushed solrep entry\n");
		prev = tmp;
		tmp = tmp->next;
		free(prev);
	}
	solrep_list = NULL;
}


/* Free all reserved memory. Destroy hashes.
 * Unlink the unix domain socket filenames.
 */
static void
clean_up(int sig)
{
	/* destroy all hashes and finish nicely */
	DEBUG(DEBUG_FLAG, "clean_up(%d)\n", sig);
	syslog(LOG_INFO, "Foreign agent daemon cleaning up (signal %d)", sig);

	fa_register(FA_DEREGISTER);

	close_sockets();
	unlink(FA_PID_FILE);

	/* Free hashtables */
	if (bindings_hash != NULL) {
		DEBUG(DEBUG_FLAG, "Removing bindings..\n");
		binding_iterator(bindings_hash, eliminate_binding_entry_force,
				 &bcounters);
		binding_destroy(bindings_hash);
	}

	if (tunnels_hash != NULL) {
		DEBUG(DEBUG_FLAG, "Removing tunnels..\n");
		tunnel_destroy_hash(tunnels_hash);
	}

	fa_hash_destroy();
	expire_denial_records(1);

	flush_solrep_list(); /* remove possible entries from the solrep list */
	cleanup_fa_config(config);
	if (config != NULL)
		free(config);
	if (fa_nai != NULL) {
		free(fa_nai);
		fa_nai = NULL;
	}

	closelog();
	exit(sig);
}


static struct interface_entry *
get_iface(int ifindex)
{
	struct interface_entry *iface;
	struct node *node;
	for (node = list_get_first(&config->interfaces); node != NULL;
	     node = list_get_next(node)) {
		iface = (struct interface_entry *) node;
		if (iface->if_index == ifindex)
			return iface;
	}

	return NULL;
}


/* Update the binding iface pointers after configuration reload.
 * old_if_index is used to get the new pointer as the interface index does not
 * change */
static int
reload_binding_iter(struct node *node, void *data)
{
	struct bindingentry *binding;
	struct tunnel_data *t_data;
	struct unconfirmed_request *unc;

	binding = (struct bindingentry *) node;
	t_data = (struct tunnel_data *) binding->data;
	assert(t_data != NULL);

	if (t_data->info.iface != NULL) {
		t_data->info.iface = get_iface(t_data->old_if_index);
		if (t_data->info.iface == NULL) {
			LOG2(LOG_ALERT, "Interface disappeared during SIGHUP "
			     "processing\n");
			/* drop the binding and continue operation */
			remove_binding_tunnels(binding, tunnels_hash,
					       &bcounters);
		}
	}

	unc = t_data->unc_req;
	while (unc != NULL) {
		unc->info.iface = get_iface(unc->old_if_index);
		if (unc->info.iface == NULL) {
			LOG2(LOG_ALERT, "Interface disappeared during SIGHUP "
			     "processing\n");
			/* drop the unconfirmed requests and continue
			 * operation */
			free_unconfirmed_data(t_data->unc_req);
			t_data->unc_req = NULL;
			break;
		}

		unc = unc->next;
	}

	return 1;
}


/* Update the interface pointers in bindings and unconfirmed requests */
static void
reload_interface_pointers(void)
{
	binding_iterator(bindings_hash, reload_binding_iter, NULL);
}

static void
reload_config(int sig)
{
	int i;

	LOG2(LOG_INFO, "Reloading configuration\n");

	for (i = 0; i < config->challenge_window; i++) {
		if (last_challenges[i] != NULL)
			free(last_challenges[i]);
	}
	free(last_challenges);
	last_challenges = NULL;

	/* remove possible entries from the solrep list; this needs to be done
	 * because the iface pointers change during configuration reloading */
	flush_solrep_list();
	cleanup_fa_config(config);
        if (load_fa(config, program_name,
		    opt_config == NULL ? FA_GLOBAL_CONF_FILE : opt_config) ==
	    FALSE) {
		LOG2(LOG_ALERT, "Reloading configuration failed\n");
		clean_up(-1);
        }

	close_sockets();
	closelog();
	init_sockets();
	init_config_data();
	reload_interface_pointers();
}


/* search for spi/addr/agent_type match from the FA SPI list
 * if spi=0, accept any spi
 * if agent_type=0, accept any type
 * Returns: pointer to the SPI entry or NULL on failure */
struct fa_spi_entry *
get_fa_spi(int spi, struct in_addr addr, int agent_type)
{
	struct fa_spi_entry *s;
	struct node *node;

	/* FIX: this might be optimized by using a hash table index by the
	 * address field */
	for (node = list_get_first(&config->fa_spi_list); node != NULL;
	     node = list_get_next(node)) {
		s = (struct fa_spi_entry *) node;
		if ((spi == 0 || s->spi == spi) &&
		    s->addr.s_addr == addr.s_addr &&
		    (agent_type == 0 || agent_type == s->agent_type)) {
			return s;
		}
	}
	return NULL;
}


/**
 * binding_get_tunnels_iter:
 * @node: pointer to bindingentry
 * @data: pointer to the message to be returned to the API caller
 *
 * Called from handle_api on an API_GET_TUNNELS request. This function is
 * called for each existing binding in the bindings list.
 *
 * Returns:
 *  1 (to continue iteration loop)
 */
static int
binding_get_tunnels_iter(struct node *node, void *data)
{
	struct bindingentry *binding = (struct bindingentry *) node;
	struct api_msg *api_msg = (struct api_msg *) data;
	dyn_tunnel_id id;

	if (api_msg->length + sizeof(dyn_tunnel_id) > API_DATA_SIZE) {
		DEBUG(DEBUG_FLAG, "API message overflow - skipping rest of "
		      "the bindings\n");
		return 0;
	}

	id.mn_addr.s_addr =  binding->mn_addr.s_addr;
	id.ha_addr.s_addr =  binding->ha_addr.s_addr;
	id.priv_ha = binding->priv_ha;

	memcpy((char *) api_msg->params + api_msg->length, &id,
	       sizeof(dyn_tunnel_id));

	api_msg->length += sizeof(dyn_tunnel_id);

	return 1;
}


/* Handle api call from socket sock. If it is privileged, the admin
 * is set nonzero. */
static void
handle_api(int sock, int admin)
{
	struct sockaddr_un addr;
	socklen_t addrlen;
	struct api_msg recv_msg;
	struct api_msg send_msg;
	struct bindingentry *binding;
	struct dynamics_tunnel_info *info;
	struct dynamics_fa_status *status;
	char *path;
	int flags;
	dyn_tunnel_id *tid;
	struct in_addr tmp_addr;
	struct bindingkey bkey;

        DEBUG(DEBUG_FLAG, "receive from api\n");
	addrlen = sizeof(addr); /* set size of address buffer */
	memset(&addr, 0, addrlen); /* zero the address */
	if (api_receive(sock, &addr, &addrlen, &recv_msg) < 0) {
		LOG2(LOG_ERR, "api receive: %s\n", strerror(errno));
		return;
	}
	if (recv_msg.type != API_CALL_MSG) return;

	send_msg.code = API_UNDEFINED;
	send_msg.length = 0;

	if (admin)
		stats.apicalls_admin++;
	else
		stats.apicalls_read++;

	switch (recv_msg.code) {
        case API_GET_TUNNELS:
		DEBUG(DEBUG_FLAG, "Received API_GET_TUNNELS\n");
		if (recv_msg.length != 0) {
			send_msg.code = API_ILLEGAL_PARAMETERS;
			break;
		}
		if (binding_iterator(bindings_hash, binding_get_tunnels_iter,
				     &send_msg))
			send_msg.code = API_SUCCESS;
		else
			send_msg.code = API_INSUFFICIENT_SPACE;
		break;

        case API_GET_TUNNEL_INFO:
		DEBUG(DEBUG_FLAG, "Received API_GET_TUNNEL_INFO\n");
		if (recv_msg.length != sizeof(dyn_tunnel_id)) {
			send_msg.code = API_ILLEGAL_PARAMETERS;
			break;
		}
		tid = (dyn_tunnel_id *) recv_msg.params;
		UNALIGNED_(&tmp_addr, &tid->ha_addr);
		if (tmp_addr.s_addr == 0) {
			UNALIGNED_(&tmp_addr, &tid->mn_addr);
			binding = binding_fetch2(bindings_hash, &tmp_addr);
		} else {
			UNALIGNED_(&bkey.mn_addr, &tid->mn_addr);
			UNALIGNED_(&bkey.ha_addr, &tid->ha_addr);
			UNALIGNED_(&bkey.priv_ha, &tid->priv_ha);
			binding = binding_fetch(bindings_hash, &bkey);
		}
		if (!binding) {
			DEBUG(DEBUG_FLAG, "\tbinding not found\n");
			send_msg.code = API_FAILED;
			break;
		}
		info = (struct dynamics_tunnel_info *) send_msg.params;
		memset(info, 0, sizeof(struct dynamics_tunnel_info));
		UNALIGNED_(&info->id.mn_addr, &binding->mn_addr);
		UNALIGNED_(&info->id.ha_addr, &binding->ha_addr);
		UNALIGNED_(&info->mn_addr, &binding->mn_addr);
		UNALIGNED_(&info->co_addr, &binding->lower_addr);
		UNALIGNED_(&info->ha_addr, &binding->ha_addr);
		UNALIGNED_(&info->priv_ha, &binding->priv_ha);
		if (binding->data != NULL) {
			struct tunnel_data *t_data = binding->data;
			UNALIGNED_(&info->confirmed, &t_data->confirmed);
		}
		UNALIGNED_(&info->expiration_time, &binding->exp_time);
		UNALIGNED_(&info->creation_time, &binding->create_time);
		UNALIGNED_(&info->refresh_time, &binding->mod_time);
		UNALIGNED_(&info->last_timestamp[0], &binding->id[0]);
		UNALIGNED_(&info->last_timestamp[1], &binding->id[1]);
		UNALIGNED_(&info->spi, &binding->spi);
		UNALIGNED_(&info->timeout, &binding->timeout);
		send_msg.code = API_SUCCESS;
		send_msg.length = sizeof(struct dynamics_tunnel_info);
		break;

	case API_DESTROY_TUNNEL:
		DEBUG(DEBUG_FLAG, "Received API_DESTROY_TUNNEL\n");
		if (!admin) {
			send_msg.code = API_NOT_PERMITTED;
			break;
		}
		if (recv_msg.length != sizeof(dyn_tunnel_id)) {
			send_msg.code = API_ILLEGAL_PARAMETERS;
			break;
		}
		tid = (dyn_tunnel_id *) recv_msg.params;
		if (tid->ha_addr.s_addr == 0)
			binding = binding_fetch2(bindings_hash, &tid->mn_addr);
		else {
			bkey.mn_addr = tid->mn_addr;
			bkey.ha_addr = tid->ha_addr;
			bkey.priv_ha = tid->priv_ha;
			binding = binding_fetch(bindings_hash, &bkey);
		}
		if (!binding) {
			send_msg.code = API_FAILED;
			break;
		}
		eliminate_binding_entry((struct node *) binding, &bcounters);
		send_msg.length = 0;
		send_msg.code = API_SUCCESS;
		break;

	case API_GET_STATUS:
		DEBUG(DEBUG_FLAG, "Received API_GET_STATUS\n");
		if (recv_msg.length != 0) {
			send_msg.code = API_ILLEGAL_PARAMETERS;
			break;
		}
		status = (struct dynamics_fa_status *) send_msg.params;
		memcpy(status, &stats, sizeof(struct dynamics_fa_status));
		UNALIGNED_(&status->tunnel_count, &bcounters.bindingcount);
		UNALIGNED_(&status->pending_count, &bcounters.pendingcount);

		send_msg.length = sizeof(struct dynamics_fa_status);
		send_msg.code = API_SUCCESS;
		break;

	case API_ATTACH_INFO_SOCKET:
		if (!admin) {
			send_msg.code = API_NOT_PERMITTED;
			break;
		}
		if (recv_msg.length < 1 ||
		    recv_msg.length >= sizeof(fa_info_addr.sun_path)) {
			send_msg.code = API_ILLEGAL_PARAMETERS;
			break;
		}
		if (fa_info_sock > -1) {
			info_send(FA_INFO_CLOSE, FA_INFO_OK, NULL, 0);
			close_socket(fa_info_sock);
		}
		path = (char *) recv_msg.params;
		path[recv_msg.length - 1] = '\0';
		DEBUG(DEBUG_FLAG, "API attach info socket[%s]\n", path);

		fa_info_sock = socket(AF_LOCAL, SOCK_DGRAM, 0);
		if (fa_info_sock < 0) {
			DEBUG(DEBUG_FLAG, "\tfailed: %s\n", strerror(errno));
			send_msg.code = API_FAILED;
			break;
		}
		/* Set non-blocking */
		flags = fcntl(fa_info_sock, F_GETFL, 0);
		if (flags == -1 ||
		    fcntl(fa_info_sock, F_SETFL, flags | O_NDELAY) == -1) {
			DEBUG(DEBUG_FLAG, "\tfcntl F_SETFL failed: %s\n",
			      strerror(errno));
			close(fa_info_sock);
			fa_info_sock = -1;
			send_msg.code = API_FAILED;
			break;
		}

		memset(&fa_info_addr, 0, sizeof(fa_info_addr));
		fa_info_addr.sun_family = AF_LOCAL;
		dynamics_strlcpy(fa_info_addr.sun_path, path, recv_msg.length);

		send_msg.code = API_SUCCESS;
		break;

	case API_DETACH_INFO_SOCKET:
		if (!admin) {
			send_msg.code = API_NOT_PERMITTED;
			break;
		}
		DEBUG(DEBUG_FLAG, "API detach info socket - ");
		if (fa_info_sock < 0) {
			DEBUG(DEBUG_FLAG, "failed\n");
			send_msg.code = API_FAILED;
			break;
		}

		info_send(FA_INFO_CLOSE, FA_INFO_OK, NULL, 0);
		close_socket(fa_info_sock);
		fa_info_sock = -1;
		send_msg.code = API_SUCCESS;
		break;

	case API_SEND_ADV:
	{
		/**
		 * Now the only parameter we use is ifindex
		 * but i think we should expand it to support
		 * direct (send to mac addr) advertisements.
		 * - roger@lifix.fi
		 */
		struct interface_entry *iface;
		struct dyn_param_adv *adv;
		if (recv_msg.length != sizeof(struct dyn_param_adv) ) {
			send_msg.code = API_ILLEGAL_PARAMETERS;
			break;
		}
		adv = (struct dyn_param_adv *) recv_msg.params;
		if ( (iface = get_iface(adv->ifindex) ) == NULL ) {
			send_msg.code = API_FAILED;
			break;
		}
		fa_send_agent_adv(iface, NULL, NULL);
		send_msg.length = 0;
		send_msg.code = API_SUCCESS;
		break;
	}

	default:
		/* just send a message indicating that function is not
		 * supported. No need to confirm that message gets through. */
		LOG2(LOG_ERR, "Received unknown API command: %d\n",
		     recv_msg.code);
		send_msg.code = API_NOT_SUPPORTED;
		break;
	}
	api_send(sock, &addr, addrlen, &send_msg);
}


static void
handle_errqueue(int s)
{
	char msg[256];
	int n;
	struct sockaddr_in cli_addr;
	struct msghdr mh;
	char cdata[256];
	struct iovec iov;
	struct cmsghdr *cmsg;
	struct dest_unreachable_data unreach;


	memset(&mh, 0, sizeof(mh));
	iov.iov_base = msg;
	iov.iov_len = sizeof(msg);
	mh.msg_name = &cli_addr;
	mh.msg_namelen = sizeof(cli_addr);
	mh.msg_iov = &iov;
	mh.msg_iovlen = 1;
	mh.msg_control = &cdata;

	for (;;) {
		struct dynamics_sock_extended_err *err = NULL;

		memset(&cli_addr, 0, sizeof(cli_addr));
		mh.msg_controllen = sizeof(cdata);

		n = recvmsg(s, &mh, MSG_ERRQUEUE);
		if (n < 0)
			break;

		memset(&unreach, 0, sizeof(unreach));
		unreach.port = -1;

		if (cli_addr.sin_family == AF_INET) {
			DEBUG(DEBUG_FLAG, "Learned unreachable address from "
			      "msg_name: %s:%i\n",
			      inet_ntoa(cli_addr.sin_addr),
			      ntohs(cli_addr.sin_port));
			unreach.addr.s_addr = cli_addr.sin_addr.s_addr;
			unreach.port = cli_addr.sin_port;
		}

		for (cmsg = CMSG_FIRSTHDR(&mh); cmsg != NULL;
		     cmsg = DYNAMICS_CMSG_NXTHDR(&mh, cmsg)) {
			if (cmsg->cmsg_level == SOL_IP &&
			    cmsg->cmsg_type == IP_RECVERR) {
				err = (struct dynamics_sock_extended_err *)
					CMSG_DATA(cmsg);
			}
		}

		if (err == NULL) {
			DEBUG(DEBUG_FLAG, "Could not get IP_RECVERR data\n");
			break;
		}

		DEBUG(DEBUG_FLAG, "\tIP_RECVERR: errno=%i origin=%i "
		      "type=%i code=%i pad=%i info=%i data=%i\n",
		      err->ee_errno, err->ee_origin, err->ee_type,
		      err->ee_code, err->ee_pad, err->ee_info,
		      err->ee_data);

		unreach.code = err->ee_code;

		if (unreach.port == -1) {
			/* Older kernels (2.2.x?) seem to give bogus port
			 * entries or do not return address in msg_name, so try
			 * to get the IP address from offender data */
			struct sockaddr_in *offender;
			offender = (struct sockaddr_in *) SO_EE_OFFENDER(err);
			if (offender == NULL ||
			    offender->sin_family != AF_INET) {
				DEBUG(DEBUG_FLAG,
				      "Could not get offender data\n");
				break;
			}

			DEBUG(DEBUG_FLAG, "\toffender %s:%i (ignoring bogus "
			      "port)\n",
			      inet_ntoa(offender->sin_addr),
			      ntohs(offender->sin_port));
			unreach.addr.s_addr = offender->sin_addr.s_addr;
			/* offender->sin_port seem to be bogus in 2.2.17 and
			 * zero in 2.4.4, so better ignore it; port number -1
			 * is handled as a special wildcard later */
		}

		handle_dest_unreach(&unreach);
	}
}


/* Rotating L2 data queue of last registration messages */
#define L2_QUEUE_LEN 10
/* How many bytes to save from each packet; this is the size of the
 * registration request header. If the message is registration reply, only
 * first 20 bytes are checked when searching for L2 data. */
#define L2_SAVE_UDP_LEN 24

struct pending_registration_l2_info {
	struct packet_from_info info;
	unsigned char udp_start[L2_SAVE_UDP_LEN];
};

static struct pending_registration_l2_info l2_info[L2_QUEUE_LEN];
static int l2_next_pos = 0;


/* Received registration message. Process it.
   returns:
     0 if successful
     1 on error (not enough memory)
*/
static int
handle_reg_msg(struct interface_entry *iface)
{
	char msg[MAXMSG];
	int res, n, check_len;
	struct sockaddr_in from;
	socklen_t fromlen;
	struct packet_from_info *info = NULL, tmpinfo;
	struct msg_extensions ext;

	memset(&from, 0, sizeof(from));

	fromlen = sizeof(from);
	n = recvfrom(iface->udp_sock, msg, sizeof(msg), 0,
		     (struct sockaddr *) &from, &fromlen);
	if (n < 0) {
		DEBUG(DEBUG_FLAG, "handle_reg_msg - recvmsg: (errno=%i) "
		      "%s\n", errno, strerror(errno));
		handle_errqueue(iface->udp_sock);

		return 0;
	}

	DEBUG(DEBUG_FLAG, "Received %d bytes from %s:%d\n", n,
	      inet_ntoa(from.sin_addr), ntohs(from.sin_port));

	/* search for L2 data from packet socket */
	if (n > 0 && msg[0] == REG_REP)
		check_len = sizeof(struct reg_rep);
	else
		check_len = L2_SAVE_UDP_LEN;

	if (n >= check_len) {
		int pos = l2_next_pos;
		for (;;) {
			if (pos == 0)
				pos = L2_QUEUE_LEN - 1;
			else
				pos--;
			if (pos == l2_next_pos)
				break;

			if (memcmp(msg, &l2_info[pos].udp_start,
				   check_len) == 0) {
				DEBUG(DEBUG_FLAG, "\tpending L2 data found "
				      "from pos %i\n", pos);
				info = &l2_info[pos].info;
				break;
			}
		}
	}

	if (info == NULL) {
		DEBUG(DEBUG_FLAG, "\tcould not find pending L2 data - using "
		      "dummy entry\n");
		memset(&tmpinfo, 0, sizeof(tmpinfo));
		info = &tmpinfo;
		info->iface = iface;
		memcpy(&info->src, &from, sizeof(info->src));
	}

	res = parse_msg(msg, n, &ext);

	if (res != 0) {
		char *reason = "N/A";
		if (res == -1) {
			reason = "unknown extension";
			stats.discarded_unknown_ext++;
		} else if (res == -2) {
			reason = "malformed message";
			stats.discarded_malformed_msg++;
		} else if (res == -3 || res == -4 || res == -5) {
			reason = "unknown vendor extension";
			stats.discarded_vendor_ext++;
		}
		report_discarded_msg(msg, n, &info->src, reason);
		DEBUG(DEBUG_FLAG, "parse_msg returned %i\n", res);
		if (res == -3 || res == -4) {
			DEBUG(DEBUG_FLAG, "Unknown critical vendor extension -"
			      " sending failure reply\n");
			send_failure_reply(REGREP_UNSUPP_VENDOR_ID_MN_FA, &ext,
					   info, NULL, 0);
		}
		if (res != -5 || ext.rep == NULL) {
			return 0;
		}
		/* let registration replies with unknown CVSE pass through to
		 * be denied properly in handle_reply() */
	}

	if (ext.req && iface->type == INTERFACE_TYPE_UP) {
		DEBUG(DEBUG_FLAG, "Received registration request from "
		      "interface type UP - rejecting\n");
		report_discarded_msg(msg, n, &info->src,
				     "registration request from UP interface");
		send_failure_reply(REGREP_ADMIN_PROHIBITED_FA, &ext,
				   info, NULL, 0);
	} else if (ext.rep && iface->type == INTERFACE_TYPE_DOWN) {
		DEBUG(DEBUG_FLAG, "Received registration reply from "
		      "interface type DOWN - rejecting\n");
		report_discarded_msg(msg, n, &info->src, "registration reply "
				     "from DOWN interface");
	} else if (ext.req) {
		res = handle_request(&ext, info, config);
		if (res) {
			stats.req_rejected++;
			report_discarded_msg(msg, n, &info->src,
					     "registration request rejected");
		} else
			stats.req_accepted++;
		info_send(FA_INFO_REQUEST,
			  res == 0 ? FA_INFO_OK : FA_INFO_FAILED, ext.req,
			  sizeof(struct reg_req));
	} else if (ext.rep) {
		res = handle_reply(&ext, info);
		if (res) {
			stats.rep_rejected++;
			report_discarded_msg(msg, n, &info->src,
					     "registration reply rejected");
		} else
			stats.rep_accepted++;
		info_send(FA_INFO_REPLY,
			  res == 0 ? FA_INFO_OK : FA_INFO_FAILED, ext.rep,
			  sizeof(struct reg_rep));
	} else if (ext.fa_req) {
		res = handle_fa_req(msg, n, &ext, &info->src, iface);
	} else if (ext.fa_rep) {
		res = handle_fa_rep(msg, n, &ext, &info->src, iface);
	} else {
		DEBUG(DEBUG_FLAG, "Unknown registration message type\n");
	}

	return 0;
}


static int
handle_reg_msg_packet(struct interface_entry *iface)
{
	char msg[MAXMSG];
	int n, i;
	socklen_t fromlen;
	struct packet_from_info *info;
	struct iphdr *ip;
	struct udphdr *udp;
	int hdrlen = sizeof(struct iphdr) + sizeof(struct udphdr);

	info = &l2_info[l2_next_pos].info;

	memset(info, 0, sizeof(*info));
	info->iface = iface;

	fromlen = sizeof(info->from);
	n = recvfrom(iface->udp_packet, msg, sizeof(msg), 0,
		     (struct sockaddr *) &info->from, &fromlen);

	if (n < 0) {
		DEBUG(DEBUG_FLAG, "handle_reg_msg_packet - recvfrom: %s\n",
		      strerror(errno));
		return -1;
	}

	if (n < hdrlen + L2_SAVE_UDP_LEN) {
		DEBUG(DEBUG_FLAG, "handle_reg_msg_packet - underflow\n");
		return -1;
	}

	DEBUG(DEBUG_FLAG, "\treceived %i bytes - saving L2 data to pos %i\n",
	      n, l2_next_pos);
	DEBUG(DEBUG_FLAG, "\tsll_ifindex=%i sll_hatype=0x%04x sll_pkttype=%i "
	      "sll_halen=%i sll_addr=",
	      info->from.sll_ifindex, ntohs(info->from.sll_hatype),
	      info->from.sll_pkttype, info->from.sll_halen);
	for (i = 0; i < sizeof(info->from.sll_addr); i++) {
		if (i > 0)
			DEBUG(DEBUG_FLAG, ":");
		DEBUG(DEBUG_FLAG, "%02x", info->from.sll_addr[i]);
	}
	DEBUG(DEBUG_FLAG, "\n");

	ip = (struct iphdr *) msg;
	if (ip->protocol != IPPROTO_UDP) {
		DEBUG(DEBUG_FLAG, "\tskipping non-UDP packet (protocol=%i)\n",
		      ip->protocol);
		return -1;
	}

	udp = (struct udphdr *) (ip + 1);
	if (ntohs(udp->dest) != 434) {
		DEBUG(DEBUG_FLAG, "\tskipping UDP packet to port other than "
		      "434 (%i)\n", ntohs(udp->dest));
		return -1;
	}

	info->ttl = ip->ttl;
	info->dst_addr.s_addr = ip->daddr;
	info->src.sin_family = AF_INET;
	info->src.sin_addr.s_addr = ip->saddr;
	info->src.sin_port = udp->source;

	memcpy(&l2_info[l2_next_pos].udp_start, udp + 1, L2_SAVE_UDP_LEN);

	l2_next_pos++;
	if (l2_next_pos >= L2_QUEUE_LEN)
		l2_next_pos = 0;

	return 0;
}


/* Open sockets and initialize them. */
static void
init_sockets(void)
{
	struct interface_entry *iface;
	int one = 1;

#ifdef INCLUDE_IPAY
	struct in_addr dummy;
	memset(&dummy, 0, sizeof(dummy));
	ipay_sock = dynamics_open_udp_socket(config, dummy, config->ipay_port,
					     NULL);
	if (ipay_sock < 0) {
		LOG2(LOG_ALERT, "Ipay socket open failed - %s\n",
		     strerror(errno));
		clean_up(-1);
	}
#endif

	/* Unix sockets for API calls */
	api_rw_sock = api_open_socket(config->fa_api_admin_socket_path,
				      config->fa_api_admin_socket_group,
				      config->fa_api_admin_socket_owner,
				      config->fa_api_admin_socket_permissions);
	if (api_rw_sock < 0) {
		syslog(LOG_ERR, "init - open_api_rw_socket: %m");
	}

	api_ro_sock = api_open_socket(config->fa_api_read_socket_path,
				      config->fa_api_read_socket_group,
				      config->fa_api_read_socket_owner,
				      config->fa_api_read_socket_permissions);
	if (api_ro_sock < 0) {
		syslog(LOG_ERR, "init - open_api_ro_socket: %m");
	}

	/* RAW sockets for ICMP agent solicitation messages */
	DEBUG(DEBUG_FLAG, "Initializing interfaces\n");
	iface = (struct interface_entry *) list_get_first(&config->interfaces);
	if (iface == NULL) {
		LOG2(LOG_ALERT, "No interfaces defined\n");
		clean_up(-1);
	}
	while (iface != NULL) {
		DEBUG(DEBUG_FLAG, "\t%s: ", iface->dev);
		iface->if_index = dyn_ip_get_ifindex(iface->dev);
		if (iface->if_index < 0) {
			LOG2(LOG_ALERT,
			     "Could not get ifindex for interface %s\n",
			     iface->dev);
			clean_up(-1);
		}
		DEBUG(DEBUG_FLAG, "ifindex=%i ", iface->if_index);

		if (iface->type == INTERFACE_TYPE_BOTH ||
		    iface->type == INTERFACE_TYPE_UP)
			up_interface = iface;

		if (iface->force_addr.s_addr != 0) {
			DEBUG(DEBUG_FLAG, "forcing address ");
			iface->addr = iface->force_addr;
		} else {
			if (dyn_ip_get_ifaddr(iface->dev, &iface->addr) != 0) {
				LOG2(LOG_ALERT, "could not get interface[%s] "
				     "address\n", iface->dev);
				clean_up(-1);
			}
		}

		iface->udp_sock =
			dynamics_open_udp_socket(config->socket_priority,
						 iface->force_addr,
						 config->udp_port,
						 iface->dev);
		if (iface->udp_sock < 0) {
			LOG2(LOG_ALERT, "UDP socket opening failed\n");
			clean_up(-1);
		}

		if (setsockopt(iface->udp_sock, SOL_IP, IP_RECVERR, 
			       (const void *) &one, sizeof(one)) < 0) {
			DEBUG(DEBUG_FLAG,
			      "init_sockets - setsockopt(SOL_IP, IP_RECVERR): "
			      " %s\n", strerror(errno));
			/* continue without IP_RECVERR */
		}

		iface->udp_packet =
			open_agent_icmp_adv_socket(iface->dev,
						   UDP_FILTER_PORT434);
		if (iface->udp_packet < 0) {
			LOG2(LOG_ALERT, "could not open packet socket for "
			     "UDP port 434\n");
			clean_up(-1);
		}

		iface->icmp_sock =
			open_agent_icmp_adv_socket(iface->dev,
						   AGENTADV_FILTER_SOL);
		if (iface->icmp_sock < 0) {
			LOG2(LOG_ALERT, "could not open ICMP socket\n");
			clean_up(-1);
		}

		DEBUG(DEBUG_FLAG, "%s => socket=%i\n", inet_ntoa(iface->addr),
		      iface->icmp_sock);

		iface = (struct interface_entry *) list_get_next(&iface->node);
	}
}


/* initializations that are run on all config reads */
static void
init_config_data(void)
{
	int i;

	/* Configure logging */
	openlog("foreign agent", LOG_PID | LOG_CONS, config->syslog_facility);

	/* Fill in the upper_fa_addr */
	upper_fa_addr.sin_family = AF_INET;
	memcpy(&upper_fa_addr.sin_addr, &config->upper_fa_addr,
	       sizeof(config->upper_fa_addr));
	upper_fa_addr.sin_port = htons(config->upper_fa_port);
	DEBUG(DEBUG_FLAG, "\tinit_data: upper_fa_addr = %s:%d\n",
	      inet_ntoa(upper_fa_addr.sin_addr),
	      ntohs(upper_fa_addr.sin_port));

	/* Get the FA public key */
	if (rsa_initialize(config->key_file) != 0) {
		fprintf(stderr, "RSA initialization failed (key file: %s)\n",
			config->key_file);
		clean_up(1);
	}

	if (fa_public_key != NULL)
		free(fa_public_key);
	fa_public_key = (struct msg_key *)
		malloc(sizeof(struct msg_key) + SPI_LEN +
		       rsa_get_public_key_len() + 1);
	if (fa_public_key == NULL) {
		fprintf(stderr, "Not enough memory for fa_public_key\n");
		clean_up(1);
	}
	if (rsa_get_public_key_len() + SPI_LEN > 255) {
		fprintf(stderr, "fa_public_key over 255 bytes long - does "
			"not fit into extension\n");
		clean_up(1);
	}
	init_key_extension(fa_public_key, VENDOR_EXT_DYNAMICS_FA_PUBKEY,
			   htonl(PUBKEY_ALG_SPI_RSA),
			   rsa_get_public_key_len());
  	memcpy(MSG_KEY_DATA(fa_public_key), rsa_get_public_key(),
  	       GET_KEY_LEN(fa_public_key));
	if (fa_nai != NULL) {
		free(fa_nai);
		fa_nai = NULL;
	}
	if (config->fa_nai[0] != '\0') {
		fa_nai = (struct fa_nai_ext *)
			malloc(sizeof(struct fa_nai_ext) +
			       strlen(config->fa_nai));
		if (fa_nai == NULL) {
			DEBUG(DEBUG_FLAG, "Not enough memory for fa_nai\n");
			clean_up(1);
		}
		fa_nai->type = VENDOR_EXT_TYPE2;
		fa_nai->reserved = 0;
		fa_nai->length = sizeof(struct fa_nai_ext) - 2 +
			strlen(config->fa_nai);
		fa_nai->vendor_id = htonl(VENDOR_ID_DYNAMICS);
		fa_nai->sub_type = htons(VENDOR_EXT_DYNAMICS_FA_NAI);
		memcpy(fa_nai + 1, config->fa_nai, strlen(config->fa_nai));
	}

	last_challenges = (struct challenge_ext **)
		malloc(config->challenge_window *
		       sizeof(struct challenge_ext *));
	if (last_challenges == NULL) {
		DEBUG(DEBUG_FLAG, "Not enough memory for last_challenges\n");
		clean_up(1);
	}
	for (i = 0; i < config->challenge_window; i++)
		last_challenges[i] = NULL;
	last_challenges_pos = 0;
}


/* initializations that are run only once (not on config reloads) */
static void
init_data(void)
{
	struct interface_entry *iface;

	iface = (struct interface_entry *) list_get_first(&config->interfaces);
	if (iface == NULL) {
		LOG2(LOG_ALERT, "init_data: no interfaces defined\n");
		clean_up(1);
	}

	allow_ipv4_forwarding();

	/* Initialize hashes */
	tunnels_hash = tunnel_init(config->tunnel_device,
				   config->routing_table_start,
				   config->routing_table_end);
	if (tunnels_hash == NULL) {
		LOG2(LOG_ALERT,
		     "init_data: Tunnels hash not initialized!\n");
		clean_up(1);
	}

	bindings_hash = binding_init(config->max_bindings,
				     config->fa_default_tunnel_lifetime);
	if (bindings_hash == NULL) {
		LOG2(LOG_ALERT, "init_data: Could not initialize"
		     " bindings_hash\n");
		clean_up(1);
	}
	bcounters.bindingcount = 0;
	bcounters.pendingcount = 0;

	if (!fa_hash_init()) {
		LOG2(LOG_ALERT, "init_data: Could not initialize fa_hash\n");
		clean_up(1);
	}

	if (mn_addr_init() < 0) {
		LOG2(LOG_ALERT, "init_data: Could not initialize "
		     "mn_addr_hash\n");
		clean_up(1);
	}

	DEBUG(DEBUG_FLAG, "\tinit_data: Hashes initialized\n");

	iface = (struct interface_entry *) list_get_first(&config->interfaces);
	srand(time(NULL) ^ (iface ? iface->addr.s_addr : 0));

	/* Reset statistic counters */
	memset(&stats, 0, sizeof(stats));
	dynamics_strlcpy(stats.version, VERSION, sizeof(stats.version));

	/* Set up signal actions */
	signal(SIGHUP, reload_config);
	signal(SIGTERM, clean_up);    /* clean death */
	signal(SIGINT, clean_up);     /* clean death */
}


/* set_expr_timer:
 * @tv: Set to time when next event is due.
 * @next_agentadv: Time when next agent advertisement is due.
 *
 * Schedule next wake up time. The next wake up time can be caused
 * by any of the following events:
 * - a regular Agent advertisement is due
 * - a binding has expired
 * - a delayed Solicitation reply is due
 * - FA registration is due
 * - delayed tunnel deletion is due
 */
static int
set_expr_timer(struct timeval *tv, struct timeval next_agentadv) 
{
	int expire_time, diff;
	struct timeval solrep;
	struct timeval now;

	gettimeofday(&now, NULL);
	if (timerisset(&next_agentadv)) {
		tv->tv_sec = next_agentadv.tv_sec - now.tv_sec ;
		tv->tv_usec = next_agentadv.tv_usec - now.tv_usec;
		if (tv->tv_usec <0) {
			tv->tv_sec--;
			tv->tv_usec += 1000000;
		}
		if (tv->tv_sec < 0) {
			tv->tv_sec = 0;
			tv->tv_usec = 0;
		}
	} else {
		tv->tv_sec = -1;
		tv->tv_usec = 0;
	}

	if ((expire_time = binding_nextexpiretime(bindings_hash)) >= 0) {
		DEBUG(DEBUG_FLAG2, "set_expr_timer: binding next expire %d\n",
		      expire_time);
		if (tv->tv_sec < 0 || expire_time < tv->tv_sec) {
			tv->tv_sec = expire_time;
			tv->tv_usec = 0;
		}
	}

	solrep.tv_sec = 0;
	solrep.tv_usec = 0;
	check_delayed_solrep(&now, &solrep);
	if (solrep.tv_sec >= 0) {
		DEBUG(DEBUG_FLAG, "set_expr_timer: remaining delayed "
		      "solicitation reply\n");
		if (tv->tv_sec < 0 || solrep.tv_sec < tv->tv_sec ||
		    (solrep.tv_sec == tv->tv_sec && 
		     solrep.tv_usec < tv->tv_usec)) {
			tv->tv_sec = solrep.tv_sec;
			tv->tv_usec = solrep.tv_usec;
		}
	}

	diff = fa_reg_next_try - now.tv_sec;
	if (fa_reg_next_try > 0 && (tv->tv_sec == -1 || diff < tv->tv_sec))
		tv->tv_sec = diff < 0 ? 0 : diff;

	if (tv->tv_sec == -1 && tunnel_delayed_exists(tunnels_hash))
		tv->tv_sec = TUNNEL_DELAYED_CHECK_INTERVAL;

	return 0;
}


struct search_data {
	struct bindingentry *binding;
	struct unconfirmed_request *unc;
	struct in_addr addr;
};

static int
handle_dest_iter(struct node *node, void *data)
{
	struct bindingentry *binding;
	struct tunnel_data *t_data;
	struct search_data *search;
	struct unconfirmed_request *unc;

	binding = (struct bindingentry *) node;
	if (binding == NULL || data == NULL) return 0;
	t_data = (struct tunnel_data *) binding->data;
	if (t_data == NULL) return 0;
	search = (struct search_data *) data;

	unc = t_data->unc_req;
	while (unc != NULL) {
		if (unc->ha_addr.s_addr == search->addr.s_addr) {
			search->binding = binding;
			search->unc = unc;
			return 0;
		}
		unc = unc->next;
	}
	return 1;
}

static void
handle_dest_unreach(struct dest_unreachable_data *data)
{
	struct search_data search;
	struct msg_extensions ext;
	struct reg_req req;
	int reply_code;
	struct tunnel_data *t_data;

	if (!config->highest_FA)
		return;

	if (data->port != -1 && data->port != htons(config->ha_udp_port))
		return;

	search.binding = NULL;
	search.unc = NULL;
	search.addr = data->addr;
	binding_iterator(bindings_hash, handle_dest_iter, &search);

	if (search.unc == NULL)
		return;

	switch (data->code) {
	case ICMP_NET_UNREACH:
		reply_code = REGREP_HN_UNCREACHABLE_FA;
		break;
	case ICMP_HOST_UNREACH:
		reply_code = REGREP_HA_HOST_UNCREACHABLE_FA;
		break;
	case ICMP_PORT_UNREACH:
		reply_code = REGREP_HA_PORT_UNCREACHABLE_FA;
		break;
	default:
		reply_code = REGREP_HA_UNREACHABLE_FA;
		break;
	}

	DEBUG(DEBUG_FLAG, "handle_dest_unreach: found unconfirmed "
	      "request that resulted in HA unreachable\n");
	memset(&ext, 0, sizeof(ext));
	ext.req = &req;
	memset(&req, 0, sizeof(req));
	req.home_addr.s_addr = search.binding->mn_addr.s_addr;
	req.ha_addr.s_addr = search.unc->ha_addr.s_addr;
	req.id[0] = htonl(search.unc->id[0]);
	req.id[1] = htonl(search.unc->id[1]);
	send_failure_reply(reply_code, &ext, &search.unc->info, NULL, 0);

	/* no need to inform MN about pending request timing out after this
	 * error reply */
	t_data = (struct tunnel_data *) search.binding->data;
	if (t_data != NULL)
		t_data->pending_request = FALSE;
}

static int
fa_send_agent_adv(struct interface_entry *iface, struct sockaddr_ll *to,
		  struct in_addr *dest)
{
	__u16 adv_opts;
	__u8 own_adv_opts;
	int ret;
	struct challenge_ext *challenge = NULL;

	/* Set Agent Advertisement data */
	adv_opts = AGENT_ADV_FOREIGN_AGENT;
	if (config->reg_required)
		adv_opts |= AGENT_ADV_REGISTRATION_REQUIRED;
	if (config->enable_reverse_tunneling)
		adv_opts |= AGENT_ADV_BIDIR_TUNNELING;
	if (bcounters.bindingcount >= config->max_bindings ||
	    (config->max_pending > 0 &&
	     bcounters.pendingcount >= config->max_pending))
		adv_opts |= AGENT_ADV_BUSY;
	own_adv_opts = 0;
	if (config->enable_triangle_tunneling)
		own_adv_opts |= AGENT_ADV_OWN_TRIANGLE_TUNNELING;

	set_agent_adv_data(
		config->fa_default_tunnel_lifetime,
		config->highest_fa_addr,
		iface->addr, adv_opts, own_adv_opts,
		iface->interval > 21845 ?
		65535 : iface->interval * 3, config->hfa_pubkey_hash_len > 0 ?
		config->hfa_pubkey_hash : NULL,
		config->hfa_pubkey_hash_len);
	set_agent_adv_nai(fa_nai);
#ifdef INCLUDE_IPAY
	set_agent_adv_ipay(config->timePrice, config->bytePrice);
#endif

	if (config->enable_challenge_response) {
		challenge = create_challenge_ext(config,
						 AGENT_ADV_CHALLENGE_EXT);
		set_agent_adv_challenge(challenge);
	}
	ret = send_agent_advertisement(iface->icmp_sock,
				       (struct sockaddr *) to, dest,
				       iface->if_index);
	if (ret < 0) {
		DEBUG(DEBUG_FLAG, "Sending agent advertisement to interface "
		      "'%s' failed\n", iface->dev);
	}
	if (config->enable_challenge_response) {
		set_agent_adv_challenge(NULL);

		/* record the last challenge_window advertised challenges */
		if (last_challenges[last_challenges_pos] != NULL)
			free(last_challenges[last_challenges_pos]);
		last_challenges[last_challenges_pos] = challenge;
		last_challenges_pos++;
		if (last_challenges_pos >= config->challenge_window)
			last_challenges_pos = 0;
	}

	return ret;
}

/* send_agent_advs:
 * 
 * Send agent advertisements on interfaces. 
 *
 * Fill in next_agentadv with the time when the next advertisement
 * is due. 
 *
 * Updates iface.last_adv in the config.interfaces list.
 */
static void
send_agent_advs(struct timeval *next_agentadv)
{
	struct interface_entry *iface;
	struct node *node;
	struct timeval now, next;
	unsigned int diff;
	int first = 1;

	/* Send ICMP Agent Advertisement message and save the timestamp */
	gettimeofday(&now, NULL);
	next.tv_sec = 0;
	next.tv_usec = 0;
	for (node = list_get_first(&config->interfaces); node != NULL;
	     node = list_get_next(node)) {
		iface = (struct interface_entry *) node;
		if (iface->agentadv != INTERFACE_AGENTADV_ALL)
			continue;

		if (now.tv_sec > iface->last_adv.tv_sec + iface->interval - 1)
		{
			if (!fa_send_agent_adv(iface, NULL, NULL))
				stats.adv_sent++;
			iface->last_adv.tv_sec = now.tv_sec;
			iface->last_adv.tv_usec = now.tv_usec;
		}

		next.tv_sec  = iface->last_adv.tv_sec + iface->interval;
		/* Add random delay */
		next.tv_usec = iface->last_adv.tv_usec +
			(int) (MAX_ADV_DELAY * rand() / (RAND_MAX + 1.0));
		if (next.tv_usec > 999999) {
			next.tv_sec++;
			next.tv_usec -= 1000000;
		}
		if (first || next.tv_sec < next_agentadv->tv_sec) {
			DEBUG(DEBUG_FLAG, "** ");
			*next_agentadv = next;
		}
		first = 0;

		/* for debuging */
		diff = ((next.tv_sec - now.tv_sec) * 1000000 + 
			next.tv_usec - now.tv_usec) / 1000; 
		DEBUG(DEBUG_FLAG, "send_agent_advs: next agentadv: "
		      "%ld.%ld diff = %d msec\n", 
		      next.tv_sec, next.tv_usec, diff);
	}
}


static int
check_delayed_solrep(struct timeval *now, struct timeval *next)
{
	struct sol_reply_list *list = solrep_list;
	unsigned int passed = 0;

	assert(next != NULL);
	while (list != NULL && cmp_timeval(now, &list->send) == 1) {
		DEBUG(DEBUG_FLAG, "Found delayed solicitation reply\n");
		fa_send_agent_adv(list->iface, &list->to, &list->addr);
		solrep_list = list->next;
		free(list);
		list = solrep_list;
	}
	

	/* if any pending replys */
	if (list != NULL) {
		passed = 1000000 * (list->send.tv_sec - now->tv_sec) +
			list->send.tv_usec - now->tv_usec;
		DEBUG(DEBUG_FLAG, "Pending agent sol reply. %u usecs\n",
		      passed);
		set_usecs(next, passed);
	} else
		next->tv_sec = -1;

	return 0;
}

static int
add_agent_sol_reply(struct interface_entry *iface, struct sockaddr_ll *to,
		    struct in_addr *addr)
{
	struct sol_reply_list *new, *prev, *list = solrep_list;
	unsigned int usec_random =
		(int) (MAX_ADV_DELAY * rand() / (RAND_MAX + 1.0));
	
	assert(iface != NULL);
	new = (struct sol_reply_list *)
		calloc(1, sizeof(struct sol_reply_list));
	if (new == NULL) {
		DEBUG(DEBUG_FLAG, "add_agent_sol_reply: Out of memory\n");
		return -1;
	}

	gettimeofday(&new->send, NULL);
	add_usecs(&new->send, usec_random);
	new->iface = iface;
	memcpy(&new->to, to, sizeof(*to));
	new->addr.s_addr = addr->s_addr;

	if (list == NULL) {
		/* add first entry */
		solrep_list = new;
		new->next = NULL;

		return 0;
	}

	/* Find proper place */
	prev = NULL;
	while (list != NULL) {
		if(cmp_timeval(&list->send, &new->send) == 1)
			break;
		prev = list;
		list = list->next;
	}

	/* Add to the list */
	if (prev == NULL) {
		/* add to the beginning */
		new->next = solrep_list;
		solrep_list = new;
	} else {
		prev->next = new;
		new->next = list;
	}
	
	return 0;
}


static void
handle_icmp(struct interface_entry *iface)
{
	int r;
	struct sockaddr_ll from;
	struct in_addr to, fromaddr;

	r = check_icmp_sol(iface->icmp_sock, &from, &to, &fromaddr);

	if (r == 0 && iface->agentadv != INTERFACE_AGENTADV_NONE) {
		/* Agent solicitation */
		add_agent_sol_reply(iface, &from, &fromaddr);
	}
}


static void
main_loop(void)
{
	fd_set set;
	struct timeval tv;
	int oldmask;
	struct interface_entry *iface;
	struct node *node;
	struct timeval next_agentadv;

	next_agentadv.tv_sec = 0;
	next_agentadv.tv_usec = 0;
	send_agent_advs(&next_agentadv);

	/* Main loop
	 * Wait for incoming ICMP and UDP packets, and API calls. */
	for (;;) {
		fa_register(FA_REGISTER);

		FD_ZERO(&set);
		for (node = list_get_first(&config->interfaces); node != NULL;
		     node = list_get_next(node)) {
			iface = (struct interface_entry *) node;
			FD_SET(iface->udp_sock, &set);
			FD_SET(iface->udp_packet, &set);
			FD_SET(iface->icmp_sock, &set);
		}
#ifdef INCLUDE_IPAY
		FD_SET(ipay_sock, &set);
#endif
		if (api_rw_sock > -1)
			FD_SET(api_rw_sock, &set);
		if (api_ro_sock > -1)
			FD_SET(api_ro_sock, &set);

		/* Wake up when the next scheduled task timer expires or there
		 * is something in the sockets. */
		set_expr_timer(&tv, next_agentadv);
		DEBUG(DEBUG_FLAG2, "FA main: timer for select %ld.%ld\n",
		      tv.tv_sec, tv.tv_usec);
		if (select(FD_SETSIZE, &set, NULL, NULL,
			   tv.tv_sec == -1 ? NULL : &tv) < 0) {
			/* don't exit on e.g. SIGHUP */
			if (errno != EINTR) {
				LOG2(LOG_ERR, "select error: %s\n",
				     strerror(errno));
				clean_up(1);
			}
			continue;
		}

		/* we don't want to reread the configuration file
		 * while we are handling requests; hold the signal
		 * until we are ready */
		oldmask = sigblock(sigmask(SIGHUP));

		/* Check if any of the bindings have expired.
		 * Due to the binding module implementation this must be done
		 * just before using binding_add in order to get the binding
		 * module to current time. */
		check_bindings(bindings_hash, tunnels_hash, &bcounters);

#ifdef INCLUDE_IPAY
		if (FD_ISSET(ipay_sock, &set))
			handle_ipay(ipay_sock);
#endif

		if (api_rw_sock > -1 && FD_ISSET(api_rw_sock, &set)) {
			/* privileged API call */
			handle_api(api_rw_sock, 1);
		}

		if (api_ro_sock > -1 && FD_ISSET(api_ro_sock, &set)) {
			/* unprivileged API call */
			handle_api(api_ro_sock, 0);
		}

		for (node = list_get_first(&config->interfaces); node != NULL;
		     node = list_get_next(node)) {
			iface = (struct interface_entry *) node;

			if (FD_ISSET(iface->udp_packet, &set)) {
				DEBUG(DEBUG_FLAG, "Got UDP message(%s, packet "
				      "socket)\n", iface->dev);
				handle_reg_msg_packet(iface);
			}

			if (FD_ISSET(iface->udp_sock, &set)) {
				DEBUG(DEBUG_FLAG, "Got UDP message(%s)\n",
				      iface->dev);
				handle_reg_msg(iface);
			}

			if (FD_ISSET(iface->icmp_sock, &set)) {
				handle_icmp(iface);
			}
		}

		if (timerisset(&next_agentadv))
			send_agent_advs(&next_agentadv);

		tunnel_check_delayed(tunnels_hash, 0);
		fa_hash_expire();
		expire_denial_records(0);

#ifdef INCLUDE_IPAY
		ipay_send_statistics(ipay_sock);
#endif

		sigsetmask(oldmask);
	}
}


static int
fa_parse_command_line(int argc, char *argv[])
{
	int c, oindex = 0;
	static struct option long_options[] = {
		/* common arguments */
		{"help", no_argument, NULL, 'h'},
		{"version", no_argument, NULL, 'v'},
		{"debug", no_argument, NULL, 0},
		{"fg", no_argument, NULL, 0},
		{"config", required_argument, NULL, 'c'},
		/* FA specific arguments */
		/* end of arguments */
		{0, 0, 0, 0}
	};

	DEBUG(DEBUG_FLAG, "FA command line parsing\n");

	while ((c = getopt_long(argc, argv, "+hv", long_options, &oindex)) !=
	       EOF) {
		
		switch (c) {
		case 'h':
			/* show FA specific parameters */
			/* N/A */
			exit(1);
			break;
			
		case '?':
			printf("Command line parsing failed - aborting\n");
			exit(1);

		case 'v':
		case 'c':
		case 0:
			break;
			
		default:
			printf("?? getopt returned character code 0%o ??\n",
			       c);
			exit(1);
		}
	}

	return 0;
}

int
main(int argc, char *argv[])
{
	program_name = parse_long_options(argc, argv, "foreign agent",
					  PACKAGE, VERSION, dynamics_usage);

	fa_parse_command_line(argc, argv);

        config = malloc(sizeof(struct fa_config));
        if (config == NULL) {
                fprintf(stderr, "Not enough memory for struct fa_config\n");
                exit(1);
        }

        if (load_fa(config, program_name,
		    opt_config == NULL ? FA_GLOBAL_CONF_FILE : opt_config) ==
	    FALSE) {
                fprintf(stderr, "Configuration file reading failed.\n");
                exit(1);
        }

	check_kernel_support(CHECK_KERNEL_ADV_ROUTING | CHECK_KERNEL_IPIP |
			     CHECK_KERNEL_NETLINK);

#ifndef NO_SUID_CHECK_FOR_FA
	if (getuid()) {
		fprintf(stderr, "This program must be run by root.\n");
		exit(1);
	}
#endif 

	/* These routines don't return any values. If something is
           wrong clean_up() is used */
	init_sockets();
	init_data();
	init_config_data();

	if (!opt_foreground && dynamics_fork_daemon() == -1)
		clean_up(1);
	dynamics_write_pid_file(FA_PID_FILE);

	syslog(LOG_INFO, "%s foreign agent daemon version %s started\n",
	       PACKAGE, VERSION);

	main_loop();
        return 0;
}
