/* $Id: ha.c,v 1.204 2001/10/20 15:41:38 jm Exp $
 * Home Agent daemon
 *
 * Dynamic hierarchial IP tunnel
 * Copyright (C) 1998-2001, Dynamics group
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 * 
 * Additional modifications Copyright 2010 Juho Paaso, Aalto 
 * University / Department of Communication and Networking"
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define DEBUG_FLAG 'H'


#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <stdio.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <net/if.h>

#include <sys/time.h>
#include <unistd.h>
#include <syslog.h>
#ifndef __USE_BSD
#define __USE_BSD /* for sigblock et al. */
#endif
#include <signal.h>
#include <time.h>
#include <stdlib.h>
#include <assert.h>
#include <errno.h>
#include <getopt.h>
#include <sys/uio.h>
#include <linux/icmp.h>

#include "ha.h"
#include "list.h"
#include "hashtable.h"
#include "fileio.h"
#include "msgparser.h"
#include "agentadv.h"
#include "proxyarp.h"
#include "binding.h"
#include "auth.h"
#include "tunnel.h"
#include "debug.h"
#include "agentapi.h"
#include "util.h"
#include "fixed_fd_zero.h"
#include "dyn_ip.h"
#include "ha_config.h"
#include "agent_utils.h"
#include "md5_mac.h"


/* util.c - common command line arguments */
extern int opt_foreground;
extern char *opt_config;
int traces = 0;

/********************
 * defines
 ********************/

#define ASSERT assert
#define MIN(x, y) ( (x) < (y) ? (x) : (y))

#define LOG2(lev, fmt, args...) { DEBUG(DEBUG_FLAG, fmt, ## args); \
  syslog(lev, fmt, ## args); }

#define MAX_ADV_DELAY 200000.0 /* wait random 0 - MAX_ADV_DELAY microseconds 
				  befora sending agent advertisement (either 
				  broadcast or unicast, answer to agent 
				  solicitation */


/*********************
 * function prototypes
 *********************/

/* called before the home agent terminates to do some cleaning */
static void clean_up();

/* functions for handling requests, creating and destroying
 * tunnels (bindings) */
static void remove_tunnel(struct bindingentry *binding);
static void destroy_binding(struct bindingentry *binding);
static struct fa_spi_entry* get_fa_spi(int spi, struct in_addr addr);
static struct spi_entry* get_mn_spi(int spi);
static void init_interfaces(void);

/********************
 * global variables
 ********************/

static struct ha_config config;
struct fa_nai_ext *ha_nai = NULL;
static HASH *tunnels;
static struct bindingtable *bindings;	
static int bindingcount = 0;
static char *program_name; 				/* the name this program was run with */
static struct dynamics_ha_status stats;

/**********************************************************************/

extern int traces;
int use_virtual_home_nw = 1;
struct in_addr binded_hoas[HA_MOBILE_HASHTABLE_SIZE];		/* MR HoAs in the binding table */
struct in_addr allocated_hoas[HA_MOBILE_HASHTABLE_SIZE];	/* MR HoAs allocated from the pool */

/**
 * Initialize all s_addr elements to 0 which is interpreted to be a free slot.
 */
int init_hoa_pool() {
	int i;
	for (i = 0; i < HA_MOBILE_HASHTABLE_SIZE; i++) {
		allocated_hoas[i].s_addr = 0;
	}
	return 0;
}
/**
 * Returns 1 if HoA is free to allocate, 0 if it's not.  
 */
int hoa_is_free(struct in_addr *candidate) {
	int i;
	for (i = 0; i < HA_MOBILE_HASHTABLE_SIZE; i++) {
		if (allocated_hoas[i].s_addr == candidate->s_addr) {
			DEBUG(DEBUG_FLAG, "hoa_is_free: HoA candidate %s is already allocated.\n", inet_ntoa(*candidate));
			return 0;
		}
	}
	DEBUG(DEBUG_FLAG, "hoa_is_free: HoA candidate %s is free to allocate.\n", inet_ntoa(*candidate));
	return 1;
}
/**
 * Finds a free slot in the table and places hoa in it.
 * 
 * Returns 0 if ok, -1 if no free slot found.
 */
int allocate_hoa(struct in_addr *hoa) {
	int i;
	for (i = 0; i < HA_MOBILE_HASHTABLE_SIZE; i++) {
		if (allocated_hoas[i].s_addr == 0) {
			allocated_hoas[i].s_addr = hoa->s_addr;
			return 0;
		}
	}
	DEBUG(DEBUG_FLAG, "allocate_hoa: No free slot found from the HoA pool container!!\n");
	return -1;
}
/**
 * Finds the slot in the pool containing the given hoa and sets it to 0.
 * 
 * Returns 0 if ok, -1 if hoa not found in the pool.
 */
int free_hoa(struct in_addr *hoa) {
	int i;
	DEBUG(DEBUG_FLAG, "free_hoa: %s\n", inet_ntoa(*hoa));
	/* Find the HoA from the pool if it's not 0.0.0.0 */
	if (hoa->s_addr != 0) {
		for (i = 0; i < HA_MOBILE_HASHTABLE_SIZE; i++) {
			if (allocated_hoas[i].s_addr == hoa->s_addr) {
				allocated_hoas[i].s_addr = 0;
				return 0;
			}
		}
	}
	DEBUG(DEBUG_FLAG, "free_hoa: HoA %s not found from the pool!\n",
		inet_ntoa(*hoa));
	return -1;
}
/** 
 * get_hoa_from_pool: Place a non allocated HoA in the new_hoa.
 * 
 * Returns: 0 if ok, -1 if an available HoA was not found.
 */
int get_hoa_from_pool(struct in_addr *new_hoa) {

	int i;
	struct in_addr temp_hoa;
	temp_hoa.s_addr = config.hoa_pool_first.s_addr;
	
	for (i = 0; i <= config.hoa_range; i++) {
		if (hoa_is_free(&temp_hoa)) {
			allocate_hoa(&temp_hoa);
			new_hoa->s_addr = temp_hoa.s_addr;
			return 0;
		} else {
			/* Increase the last byte of the IP address by one. */
			((unsigned char *) &temp_hoa)[3]++;
		}
	}
	DEBUG(DEBUG_FLAG, "get_hoa_from_pool - No available HoA found from the pool!!\n");
	return -1;
}
/**
 * To global counter bindingcount has to be decreaset immediately after 
 * calling this function!
 */
int remove_from_binded_hoas(struct in_addr *mn_addr) {
	int i;
	
	for (i = 0; i < bindingcount; i++) {
		if (binded_hoas[i].s_addr == mn_addr->s_addr) {
			for (i; i < bindingcount - 1; i++) {
				binded_hoas[i].s_addr = binded_hoas[i + 1].s_addr;
			}
			return 1;
		}
	}
	DEBUG(DEBUG_FLAG, "[ha.c] remove_from_binded_hoas- HoA %s not found from binded_hoas!\n",
		inet_ntoa(*mn_addr));
	return -1;
}

/**
 * set_ha_nai:
 *
 * Setup global variable ha_nai according to config.ha_nai. set_ha_nai()
 * release the previously allocated buffer and allocates a new buffer for the
 * new HA NAI extension.
 */
static void
set_ha_nai(void)
{
	if (ha_nai != NULL) {
		free(ha_nai);
		ha_nai = NULL;
	}
	if (config.ha_nai[0] != '\0') {
		ha_nai = (struct fa_nai_ext *)
			malloc(sizeof(struct fa_nai_ext) +
			       strlen(config.ha_nai));
		if (ha_nai == NULL) {
			DEBUG(DEBUG_FLAG, "Not enough memory for ha_nai\n");
			clean_up(1);
		}
		ha_nai->type = VENDOR_EXT_TYPE2;
		ha_nai->reserved = 0;
		ha_nai->length = sizeof(struct fa_nai_ext) - 2 +
			strlen(config.ha_nai);
		ha_nai->vendor_id = htonl(VENDOR_ID_DYNAMICS);
		ha_nai->sub_type = htons(VENDOR_EXT_DYNAMICS_FA_NAI);
		memcpy(ha_nai + 1, config.ha_nai, strlen(config.ha_nai));
	}
}


/**
 * reload_config:
 *
 * SIGHUP handler for reloading the configuration.
 */
static void
reload_config(int sig)
{
	LOG2(LOG_INFO, "Received signal %d, rereading configuration\n", sig);
	cleanup_config(&config);

        if (load_config(&config, program_name,
			opt_config == NULL ? HA_GLOBAL_CONF_FILE : opt_config)
	    == FALSE) {
		LOG2(LOG_ALERT, "Reloading configuration file failed\n");
		clean_up(1);
	}
	init_interfaces();
	set_ha_nai();
	/* If the Port configuration is changed in the configuration
	 * file, it will have no effect on a reload. If this functionality
	 * is needed, the code should close the old sockets and reopen them
	 * here. */
}


/**********************************************************************
 * clean up stuff
 **********************************************************************/

/**
 * binding_hash_iterator:
 * @node: bindingentry
 * @dummy: not used
 *
 * Hashtable iterator for clean_up() function; releases the resources reserved
 * for the bindings.
 *
 * Returns:
 *  1 (to continue iterator loop)
 */
static int
binding_cleanup_iterator(struct node *node, void *dummy)
{
	struct bindingentry *data = (struct bindingentry *) node;

	ASSERT(data != NULL);
	
	binding_remove(data);
	remove_tunnel(data);
	destroy_binding(data);

	return 1;
}

/**
 * sig_handler:
 *
 * Called on various signals to end the process.
 */
static void
sig_handler(int sig)
{
	LOG2(LOG_INFO, "signal %d received, cleaning up\n", sig);
	clean_up(sig); /* calls clean_up */
}	


/**
 * clean_up:
 *
 * Clean up allocated resources and terminate HA daemon execution.
 */
static void
clean_up(int value)
{
	/* destroy all hashes and finish nicely */
	DEBUG(DEBUG_FLAG, "[ha.c] clean_up()\n");
	if (bindings != NULL) {
		binding_iterator(bindings, binding_cleanup_iterator, NULL);
		binding_destroy(bindings);
	}
	if (tunnels != NULL)
		tunnel_destroy_hash(tunnels);

	/* remove old unix sockets */
	if (config.ha_api_read_socket_path[0] != '\0')
		unlink(config.ha_api_read_socket_path);
	if (config.ha_api_admin_socket_path[0] != '\0')
		unlink(config.ha_api_admin_socket_path);

	if (unlink(HA_PID_FILE) < 0) {
		LOG2(LOG_INFO, "clean_up() could not remove PID file ("
		     HA_PID_FILE "): %s\n", strerror(errno));
	}

	cleanup_config(&config);
	if (ha_nai != NULL) {
		free(ha_nai);
		ha_nai = NULL;
	}

	syslog(LOG_INFO, "%s home agent daemon version %s stopped\n",
	       PACKAGE, VERSION);

	closelog();
	exit(value);
}

/**********************************************************************
 * API handling
 **********************************************************************/

/**
 * binding_get_tunnels_iter:
 * @node: pointer to bindingentry
 * @data: pointer to the message to be returned to the API caller
 *
 * Called from handle_api on an API_GET_TUNNELS request. This function is
 * called for each existing binding in the bindings list.
 *
 * Returns:
 *  1 (to continue iteration loop)
 */
static int
binding_get_tunnels_iter(struct node *node, void *data)
{
	struct bindingentry *binding = (struct bindingentry *) node;
	struct api_msg *api_msg = (struct api_msg *) data;
	dyn_tunnel_id id;

	if (api_msg->length + sizeof(dyn_tunnel_id) > API_DATA_SIZE) {
		DEBUG(DEBUG_FLAG, "API message overflow - skipping rest of "
		      "the bindings\n");
		return 0;
	}

	id.mn_addr.s_addr = binding->mn_addr.s_addr;
	id.ha_addr.s_addr = binding->ha_addr.s_addr;
	id.priv_ha = 0;

	memcpy((char *) api_msg->params + api_msg->length, &id,
	       sizeof(dyn_tunnel_id));

	api_msg->length += sizeof(dyn_tunnel_id);

	return 1;
}


/**
 * api_fetch_binding:
 * @msg: pointer to received API message
 *
 * Check that the dynamics_tunnel_id is in correct format and fetch the
 * requested binding.
 *
 * Returns:
 *   Pointer to the requested binding or NULL on failure.
 */
static struct bindingentry*
api_fetch_binding(struct api_msg *msg)
{
	dyn_tunnel_id *tid;
	struct bindingkey bkey;
	struct bindingentry *binding;

	if (msg->length != sizeof(dyn_tunnel_id)) {
		LOG2(LOG_ERR, "API message has wrong length (%i, expected %i)"
		     "\n", msg->length, sizeof(dyn_tunnel_id));
		return NULL;
	}
	tid = (dyn_tunnel_id *) msg->params;
	bkey.mn_addr = tid->mn_addr;
	if (config.sha_addr.s_addr != 0)
		bkey.ha_addr = config.sha_addr;
	else if (tid->ha_addr.s_addr != 0)
		bkey.ha_addr = tid->ha_addr;
	else {
		struct interface_entry *iface;
		iface = (struct interface_entry *)
			list_get_first(&config.interfaces);
		if (iface == NULL)
			return NULL;
		bkey.ha_addr = iface->addr;
	}
	bkey.priv_ha = config.priv_ha;
	binding = binding_fetch(bindings, &bkey);
	if (!binding) {
		DEBUG(DEBUG_FLAG, "   API call could not find the request "
		      "binding\n");
	}
	return binding;
}


/**
 * handle_api:
 * @socket: the socket where the request arrived
 * @priv: whether the request is privileged to do administrative actions
 *
 * Receives an API request, executes the requested tasks, and sends the result.
 *
 * Returns:
 *   0 on success, -1 on failure
 */
static int
handle_api(int socket, int priv)
{
	struct sockaddr_un cli_addr;
	socklen_t cli_len;
	int n;
	struct api_msg recv_msg;
	struct api_msg send_msg;
	struct bindingentry *binding;
	struct dynamics_tunnel_info *info;
	struct dynamics_ha_status *status;

	memset(&cli_addr, 0, sizeof(cli_addr));
	cli_len = sizeof(cli_addr);

	n = recvfrom(socket, &recv_msg, sizeof(struct api_msg), 0,
		     (struct sockaddr*)&cli_addr,
		     &cli_len);

	LOG2(LOG_DEBUG, "API: received %d bytes from %s\n", n,
	     cli_addr.sun_path);

	if (n < 0) {
		LOG2(LOG_ERR, "handle_api - recvfrom failed - %s\n",
		     strerror(errno));
		return -1;
	}

	if (priv)
		stats.apicalls_admin++;
	else
		stats.apicalls_read++;

	memset(&send_msg, 0, sizeof(struct api_msg));
	send_msg.type = API_REPLY_MSG;
	send_msg.code = API_UNDEFINED;
	send_msg.length = 0;

	switch (recv_msg.code) {
        case API_GET_TUNNELS:
			DEBUG(DEBUG_FLAG, "Received API_GET_TUNNELS\n");
			if (recv_msg.length != 0) {
				LOG2(LOG_ERR,
				     "API_GET_TUNNELS message has wrong length\n");
				send_msg.code = API_ILLEGAL_PARAMETERS;
				break;
			}
	
			/* get the list of bindings into send_msg */
			if (binding_iterator(bindings, binding_get_tunnels_iter,
					     &send_msg))
				send_msg.code = API_SUCCESS;
			else
				send_msg.code = API_INSUFFICIENT_SPACE;
			break;

        case API_GET_TUNNEL_INFO:
			DEBUG(DEBUG_FLAG, "Received API_GET_TUNNEL_INFO\n");
			binding = api_fetch_binding(&recv_msg);
			if (!binding) {
				send_msg.code = API_FAILED;
				break;
			}
			info = (struct dynamics_tunnel_info *) send_msg.params;
			memset(info, 0, sizeof(struct dynamics_tunnel_info));
			info->id.mn_addr.s_addr = binding->mn_addr.s_addr;
			info->id.ha_addr.s_addr = binding->ha_addr.s_addr;
			info->mn_addr.s_addr = binding->mn_addr.s_addr;
			info->co_addr.s_addr = binding->lower_addr.s_addr;
			info->ha_addr.s_addr = binding->ha_addr.s_addr;
			info->expiration_time = binding->exp_time;
			info->creation_time = binding->create_time;
			info->refresh_time = binding->mod_time;
			info->last_timestamp[0] = ntohl(binding->id[0]);
			info->last_timestamp[1] = ntohl(binding->id[1]);
			info->spi = binding->spi;
			info->confirmed = 1;
			info->timeout = binding->timeout;
			send_msg.code = API_SUCCESS;
			send_msg.length = sizeof(struct dynamics_tunnel_info);
			break;

		case API_DESTROY_TUNNEL:
			DEBUG(DEBUG_FLAG, "Received API_DESTROY_TUNNEL\n");
			if (!priv) {
				send_msg.code = API_NOT_PERMITTED;
				break;
			}
			binding = api_fetch_binding(&recv_msg);
			if (!binding) {
				send_msg.code = API_FAILED;
				break;
			}
			binding_remove(binding);
			remove_tunnel(binding);
			destroy_binding(binding);
			send_msg.length = 0;
			send_msg.code = API_SUCCESS;
			LOG2(LOG_NOTICE, "API call removed binding for MN %s\n",
			     inet_ntoa(binding->mn_addr));
			break;

		case API_GET_STATUS:
			DEBUG(DEBUG_FLAG, "Received API_GET_STATUS\n");
			if (recv_msg.length != 0) {
				LOG2(LOG_ERR, "API_GET_STATUS message has wrong "
				     "length\n");
				break;
			}
			status = (struct dynamics_ha_status *) send_msg.params;
			memcpy(status, &stats, sizeof(struct dynamics_ha_status));
			UNALIGNED_(&status->tunnel_count, &bindingcount);
			send_msg.length = sizeof(struct dynamics_ha_status);
			send_msg.code = API_SUCCESS;
			break;

		case API_RELOAD_CONFIG:
			/* Force a reload of the configuration */
			DEBUG(DEBUG_FLAG, "Received API_RELOAD_CONFIG\n");
			if (!priv) {
				send_msg.code = API_NOT_PERMITTED;
				break;
			}
			reload_config(1);
			DEBUG(DEBUG_FLAG, "Configuration_reloaded\n");
			send_msg.code = API_SUCCESS;
			break;

		default:
			LOG2(LOG_ERR, "Received unknown API command: %d\n",
			     recv_msg.code);
			send_msg.code = API_NOT_SUPPORTED;
			break;
	}

	return api_send(socket, &cli_addr, cli_len, &send_msg);
}

/**********************************************************************
 * Registration request handling
 **********************************************************************/

/**
 * pubkey_hash_ok:
 * @pubkey: pointer to public key extension
 * @hash: pointer to public key hash extension
 *
 * Check whether public key @pubkey matches with its @hash value.
 *
 * Returns:
 *   1 if key hash is correct or 0 if check failed
 */
static int
pubkey_hash_ok(struct msg_key *pubkey, struct msg_key *hash)
{
	unsigned char mac[MD5_MAC_LEN];

	if (GET_KEY_LEN(hash) != MD5_MAC_LEN) {
		DEBUG(DEBUG_FLAG, "pubkey_hash_ok: invalid hash length: %i, "
		      "expected %i\n", GET_KEY_LEN(hash), MD5_MAC_LEN);
		return 0;
	}

	md5_mac((unsigned char *) "", 0, (unsigned char *) pubkey,
		GET_KEY_EXT_LEN(pubkey), mac);
	return (memcmp(mac, MSG_KEY_DATA(hash), MD5_MAC_LEN) == 0);
}


/**
 * check_fa_auth_ext:
 * @addr: source IP address of the registration message
 * @ext: registration message extensions
 *
 * Check whether registration message FA/SHA-HA authentication is valid.
 *
 * Returns:
 *   0 if auth. ext. okay or 1 if fh_auth missing/invalid
 */
static int
check_fa_auth_ext(struct in_addr addr, const struct msg_extensions *ext)
{
	struct fa_spi_entry *fa_spi;

	fa_spi = get_fa_spi(0, addr);
	if (fa_spi == NULL) {
		/* no security association with the sender - authentication
		 * extension not required */
		if (ext->fh_auth != NULL) {
			DEBUG(DEBUG_FLAG, "No FA-HA security association, but "
			      "request had FA-HA authentication extension\n");
			return 1;
		}
		return 0;
	}

	if (ext->fh_auth == NULL && ext->sha_ha_auth != NULL &&
	    config.sha_addr.s_addr != 0) {
		DEBUG(DEBUG_FLAG, "Checking sha_ha_auth\n");
		if (auth_check_vendor(fa_spi->alg, fa_spi->shared_secret,
				      fa_spi->shared_secret_len,
				      (unsigned char *) ext->req,
				      ext->sha_ha_auth))
			return 0;
		return 1;
	}

	DEBUG(DEBUG_FLAG, "Checking fh_auth\n");
	if (ext->fh_auth == NULL ||
	    !auth_check(fa_spi->alg, fa_spi->shared_secret,
			fa_spi->shared_secret_len, (unsigned char *) ext->req,
			ext->fh_auth))
		return 1;
	else
		return 0;
}

#define AUTH_MAC_RFC2002 -100

/**
 * validate_request:
 * @binding: binding entry of the MN from which the registration message is
 * originating or %NULL if no binding for this MN available
 * @faaddr: source IP address of the registration message - jpaaso - == CoA
 * @msg: pointer to the start of registration message
 * @msg_len: length of the registration message
 * @ext: received message extensions
 * @code: pointer to buffer for failure code
 * @auth_type: pointer to buffer for used MN-HA authentication type
 *
 * Validate registration request. Check that it:
 * - has the necessary extensions
 * - comes from an authorized mobile node
 * - has a valid MAC in the authentication extension
 * - has the right HA address
 *
 * Returns:
 *  If everything is OK a pointer to the used SPI entry will be
 *  returned, and code will be REGREP_ACCEPTED or REGREP_ACCEPTED_NO_SB.
 *
 *  On failure @code is set to the value that should be sent in the failure
 *  reply. If code is negative no reply is to be sent. The returned value
 *  is %NULL if the SPI entry for MN is not found, otherwise it points to the
 *  SPI entry.
 *
 *  The auth_type is set to a special code AUTH_MAC_RFC2002 if
 *  the MAC of the mh_auth did not include the SPI field (RFC 2002).
 */
static struct spi_entry *
validate_request(struct bindingentry *binding, struct in_addr faaddr,
		 const char *msg, int msg_len,
		 const struct msg_extensions *ext, int *code, int *auth_type)
{
	struct node *node;
	struct spi_entry *spi = NULL;
	time_t current_time;
	enum { NOT_FOUND, UNICAST_FOUND, BROADCAST_FOUND } own_addr_found;
	struct authorized_entry *auth = NULL;
	char faaddrstr[16];		/* jpaaso - faaddr == CoA */
	char mnaddrstr[16];
	struct ha_tunnel_data *t_data = NULL;

	if (binding != NULL)
		t_data = (struct ha_tunnel_data *) binding->data;
	
	dynamics_strlcpy(faaddrstr, inet_ntoa(faaddr), sizeof(faaddrstr));

	*code = REGREP_ACCEPTED;
	*auth_type = 0;

	/* check that it has the right extensions */
	if (ext->req == NULL || ext->mh_auth == NULL) {
		LOG2(LOG_WARNING,
		     "FA %s: message does not contain either request "
		     "and/or message authentication\n", faaddrstr);
		if (ext->mh_auth == NULL)
			*code = REGREP_MN_FAILED_AUTH_HA;
		else
			*code = REGREP_BAD_REQUEST_HA;
		return NULL;
	}
	
	dynamics_strlcpy(mnaddrstr, inet_ntoa(ext->req->home_addr),
			 sizeof(mnaddrstr));
			 
	char ip_str[INET_ADDRSTRLEN];
	int i = 0;
	inet_ntop(AF_INET, &ext->req->home_addr, ip_str, INET_ADDRSTRLEN);
	DEBUG(DEBUG_FLAG, "Reg Home Addr: %s\n", ip_str);
	DEBUG(DEBUG_FLAG, "Auth list: ");
	/* check that the MN is authorized */
	for (node = list_get_first(&config.authorized_list); node != NULL;
	     node = list_get_next(node)) {
		auth = (struct authorized_entry *) node;
		// network to printable
		inet_ntop(AF_INET, &auth->network, ip_str, INET_ADDRSTRLEN);
		DEBUG(DEBUG_FLAG, "%d: %s, ", i, ip_str);
		i++;
		if (ntohl(ext->mh_auth->spi) >= auth->spi_low &&
		    ntohl(ext->mh_auth->spi) <= auth->spi_high &&
		    (ext->req->home_addr.s_addr & auth->netmask.s_addr) == auth->network.s_addr) {
			break;
		}
	}
	DEBUG(DEBUG_FLAG, "\n");
	if (node == NULL) {
		LOG2(LOG_WARNING, "FA %s, MN %s: Not authorized\n", 
		     faaddrstr, mnaddrstr);
		*code = REGREP_ADMIN_PROHIBITED_HA;
		return NULL;
	}

	/* check MAC (MN-HA authentication extension) */
	spi = get_mn_spi(ntohl(ext->mh_auth->spi));
	if (spi == NULL) {
		LOG2(LOG_WARNING, "FA %s, MN %s: SPI %d not found\n", 
		     faaddrstr, mnaddrstr, ntohl(ext->mh_auth->spi));
		*code = REGREP_MN_FAILED_AUTH_HA;
		return NULL;
	}
	
	if (!auth_check(spi->auth_alg, spi->shared_secret,
			spi->shared_secret_len,
			(unsigned char *) msg, ext->mh_auth)) {
		if (spi->auth_alg != AUTH_ALG_MD5 ||
		    !auth_check(AUTH_ALG_MD5_RFC2002, spi->shared_secret,
				spi->shared_secret_len,
				(unsigned char *) msg, ext->mh_auth)) {
			LOG2(LOG_WARNING, "FA %s, MN %s, SPI %d: Invalid MN-HA"
			     " authentication extension\n",
			     faaddrstr, mnaddrstr, spi->spi);
			*code = REGREP_MN_FAILED_AUTH_HA;
			return spi;
		} else {
			/* non-standard authentication method used */
			DEBUG(DEBUG_FLAG, "RFC 2002 MAC calculation "
			      "detected - trying to use the same method\n");
			*auth_type = AUTH_MAC_RFC2002;
		}
	}

	if (ext->double_auth_ext > 0) {
		LOG2(LOG_WARNING, "FA %s, MN %s: duplicate authentication "
		     "extension\n", faaddrstr, mnaddrstr);
		if (ext->double_auth_ext & DOUBLE_MH_AUTH)
			*code = REGREP_MN_FAILED_AUTH_HA;
		else if (ext->double_auth_ext & DOUBLE_FH_AUTH)
			*code = REGREP_FA_FAILED_AUTH_HA;
		else
			*code = REGREP_REASON_UNSPEC_HA;
		return spi;
	}

	/* check possible Foreign-Home Authentication Extension */
	if (check_fa_auth_ext(faaddr, ext) != 0) {
		LOG2(LOG_WARNING, "FA %s, MN %s: invalid/missing fh_auth\n",
		     faaddrstr, mnaddrstr);
		*code = REGREP_FA_FAILED_AUTH_HA;
		return spi;
	}

	/* check that MN-HA auth. ext. protects correct extensions */
	if (!auth_is_protected(ext->pubkey_hash, ext->mh_auth) ||
	    !auth_is_protected(ext->priv_ha, ext->mh_auth) ||
	    !auth_is_protected(ext->mn_keyreq, ext->mh_auth) ||
	    !auth_is_protected(ext->mn_nai, ext->mh_auth)) {
		LOG2(LOG_WARNING, "FA %s, MN %s: MN-HA auth. ext. does not "
		     "protect all the required extensions\n",
		     faaddrstr, mnaddrstr);
		*code = REGREP_MN_FAILED_AUTH_HA;
	}

	/* check that MN-AAA auth. ext. protects correct extensions */
	if (ext->mn_fa_key_req_aaa != NULL &&
	    (ext->mn_aaa_auth == NULL ||
	     !auth_is_protected(ext->mn_fa_key_req_aaa, ext->mn_aaa_auth))) {
		LOG2(LOG_WARNING, "FA %s, MN %s: MN-AAA auth. ext. does not "
		     "protect MN-FA keyreq from AAA extensions\n",
		     faaddrstr, mnaddrstr);
		*code = REGREP_MN_FAILED_AUTH_HA;
	}
	if (ext->mn_ha_key_req_aaa != NULL &&
	    (ext->mn_aaa_auth == NULL ||
	     !auth_is_protected(ext->mn_ha_key_req_aaa, ext->mn_aaa_auth))) {
		LOG2(LOG_WARNING, "FA %s, MN %s: MN-AAA auth. ext. does not "
		     "protect MN-HA keyreq from AAA extensions\n",
		     faaddrstr, mnaddrstr);
		*code = REGREP_MN_FAILED_AUTH_HA;
	}

	if (ext->fh_auth != NULL &&
	    !auth_is_protected(ext->challenge, ext->fh_auth)) {
		LOG2(LOG_WARNING, "FA %s, MN %s: FA-HA auth. ext. does not "
		     "protect all the required extensions\n",
		     faaddrstr, mnaddrstr);
		*code = REGREP_FA_FAILED_AUTH_HA;
	}

	/* check own (HA) address */
	own_addr_found = NOT_FOUND;
	for (node = list_get_first(&config.interfaces); node != NULL;
	     node = list_get_next(node)) {
		struct interface_entry *iface =
			(struct interface_entry *) node;
		if (iface->addr.s_addr == ext->req->ha_addr.s_addr) {
			own_addr_found = UNICAST_FOUND;
			break;
		}
		if (iface->ha_disc &&
		    iface->bcaddr.s_addr == ext->req->ha_addr.s_addr) {
			own_addr_found = BROADCAST_FOUND;
			break;
		}
	}

	if (own_addr_found != 1 &&
	    (config.sha_addr.s_addr == 0 ||
	     ext->req->ha_addr.s_addr != config.sha_addr.s_addr)) {
		if (own_addr_found == BROADCAST_FOUND)
			DEBUG(DEBUG_FLAG,
			      "==> HA discovery, deny with unknown HA\n");
		else
			LOG2(LOG_WARNING, "FA %s, MN %s: HA address does not "
			     "match\n", faaddrstr, mnaddrstr);
		*code = REGREP_UNKNOWN_HA_HA;
		return spi;
	}

	/* check option bits */
	if (ext->req->opts & REGREQ_SIMULTANEOUS_BINDINGS) {
		LOG2(LOG_INFO,
		     "FA %s, MN %s: Received request for unsupported "
		    "simultaneous_bindings\n", faaddrstr, mnaddrstr);
		/* this is ok, but inform the MN */
		*code = REGREP_ACCEPTED_NO_SB;
	}

	current_time = time(NULL);
	if (spi->replay_method == REPLAY_PROTECTION_TIMESTAMP &&
	    (ntohl(ext->req->id[0]) < current_time + UNIX_NTP_DIFF -
	     spi->timestamp_tolerance ||
	     ntohl(ext->req->id[0]) > current_time + UNIX_NTP_DIFF +
	     spi->timestamp_tolerance)) {
		time_t t;
		t = ntohl(ext->req->id[0]) - UNIX_NTP_DIFF;
		LOG2(LOG_WARNING, "ID mismatch - timestamp difference=%i, "
		     "tolerance=%i\n", abs(t - current_time),
		     spi->timestamp_tolerance);
		*code = REGREP_ID_MISMATCH_HA;
	} else if (spi->replay_method == REPLAY_PROTECTION_NONCE &&
		   t_data != NULL && ext->req->id[0] != t_data->nonce) {
		LOG2(LOG_WARNING, "ID mismatch (nonce)\n");
		DEBUG(DEBUG_FLAG, "\treq->id: %08x, expected nonce %08x\n",
		      ext->req->id[0], t_data->nonce);
		*code = REGREP_ID_MISMATCH_HA;
	}

	if (ext->req->opts & REGREQ_BROADCAST_DATAGRAMS) {
		DEBUG(LOG_INFO,
		      "FA %s, MN %s: Received request for unsupported "
		    "broadcast message forwarding\n", faaddrstr, mnaddrstr);
		/* this is not yet implemented in Dynamics HA, but accept the
		 * connection anyway; there is no accepted reply code that
		 * would indicate this situation, so MN will think that the
		 * request was accepted.. */
	}

	/* REGREQ_VJ_HC is ignored in HA, see RFC 2002, 3.8.2 */
	if (ext->req->opts & (REGREQ_MINIMAL_ENCAPS | REGREQ_GRE_ENCAPS)) {
		LOG2(LOG_INFO, "FA %s, MN %s: Received unsupported option "
		     "request (opts=0x%02x)\n", faaddrstr, mnaddrstr,
		     ext->req->opts);
		/* RFC 2344 defines new denial code for unavailable
		 * encapsulation */
		if ((ext->req->opts & REGREQ_REVERSE_TUNNEL) != 0)
			*code = REGREP_ENCAP_UNAVAIL_HA;
		else
			*code = REGREP_REASON_UNSPEC_HA;
		return spi;
	}

	if (!config.enable_reverse_tunneling &&
	    (ext->req->opts & REGREQ_REVERSE_TUNNEL) != 0) {
		LOG2(LOG_INFO,
		     "FA %s, MN %s: Asked forbidden reverse tunneling\n",
		     faaddrstr, mnaddrstr);
		*code = REGREP_REVERSE_TUNNEL_UNAVAIL_HA;
		return spi;
	}

	if (!config.enable_triangle_tunneling &&
	    (ext->req->opts & REGREQ_REVERSE_TUNNEL) == 0) {
		LOG2(LOG_INFO,
		     "FA %s, MN %s: Reverse tunneling mandatory\n",
		     faaddrstr, mnaddrstr);
		*code = REGREP_REVERSE_TUNNEL_MANDATORY_HA;
		return spi;
	}

	if (ext->req->opts & REGREQ_RESERVED) {
		LOG2(LOG_INFO, "FA %s, MN %s: Received a request with non-zero"
		     " reserved field\n", faaddrstr, mnaddrstr);
		*code = REGREP_BAD_REQUEST_HA;
		return spi;
	}

	if (config.pubkey_hash_method == HASH_METHOD_REQUIRE &&
	    ext->fa_pubkey != NULL && ext->pubkey_hash == NULL) {
		LOG2(LOG_INFO, "FA %s, MN %s: Received a request with public "
		     "key, but without public key hash extension\n",
		     faaddrstr, mnaddrstr);
		*code = REGREP_ADMIN_PROHIBITED_HA;
		return spi;
	}

	if (config.pubkey_hash_method != HASH_METHOD_NONE &&
	    ext->fa_pubkey != NULL &&
	    ext->pubkey_hash != NULL &&
	    !pubkey_hash_ok(ext->fa_pubkey, ext->pubkey_hash)) {
		LOG2(LOG_INFO, "FA %s, MN %s: Received a request with public "
		     "key and mismatching public key hash extension\n",
		     faaddrstr, mnaddrstr);
		*code = REGREP_ADMIN_PROHIBITED_HA;
		return spi;
	}

	/* request is OK */
	return spi;
}


/**
 * new_session_key:
 * @binding: pointer to bindingentry
 * @mn_spi: pointer to SPI entry for the MN
 *
 * Generate a new session key for the given MN (@binding,@mn_spi).
 *
 * Returns:
 *  0 on success or -1 on failure
 */
static int
new_session_key(struct bindingentry *binding, struct spi_entry *mn_spi)
{
#ifndef NODEBUGMODE
	int i;
#endif

	/* remove the (possibly) cached previous key reply using FA's public
	 * key, because the session key changes */
	if (binding->last_sent_fa_pubkeyrep)
		free(binding->last_sent_fa_pubkeyrep);
	binding->last_sent_fa_pubkeyrep = NULL;

	/* generate a new session key */
	binding->keylen =
		generate_session_key(mn_spi->auth_alg, mn_spi->shared_secret,
				     mn_spi->shared_secret_len, binding->key);
	if (binding->keylen == -1) {
		LOG2(LOG_ERR, "Could not generate session key\n");
		return -1;
	}

#ifndef NODEBUGMODE
	DEBUG(DEBUG_FLAG, "Session key: ");
	for (i = 0; i < binding->keylen; i++)
		DEBUG(DEBUG_FLAG, "%02x", binding->key[i]);
	DEBUG(DEBUG_FLAG, "\n");
#endif

	return 0;
}


struct gratuitous_arp_data {
	struct gratuitous_arp_data *next;
	struct timeval tv;
	struct in_addr addr;
	char interface[IFNAMSIZ + 1];
};

static struct gratuitous_arp_data *gratuitous_arp_queue = NULL;


/**
 * send_gratuitous_arp:
 * @addr: IP address for the ARP packet
 * @interface: interface to which the packet is sent
 *
 * Send a gratuitous ARP for an MN.
 */
static void
send_gratuitous_arp(struct in_addr addr, char *interface)
{
	struct gratuitous_arp_data *arp2, *arp3, *arp;

	/* HW broadcast is not usually guaranteed, so the gratuitous ARP is
	 * transmitted a small number of times (see RFC2002, sec. 4.6) */

	/* send first packet immediately */
	proxyarp_gratuitous(addr, interface);

	arp2 = (struct gratuitous_arp_data *)
		malloc(sizeof(struct gratuitous_arp_data));
	arp3 = (struct gratuitous_arp_data *)
		malloc(sizeof(struct gratuitous_arp_data));
	if (arp2 == NULL || arp3 == NULL) {
		DEBUG(DEBUG_FLAG, "send_gratuitous_arp: malloc() failed\n");
		if (arp2 != NULL)
			free(arp2);
		if (arp3 != NULL)
			free(arp3);
		return;
	}

	/* send second packet after 100 ms delay */
	memset(arp2, 0, sizeof(struct gratuitous_arp_data));
	gettimeofday(&arp2->tv, NULL);
	add_usecs(&arp2->tv, 100000);
	arp2->addr.s_addr = addr.s_addr;
	dynamics_strlcpy(arp2->interface, interface, sizeof(arp2->interface));

	/* send third packet after 200 ms delay */
	memcpy(arp3, arp2, sizeof(struct gratuitous_arp_data));
	add_usecs(&arp3->tv, 100000);

	DEBUG(DEBUG_FLAG, "send_gratuitous_arp - adding packet to queue at "
	      "%li.%06li and %li.%06li\n", arp2->tv.tv_sec, arp2->tv.tv_usec,
	      arp3->tv.tv_sec, arp3->tv.tv_usec);

	if (gratuitous_arp_queue == NULL) {
		/* empty queue - make new queue of the new items */
		gratuitous_arp_queue = arp2;
		arp2->next = arp3;
		arp3->next = NULL;
		return;
	}

	if (cmp_timeval(&gratuitous_arp_queue->tv, &arp2->tv) > 0) {
		/* arp2 goes before the old head of the queue */
		arp2->next = gratuitous_arp_queue;
		gratuitous_arp_queue = arp2;
	} else {
		/* insert arp2 into queue */
		arp = gratuitous_arp_queue;
		while (arp->next != NULL &&
		       cmp_timeval(&arp->next->tv, &arp2->tv) <= 0)
			arp = arp->next;
		arp2->next = arp->next;
		arp->next = arp2;
	}

	/* insert arp3 into queue */
	arp = arp2;
	while (arp->next != NULL &&
	       cmp_timeval(&arp->next->tv, &arp3->tv) <= 0)
		arp = arp->next;
	arp3->next = arp->next;
	arp->next = arp3;
}


/**
 * check_queued_gratuitous_arp:
 *
 * Send any pending gratuitous ARP packets from the queue.
 */
static void
check_queued_gratuitous_arp(void)
{
	struct timeval now;
	struct gratuitous_arp_data *arp;

	if (gratuitous_arp_queue == NULL)
		return;

	gettimeofday(&now, NULL);
	while (gratuitous_arp_queue != NULL &&
	       cmp_timeval(&gratuitous_arp_queue->tv, &now) <= 0) {
		arp = gratuitous_arp_queue;
		DEBUG(DEBUG_FLAG, "Sending queued gratuitous ARP "
		      "(tv=%li.%06li addr=%s interface=%s)\n",
		      arp->tv.tv_sec, arp->tv.tv_usec, inet_ntoa(arp->addr),
		      arp->interface);
		proxyarp_gratuitous(arp->addr, arp->interface);
		gratuitous_arp_queue = arp->next;
		free(arp);
	}
	if (gratuitous_arp_queue == NULL)
		DEBUG(DEBUG_FLAG, "Gratuitous ARP queue empty.\n");
	else
		DEBUG(DEBUG_FLAG, "Entries remaining in gratuitous ARP queue "
		      "(next at %li.%06li)\n",
		      gratuitous_arp_queue->tv.tv_sec,
		      gratuitous_arp_queue->tv.tv_usec);
}

/**
 * Checks that the networks don't overlap.
 * 
 * Returns: 
 *  1 if addresses don't belong to the same network, 0 if they do.
 */
int separate_networks(struct in_addr *addr1, struct in_addr *addr2) {
	int i;
	unsigned char octet1;
	unsigned char octet2;
	char ip1_str[INET6_ADDRSTRLEN];
	char ip2_str[INET6_ADDRSTRLEN];
	
	inet_ntop(AF_INET, addr1, ip1_str, INET6_ADDRSTRLEN);
	inet_ntop(AF_INET, addr2, ip2_str, INET6_ADDRSTRLEN);
		
	//DEBUG(DEBUG_FLAG, "separate_networks - %s == %s ? ",
	//	ip1_str, ip2_str);
	
	for (i = 0; i <= 3; i++) {
		octet1	= ((unsigned char *) addr1)[i];
		octet2	= ((unsigned char *) addr2)[i];
		/* Octets differ and are not zero */
		if (octet1 != octet2 && octet1 != 0 && octet2 != 0) {
			//DEBUG(DEBUG_FLAG, "Yes (octet number %d different: %u != %u).\n", 
			//	i, octet1, octet2);
			return 1;	
		}
	}
	//DEBUG(DEBUG_FLAG, "No.\n");
	return 0;
}

/**
 * nemo_nw_in_use:
 * @node: the new binding_entry
 * @data: variable of type struct network representing the NEMO network
 * 
 * Function binding_iterator calls this function for each binding. 
 */
int nemo_nw_in_use(struct node *node, void *data) {
	
	struct bindingentry *binding 	= (struct bindingentry *) node;
	struct network *nw_param		= (struct network *) data;
	struct network *nw_iter;
	int n;
	
	/* Iterate binding's nemo_nws */
	for (n = 0; n < binding->nemo_nw_count; n++) {
		nw_iter = &binding->nemo_nws[n];
		if (!separate_networks(&nw_param->prefix, &nw_iter->prefix)) {
			DEBUG(DEBUG_FLAG, "\t\tAddress %s already reserved for HoA %s!\n", 
				inet_ntoa(nw_param->prefix), inet_ntoa(binding->mn_addr));
			return 0;
		}
	}
	return 1;
}
/**
 * Copies nw prefix and length into the binding->nemo_nws[binding->nemo_nw_count] element. 
 */
int cp_to_nemo_nws(struct bindingentry *binding, struct in_addr *prefix, int prefix_length, 
					char *nai, int nai_length) 
{
	if (binding == NULL || prefix == NULL || nai == NULL) {
		DEBUG(DEBUG_FLAG, "\t\tcp_to_nemo_nws - Empty binding, prefix or nai (NULL)!!\n");
		return -1;
	}
	DEBUG(DEBUG_FLAG, "\t\tcp_to_nemo_nws - %s/%d, NAI %s (len %d)\n",
		inet_ntoa(*prefix), prefix_length, nai, nai_length);
	
	/* prefix */
	memcpy(&binding->nemo_nws[binding->nemo_nw_count].prefix, 
		prefix, sizeof(struct in_addr));
	
	/* prefix length */
	binding->nemo_nws[binding->nemo_nw_count].prefix_length = prefix_length;
	
	/* NEMO nw NAI */
	memcpy(binding->nemo_nws[binding->nemo_nw_count].nai, nai, nai_length);
	
	/* NEMO nw NAI length*/
	binding->nemo_nws[binding->nemo_nw_count].nai_length = nai_length;
	binding->nemo_nw_count++;
	return 0;
}

int cp_to_rejected_nemo_nws(struct bindingentry *binding, struct in_addr *prefix, int prefix_length,
							char *nai, int nai_length) 
{
	if (binding == NULL || prefix == NULL) {
			DEBUG(DEBUG_FLAG, "\t\treject_nemo_nw - Empty binding or prefix!!\n");
			return -1;
	}
	DEBUG(DEBUG_FLAG, "\t\tcp_to_rejected_nemo_nws - %s/%d, NAI %s (len %d)\n",
		inet_ntoa(*prefix), prefix_length, nai, nai_length);
	
	/* prefix */
	memcpy(&binding->rejected_nemo_nws[binding->rejected_nemo_nw_count].prefix, 
		prefix, sizeof(struct in_addr));
	
	/* prefix length */
	binding->rejected_nemo_nws[binding->rejected_nemo_nw_count].prefix_length =	prefix_length;
	
	/* NEMO nw NAI */
	memcpy(binding->rejected_nemo_nws[binding->rejected_nemo_nw_count].nai, nai, nai_length);
	
	/* NEMO nw NAI length*/
	binding->rejected_nemo_nws[binding->rejected_nemo_nw_count].nai_length = nai_length;
	binding->rejected_nemo_nw_count++;
}

/**
 * create_binding:
 * @ext: parsed registration request extensions
 * @mn_spi: pointer to SPI entry for the MN
 * @create_tunnels: 1=create tunnels, 0=create only binding data
 *
 * Creates a binding matching the given request.
 *
 * Returns:
 *  On success, a memory for the binding entry is allocated and a
 *  pointer to it returned. On failure, %NULL is returned.
 */
static struct bindingentry *
create_binding(struct msg_extensions *ext, struct spi_entry *mn_spi,
	       int create_tunnels)
{
	struct bindingentry *binding;
	struct ha_tunnel_data *t_data;
	char mn_addrstr[17];
	char co_addrstr[17];
	int ok = TRUE;
	TUNNEL *tunnel_entry_ptr;	
	
	DEBUG(DEBUG_FLAG, "Creating a new binding for MN\n");

	ASSERT(ext != NULL && ext->req != NULL);
	ASSERT(mn_spi != NULL);
	if (bindingcount >= config.max_bindings) {
		LOG2(LOG_WARNING, "max bindings reached\n");
		return NULL;
	}

	/* create the binding */
	binding = malloc(sizeof(struct bindingentry));
	if (binding == NULL) {
		LOG2(LOG_ERR, "malloc[binding] - %s\n", strerror(errno));
		return NULL;
	}
	memset(binding, 0, sizeof(struct bindingentry));
	binding->data = malloc(sizeof(struct ha_tunnel_data));
	if (binding->data == NULL) {
		LOG2(LOG_ERR, "malloc[ha_tunnel_data] - %s\n",
		     strerror(errno));
		free(binding);
		return NULL;
	}
	memset(binding->data, 0, sizeof(struct ha_tunnel_data));
	t_data = (struct ha_tunnel_data *) binding->data;

	binding->spi = mn_spi->spi;

	if ((ext->fa_keyreq || ext->fa_pubkey || ext->mn_keyreq) &&
	    new_session_key(binding, mn_spi) < 0)
		ok = FALSE;

	binding->mn_addr.s_addr = ext->req->home_addr.s_addr;
	binding->ha_addr.s_addr = ext->req->ha_addr.s_addr;
	binding->priv_ha = config.priv_ha;
	binding->lower_addr.s_addr = config.sha_addr.s_addr != 0 ?	/* jpaaso - CoA */
		config.sha_addr.s_addr : ext->req->co_addr.s_addr;
	dynamics_strlcpy(mn_addrstr, inet_ntoa(ext->req->home_addr),
			 sizeof(mn_addrstr));
	dynamics_strlcpy(co_addrstr, inet_ntoa(ext->req->co_addr),
			 sizeof(co_addrstr));
	memcpy(&binding->id, ext->req->id, REG_REQ_ID_LEN);
	binding->create_time = time(NULL);
	binding->lower_port = ext->mr_port;
	DEBUG(DEBUG_FLAG, "create_binding: create tunnel to %s:%d\n",
		inet_ntoa(binding->lower_addr), ntohs(binding->lower_port));
		
	tunnel_entry_ptr = tunnel_add2(tunnels, binding->lower_addr, binding->lower_port, 
		binding->tun_dev, ext->req->ha_addr, 1, TUNNEL_IPIP, 0);
	
	/* Save the socket to this MR CoA */
	tunnel_entry_ptr->tun_socket = ext->registration_socket;
	
	if (ok && create_tunnels &&
	     tunnel_entry_ptr == NULL) {
		LOG2(LOG_ERR, "Could not add tunnel to FA %s (COA=%s) for "
		     "MN %s\n", inet_ntoa(binding->lower_addr), co_addrstr,
		     mn_addrstr);
		ok = FALSE;
	}

	if (ext != NULL && ext->mn_nai != NULL) {
		strncpy(binding->nai, MSG_MN_NAI_DATA(ext->mn_nai), ext->mn_nai->length);
		binding->nai[ext->mn_nai->length] = '\0';
		binding->nai_length = ext->mn_nai->length;
		DEBUG(DEBUG_FLAG, "MN NAI: %s\n", binding->nai);
	}

	if (!use_virtual_home_nw) {
		if (ok && create_tunnels &&
		    dyn_ip_route_get(binding->mn_addr, t_data->arp_if, IFNAMSIZ)
		    != 0) {
			LOG2(LOG_ERR, "Could not get arp interface for MN %s\n",
			     mn_addrstr);
			ok = FALSE;
		}
	}

	/*--------------------------------------------------------------*
	 | Add route to the MR HoA through the tunnel device.			|
 	 *--------------------------------------------------------------*/
	if (ok && create_tunnels &&
	    dyn_ip_route_replace(binding->mn_addr, binding->tun_dev) != 0) 
	{
		LOG2(LOG_ERR, "Could not add route to MN %s\n",
		     mn_addrstr);
		ok = FALSE;
	}

	/*------------------------------------------------------------------*
	 | Allocate and add a route for the NEMO nw, if the req was ok.		|
 	 *------------------------------------------------------------------*/
	binding->nemo_nw_count = 0;
	binding->rejected_nemo_nw_count = 0;
	if (ok && create_tunnels && ext->nemo_req != NULL) 
	{
		int i;
		struct network nemo_nw;
		DEBUG(DEBUG_FLAG, "NEMO req extension - MN has %d mobile network(s)\n", ext->nemo_nw_amount);
		for (i = 0; i < ext->nemo_nw_amount; i++) {
			nemo_nw.prefix.s_addr = ext->nemo_req[i]->prefix;
			nemo_nw.prefix_length = ext->nemo_req[i]->prefix_length;
			
			DEBUG(DEBUG_FLAG, "\tnw %d: %s/%d..\n", i, inet_ntoa(nemo_nw.prefix), nemo_nw.prefix_length);
			
			/* Iterate through each binding entry and call nemo_nw_in_use for each with nemo_nw
			 * as the second argument. Returns TRUE if the iteration was not stopped by iterator 
			 * function and FALSE if it was stopped.
			 */
			if (!binding_iterator(bindings, nemo_nw_in_use, &nemo_nw)) {
				/*------------------------------------------------------------------*
	 			 | NEMO nw already in use.											|
 	 			 *------------------------------------------------------------------*/				
				/* Insert in binding->rejected_nemo_nws so that MN can be informed about 
				 * it with a NEMO Ack Extension. */
				cp_to_rejected_nemo_nws(binding, (struct in_addr *) &ext->nemo_req[i]->prefix, 
					ext->nemo_req[i]->prefix_length, "nai_x", 6); /* ----- NAI ------------ */
			} else {
				/*------------------------------------------------------------------*
	 			 | NEMO nw to be allocated for this MR.								|
 	 			 *------------------------------------------------------------------*/	
				DEBUG(DEBUG_FLAG, "\t\tAdd a route to NEMO network %s/%d via device %s\n", 
					inet_ntoa(*(struct in_addr *) &ext->nemo_req[i]->prefix), 
						ext->nemo_req[i]->prefix_length, binding->tun_dev);
				/*------------------------------------------------------------------*
	 			 | Add route to the NEMO network through the TUN device of this MR.	|
 	 			 *------------------------------------------------------------------*/
				/*if (dyn_ip_route_add_net(binding->tun_dev, (struct in_addr *) &ext->nemo_req[i]->prefix, 
					(int) ext->nemo_req[i]->prefix_length) != 0)*/
				if (dyn_ip_route_replace_net(binding->tun_dev, (struct in_addr *) &ext->nemo_req[i]->prefix, 
					(int) ext->nemo_req[i]->prefix_length) != 0)
				{
					cp_to_rejected_nemo_nws(binding, (struct in_addr *) &ext->nemo_req[i]->prefix, 
						ext->nemo_req[i]->prefix_length, "nai_x", 6); /* ----- NAI ------------ */
					DEBUG(DEBUG_FLAG, "\t\t- NEMO route couldn't be added!\n");
				} else {
					cp_to_nemo_nws(binding, (struct in_addr *) &ext->nemo_req[i]->prefix, 
						ext->nemo_req[i]->prefix_length, "nai_x", 6); /* ----- NAI ----- */
				}
			}
		}
	}
	
	/*------------------------------------------------------------------*
	 | Handle HAaRO extensions.											|
 	 *------------------------------------------------------------------*/
	if (ext->ro_capab.s_on 	!= NULL && 
		ext->ro_capab.s_off == NULL) 
	{
		/* S = 1 */
		DEBUG(DEBUG_FLAG, "MR Route Optimization Capability (HAaRO) Extension with S=1.\n");
		binding->ro_capable 		= 1;
		binding->adv_my_networks 	= ext->ro_capab.s_on->flags & RO_CAPAB_A;
		binding->not_reachable		= ext->ro_capab.s_on->flags & RO_CAPAB_O;
		
	} else if ( ext->ro_capab.s_off != NULL && 
				ext->ro_capab.s_on 	== NULL) 
	{
		/* S = 0 */ 
		DEBUG(DEBUG_FLAG, "MR Route Optimization Capability (HAaRO) Extension with S=0.\n");
		binding->ro_capable 		= 1;
		binding->adv_my_networks 	= ext->ro_capab.s_off->flags & RO_CAPAB_A;
		binding->not_reachable		= ext->ro_capab.s_off->flags & RO_CAPAB_O;
		
	} else if (	ext->ro_capab.s_off != NULL && 
				ext->ro_capab.s_on 	!= NULL) 
	{
		DEBUG(DEBUG_FLAG, "Warning: MR RO Capab (HAaRO) Ext parsed incorrectly!\n");
		
	} else {
		DEBUG(DEBUG_FLAG, "MR not capable for Route Optimization (HAaRO).\n");
		binding->ro_capable = 0;
	}
	
	/*--------------------------------------------------------------*
	 | If not ok, free the memory and return NULL					|
 	 *--------------------------------------------------------------*/
	if (!ok) {
		free(binding->data);
		free(binding);
		return NULL;
	} else {
		memcpy(&binded_hoas[bindingcount], &binding->mn_addr, sizeof(struct in_addr));
		DEBUG(DEBUG_FLAG, "binded_hoas[%d]: %s\n", bindingcount, inet_ntoa(binded_hoas[bindingcount]));
	}
	if (create_tunnels) {
		DEBUG(DEBUG_FLAG, "tunnel_add => tun_dev=[%s]\n",
		      binding->tun_dev);
		if (!use_virtual_home_nw) {
			DEBUG(DEBUG_FLAG, "dyn_ip_route_get: %s => %s\n",
			      inet_ntoa(binding->mn_addr), t_data->arp_if);
			proxyarp_add_item(binding->mn_addr, t_data->arp_if);
			send_gratuitous_arp(binding->mn_addr, t_data->arp_if);
		}	
	}

	bindingcount++;
	
	return binding;
}


/**
 * remove_tunnel:
 * @binding: bindingentry whose tunnel is to be removed
 *
 * Removes the route, tunnel, and proxy ARP entry for the @binding that were
 * created in create_binding() and possibly updated with switch_tunnel().
 */
static void
remove_tunnel(struct bindingentry *binding)
{
	struct ha_tunnel_data *t_data;

	ASSERT(binding != NULL && binding->data != NULL);

	t_data = (struct ha_tunnel_data *) binding->data;
	
	if (dyn_ip_route_del(binding->mn_addr, binding->tun_dev) < 0)
		LOG2(LOG_WARNING, "remove_tunnel: dyn_ip_route_del failed\n");
	if (tunnel_delete(tunnels, binding->lower_addr, 0, TUNNEL_IPIP, 0) < 0) {
		LOG2(LOG_WARNING, "remove_tunnel: tunnel_delete failed\n");
	}
	if (!use_virtual_home_nw) {
		if (proxyarp_del_item(binding->mn_addr, t_data->arp_if) < 0)
			LOG2(LOG_WARNING, "remove_tunnel: proxyarp_del_item failed\n");
	}
}


/**
 * switch_tunnel:
 * @binding: the bindingentry to be updated
 * @ext: parsed registration request extensions
 * @mn_spi: pointer to SPI entry for the MN
 *
 * Switch the tunnel, i.e., do optimized remove_tunnel() and create_binding().
 *
 * Returns:
 *  0 on success or -1 on failure
 */
static int
switch_tunnel(struct bindingentry *binding, struct msg_extensions *ext,
	      struct spi_entry *mn_spi)
{
	char mn_addrstr[17];
	char co_addrstr[17];
	char old_lower_tunl[IFNAMSIZ];
	TUNNEL *new_tunnel;	
	
	ASSERT(binding != NULL && ext != NULL && ext->req != NULL);

	dynamics_strlcpy(mn_addrstr, inet_ntoa(ext->req->home_addr),
			 sizeof(mn_addrstr));
	dynamics_strlcpy(co_addrstr, inet_ntoa(ext->req->co_addr),
			 sizeof(co_addrstr));
	memcpy(old_lower_tunl, binding->tun_dev, IFNAMSIZ);

	/* remove old tunnel (delayed deletion) */
	if (tunnel_delete(tunnels, binding->lower_addr, 0, TUNNEL_IPIP, 0) < 0)
	{
		LOG2(LOG_ERR,
		     "Could not delete old tunnel to FA %s for MN %s\n",
		     co_addrstr, mn_addrstr);
	} 

	binding->lower_addr = config.sha_addr.s_addr != 0 ?
		config.sha_addr : ext->req->co_addr;
		
	DEBUG(DEBUG_FLAG, "switch_tunnel: create tunnel to %s:%d\n",
		inet_ntoa(binding->lower_addr), ntohs(binding->lower_port));
	
	binding->lower_port = ext->mr_port;
	/* add the new tunnel */
	new_tunnel = tunnel_add2(tunnels, binding->lower_addr, binding->lower_port, 
		binding->tun_dev, ext->req->ha_addr, 1, TUNNEL_IPIP, 0);
	if (new_tunnel == NULL) {
		LOG2(LOG_ERR, "Could not add tunnel to FA %s (COA=%s) for "
		     "MN %s\n", inet_ntoa(binding->lower_addr), co_addrstr,
		     mn_addrstr);
		destroy_binding(binding);
		return -1;
	}
	new_tunnel->tun_socket = ext->registration_socket;

	/*----------------------------------------------------------*
	 | Copy the new tun_fd into the bindingtable				|
	 *----------------------------------------------------------*/
	memcpy(binding->tun_dev, new_tunnel->device, sizeof(binding->tun_dev));
	
	DEBUG(DEBUG_FLAG, "tunnel_add => tun_dev %s\n", 
		binding->tun_dev);
	
	/*----------------------------------------------------------*
	 | Add route to MR CoA and possible NEMO networks.			|
	 *----------------------------------------------------------*/
	DEBUG(DEBUG_FLAG, "Add route to the new MR CoA and to it's %d NEMO nws through device %s.\n",
		binding->nemo_nw_count, binding->tun_dev);
	
	/* Replace the route to MR HoA to go to dev. */
	if (dyn_ip_route_replace(binding->mn_addr, binding->tun_dev) != 0) {
		LOG2(LOG_WARNING,
		     "Route replace failed (COA=%s, MN=%s, dev=%s)\n",
		     co_addrstr, mn_addrstr, binding->tun_dev);
	}
	int n;
	for (n = 0; n < binding->nemo_nw_count; n++) 
	{
		if (dyn_ip_route_replace_net(binding->tun_dev, &binding->nemo_nws[n].prefix, 
				binding->nemo_nws[n].prefix_length) != 0) 
		{
			DEBUG(DEBUG_FLAG, "Route replace failed (route to NEMO nw %s/%d through %s)\n",
				inet_ntoa(binding->nemo_nws[n].prefix), binding->nemo_nws[n].prefix_length,
					binding->tun_dev);
		}
	}

	if ((ext->fa_keyreq || ext->fa_pubkey || ext->mn_keyreq) &&
	    new_session_key(binding, mn_spi) < 0) {
		destroy_binding(binding);
		return -1;
	}

	return 0;
}

/**
 * destroy_binding:
 * @binding: the bindingentry to be freed
 *
 * Frees the associated memory for a binding.
 * remove_tunnel() should be called before this function.
 */
static void
destroy_binding(struct bindingentry *binding)
{
	remove_from_binded_hoas(&binding->mn_addr);
	free_hoa(&binding->mn_addr);

	if (binding->data) /* ccmalloc needs this */
		free(binding->data);
	if (binding->fa_pubkey)
		free(binding->fa_pubkey);
	if (binding->last_sent_fa_pubkeyrep)
		free(binding->last_sent_fa_pubkeyrep);
	free(binding);
	
	bindingcount--;
}


/**
 * check_bindings:
 *
 * Checks if any bindings have expired since the last call to the function.
 * Any expired bindings will be removed.
 */
static void
check_bindings(void)
{
	struct bindingentry *binding;
	time_t diff;
	static int first_time = 1;
	static time_t advance_expire_time;
	time_t time2;

	if (first_time) {
		advance_expire_time = time(NULL);
		first_time = 0;
	}
	time2 = time(NULL);
	diff = time2 - advance_expire_time;
	advance_expire_time = time2;
	if (diff < 0) {
		DEBUG(DEBUG_FLAG,
		      "check_bindings: diff(%li) < 0 - system time changed?\n",
		      diff);
		diff = 0;
	}
	binding = binding_getexpired(bindings, diff);
	while (binding != NULL) {
		DEBUG(DEBUG_FLAG, "Binding to MN %s expired\n",
		      inet_ntoa(binding->mn_addr));
		
		remove_tunnel(binding);
		destroy_binding(binding);
		binding = binding_getexpired(bindings, 0);
	}
}


/**
 * get_fa_spi:
 * @spi: security parameter index
 * @addr: IP address of the FA
 *
 * Fetches SPI structure for an FA using key (@spi,@addr). If @spi is zero,
 * any item with a matching IP address is returned regardless of the configured
 * SPI value.
 *
 * Returns:
 *  pointer to the SPI structure or %NULL if requested item not found
 */
static struct fa_spi_entry *
get_fa_spi(int spi, struct in_addr addr)
{
	struct fa_spi_entry *s;
	struct node *node;

	for (node = list_get_first(&config.fa_spi_list); node != NULL;
	     node = list_get_next(node)) {
		s = (struct fa_spi_entry *) node;
		if ((spi == 0 || s->spi == spi) &&
		    s->addr.s_addr == addr.s_addr) {
			return s;
		}
	}
	return NULL;
}


/**
 * get_mn_spi:
 * @spi: security parameter index (in host byte order)
 *
 * Fetches SPI structure for an MN using key @spi.
 *
 * Returns:
 *  pointer to the SPI structure or %NULL if requested item not found
 */
static struct spi_entry *
get_mn_spi(int spi)
{
	struct spi_entry *s;
	struct node *node;

	for (node = list_get_first(&config.spi_list); node != NULL;
	     node = list_get_next(node)) {
		s = (struct spi_entry *) node;
		if (s->spi == spi)
			return s;
	}
	return NULL;
}


/**
 * ha_add_ext_start:
 * @start: pointer to the beginning of the authenticated data
 * @pos: memory position for the added authentication extension
 * @left: amount of free base in buffer
 * @ext: parsed registration request or %NULL if not available
 * @lifetime: lifetime of the binding; zero for deregistration or failure reply
 * @mn_spi: SPI structure for the MN or %NULL if not available
 * @binding: bindingentry for the MN or %NULL if not available
 * @auth_type: authentication data area (%AUTH_RFC2002 / %AUTH_RFC2002BIS)
 *
 * Add private HA extension (if SHA is used), copy MN NAI extension from the
 * request (if available), add a key reply extension (if requested), and
 * add MN-HA authentication extension if security associated available.
 *
 * Returns:
 *  Number of used bytes in the buffer or -1 on failure.
 */
static int
ha_add_ext_start(unsigned char *start, unsigned char *pos, int left,
		 struct msg_extensions *ext, int lifetime,
		 struct spi_entry *mn_spi, struct bindingentry *binding,
		 int auth_type)
{
	int used = 0, n;

	if (config.priv_ha > 0) {
		struct priv_ha_ext *priv = (struct priv_ha_ext *) (pos + used);
		DEBUG(DEBUG_FLAG, " * priv_ha\n");
		if (left < sizeof(struct priv_ha_ext))
			return -1;
		memset(priv, 0, sizeof(struct priv_ha_ext));
		priv->type = VENDOR_EXT_TYPE2;
		priv->length = sizeof(struct priv_ha_ext) - 2;
		priv->vendor_id = htonl(VENDOR_ID_DYNAMICS);
		priv->sub_type = htons(VENDOR_EXT_DYNAMICS_PRIV_HA);
		priv->priv_ha = htonl(config.priv_ha);
		used += GET_PRIV_HA_EXT_LEN(priv);
		left -= GET_PRIV_HA_EXT_LEN(priv);
	}

	/* RFC 2794 requires that HA copies MN NAI extension to the reply */
	if (ext != NULL && ext->mn_nai != NULL) {
		DEBUG(DEBUG_FLAG, " * mn_nai\n");
		if (left < GET_MN_NAI_EXT_LEN(ext->mn_nai))
			return -1;
		memcpy(pos + used, ext->mn_nai,
		       GET_MN_NAI_EXT_LEN(ext->mn_nai));
		used += GET_MN_NAI_EXT_LEN(ext->mn_nai);
		left -= GET_MN_NAI_EXT_LEN(ext->mn_nai);
	}

	/* Add the session key to mobile node, if it requested for a key and
	 * the reply is not a deregistration or failure reply */
	if (ext && ext->mn_keyreq != NULL && mn_spi && lifetime != 0) {
		struct msg_key *key;
		DEBUG(DEBUG_FLAG, " * mn_keyrep\n");
		if (left < sizeof(struct msg_key) + MAX_SK_LEN)
			return -1;
		key = (struct msg_key *) (pos + used);
		n = auth_encrypt(mn_spi->auth_alg, mn_spi->shared_secret,
				 mn_spi->shared_secret_len, binding->key,
				 key, (struct reg_rep *) start,
				 VENDOR_EXT_DYNAMICS_MN_KEYREP,
				 htonl(mn_spi->spi));
		used += n;
		left -= n;
	}

	if (ext && mn_spi) {
		struct msg_auth *auth;
		DEBUG(DEBUG_FLAG, " * mh_auth\n");
		if (left < sizeof(struct msg_auth) + MAX_SK_LEN)
			return -1;
		/* Add message authentication for MN */
		auth = (struct msg_auth *) (pos + used);
		n = auth_add((auth_type == AUTH_RFC2002BIS ?
			      AUTH_ALG_MD5_RFC2002 : mn_spi->auth_alg),
			     mn_spi->shared_secret,
			     mn_spi->shared_secret_len, start,
			     (struct msg_auth *) (pos + used), MH_AUTH,
			     htonl(mn_spi->spi));
		used += n;
		left -= n;
	}

	/* RFC 3012 requires that HA copies challenge extension to the reply
	 * after Mobile-Home auth. ext.; this will be authenticated with
	 * FA-HA auth. ext. if security association for this is available */
	if (ext != NULL && ext->challenge != NULL) {
		DEBUG(DEBUG_FLAG, " * challenge\n");
		n = GET_CHALLENGE_EXT_LEN(ext->challenge);
		if (left < n)
			return -1;
		memcpy(pos + used, ext->challenge, n);
		used += n;
		left -= n;
	}

	return used;
}


/**
 * ha_add_ext_end:
 * @start: pointer to the beginning of the authenticated data
 * @pos: memory position for the added authentication extension
 * @addr: IP address from which the registration request was received
 * @left: amount of free base in buffer
 * @ext: parsed registration request or %NULL if not available
 *
 * Copy nonce extension from the request (if available) and add an
 * authentication extension for SHA or FA if a security association is
 * configured.
 *
 * Returns:
 *  Number of used bytes in the buffer or -1 on failure.
 */
static int
ha_add_ext_end(unsigned char *start, unsigned char *pos, struct in_addr addr,
	       int left, struct msg_extensions *ext)
{
	struct fa_spi_entry *fa_spi;
	int used = 0;

	if (config.priv_ha && ext && ext->nonce) {
		if (ext->nonce->length != sizeof(struct nonce_ext) - 2) {
			DEBUG(DEBUG_FLAG, " - invalid nonce ext len\n");
		} else {
			DEBUG(DEBUG_FLAG, " * nonce\n");
			if (left < GET_NONCE_EXT_LEN(ext->nonce))
				return -1;
			memcpy(pos + used, ext->nonce,
			       GET_NONCE_EXT_LEN(ext->nonce));
			used += GET_NONCE_EXT_LEN(ext->nonce);
			left -= GET_NONCE_EXT_LEN(ext->nonce);
		}
	}

	fa_spi = get_fa_spi(0, addr);
	if (fa_spi != NULL && config.priv_ha > 0) {
		DEBUG(DEBUG_FLAG, " * sha_ha_auth\n");
		if (left < sizeof(struct vendor_msg_auth) + MAX_SK_LEN)
			return -1;
		used += auth_add_vendor(
			AUTH_ALG_MD5, fa_spi->shared_secret,
			fa_spi->shared_secret_len, start,
			(struct vendor_msg_auth *) (pos + used),
			VENDOR_EXT_DYNAMICS_SHA_HA_AUTH, htonl(fa_spi->spi));
	} else if (fa_spi != NULL) {
		DEBUG(DEBUG_FLAG, " * fh_auth\n");
		if (left < sizeof(struct msg_auth) + MAX_SK_LEN)
			return -1;
		used += auth_add(fa_spi->alg, fa_spi->shared_secret,
				 fa_spi->shared_secret_len, start,
				 (struct msg_auth *) (pos + used), FH_AUTH,
				 htonl(fa_spi->spi));
	}

	return used;
}


int fill_nemo_ack(struct mobile_nw_ack_ext *nemo_ack, unsigned char code, 
	struct in_addr *prefix,	unsigned char prefix_length)
{	
	if (nemo_ack == NULL) {
		DEBUG(DEBUG_FLAG, "[ha.c] fill_nemo_ack() - nemo_ack == NULL !!!");
		return -1;	
	}
	nemo_ack->type			= 148;
	nemo_ack->length		= 6;
	nemo_ack->subtype		= 1;				// 1 == explicit, 2 == implicit mode ack
	nemo_ack->code			= code;
	nemo_ack->prefix_length	= prefix_length;
	
	nemo_ack->reserved		= 0;				// sent as 0, ignored on reception
	nemo_ack->prefix		= prefix->s_addr;
	
	return 0;
}



/**
 * ro_prefix_adv_into_buf: Add Route Optimization Prefix Advertisement into the 
 * buffer.
 * 
 * @mr_hoa: HoA of the MR the extension is sent to (this will not be added into the buf) 
 * @buf: buffer to fill
 * @opt_solicit_addr: possible address that was solicited by the MR
 * 
 * The buffer has to have memory alloced when calling the function. It will be filled 
 * with the information from the binding table. The format is:
 *   [ MR1 HoA | MR1 NEMO nw1 | MR1 NEMO nw1 NAI | MR1 NEMO nw2 | MR1 NEMO nw1 NAI | 
 *     MR2 HoA | MR2 NEMO nw1 | MR2 NEMO nw1 NAI | ... ]
 * 
 * If opt_solicit_addr is bigger than 0, only this MR's networks are placed in the 
 * datagram (related to the S bit in the RO Capability Extension).
 * 
 * Returns: The RO Adv header length starting from the length field.
 */
int ro_prefix_adv_into_buf(struct in_addr mr_hoa, unsigned char *buf, 
				struct in_addr opt_solicit_addr) 
{
	if (bindingcount < 1) return -1;
	
	struct ro_prefix_adv_head *ro_adv_head;
	struct ro_prefix_adv_data *ro_adv_data;
	struct bindingentry *binding;
	int offset; // in bytes
	int b, i;
	__u16 ext_length; // RO Adv header length starting from the length field
	int octets_in_prefix;
	char str_end = '\0';
	
	offset 			= sizeof(struct ro_prefix_adv_head); // in bytes
	ro_adv_head 	= (struct ro_prefix_adv_head *) buf;
	ro_adv_head->type			= TBA_T3;
	ro_adv_head->subtype 		= TBA_ST_2_3;
	ext_length					= 0;
	
	DEBUG(DEBUG_FLAG, "ro_prefix_adv_into_buf: called by %s", inet_ntoa(mr_hoa));
	if (opt_solicit_addr.s_addr > 0) 
	{
		DEBUG(DEBUG_FLAG, ", solicited only the networks of address %s", 
			inet_ntoa(opt_solicit_addr));
	}
	DEBUG(DEBUG_FLAG, "\n");
	
	for (b = 0; b < bindingcount; b++) 
	{ 
		DEBUG(DEBUG_FLAG, "\tbinded_hoas[%d]: %s\n", b, inet_ntoa(binded_hoas[b]));	
		if (binded_hoas[b].s_addr == mr_hoa.s_addr) 
		{
			DEBUG(DEBUG_FLAG, "\t\tThe caller's HoA --> next.\n");
			continue;
		}
		
		/* Check the optional solicitation address */
		if (opt_solicit_addr.s_addr > 0 && 
			binded_hoas[b].s_addr != opt_solicit_addr.s_addr) 
		{
			DEBUG(DEBUG_FLAG, "\t\tThis CR HoA is not the solicited HoA --> next.\n");
			continue;			
		} else {
			DEBUG(DEBUG_FLAG, "\t\tThis CR HoA is the solicited HoA --> load into buffer.\n");
		}
		binding = binding_fetch2(bindings, &binded_hoas[b]);
		if (binding == NULL) 
		{
			DEBUG(DEBUG_FLAG, "\t\tbinding_fetch2(bindings, &binded_hoas[%d]) "
				"== NULL!!\n", b);
			return -1;
		}
		if (!binding->adv_my_networks) {
			DEBUG(DEBUG_FLAG, "This CR doesn't want it's networks to be advertised --> next.\n");
			continue;
		}
		
		/*------------------------------------------------------------------*
	 	 | Flags + Info (1 B) | HoA (4 B) 									|
		 *------------------------------------------------------------------*/
		ro_adv_data = (struct ro_prefix_adv_data *) (buf + offset);
		ro_adv_data->flags_plen_info	= 0 | RO_ADV_M;	// |D|M| Plen/Info |
		ro_adv_data->hoa_or_prefix		= binding->mn_addr.s_addr;
		
		/* If this CR is unreachable (behind a NAT or a firewall) -> info = 1 */
		if (binding->not_reachable) {
			ro_adv_data->flags_plen_info |= 0x01;
		}
		/* Jump over the Info + HoA */
		offset		+= 5;
		ext_length 	+= 5;
		
		/* MR NEMO nws into the buffer. */
		for (i = 0; i < binding->nemo_nw_count; i++) 
		{
			DEBUG(DEBUG_FLAG, "\t\tNEMO nw %d: %s/%d", i, 
				inet_ntoa(binding->nemo_nws[i].prefix), 
				binding->nemo_nws[i].prefix_length);
			
			if (binding->nemo_nws[i].prefix_length > 32) {
				DEBUG(DEBUG_FLAG, "\t\tInvalid NEMO plen %d\n",
					binding->nemo_nws[i].prefix_length);
				return -1;	
			}
						
			/*----------------------------------------------------------*
 			 | Flags + Plen (1 B) | Prefix (1-4 B) | Realm (x B)		|
			 *----------------------------------------------------------*/
			/* Delta mode not used here so the D bit is always 0. 
			 * Data item is Prefix + Realm -> Plen, M=0. */
			ro_adv_data = (struct ro_prefix_adv_data *) (buf + offset);
			ro_adv_data->flags_plen_info = 0;
			ro_adv_data->flags_plen_info	
				= (unsigned char) binding->nemo_nws[i].prefix_length;
			
			/* Jump over flags field */
			offset++;
			ext_length++;
			
			/* Copy the needed amount of bytes from the prefix into the buffer */
			octets_in_prefix = divide_and_round_up(ro_adv_data->flags_plen_info, 8);
			DEBUG(DEBUG_FLAG, " -> send %d octets\n", octets_in_prefix);
			memcpy((buf + offset), &binding->nemo_nws[i].prefix, octets_in_prefix);
			
			/* Jump over the prefix (1-4 B) */
			ext_length	+= octets_in_prefix;
			offset 		+= octets_in_prefix;
			
			/*--------------------------------------------------------------*
		 	 | NEMO NAI (realm) (x B)										|
			 *--------------------------------------------------------------*/
			/* If NAI length is not defined, make it to be 1 */
			if (binding->nemo_nws[i].nai_length < 1) 
			{
				binding->nemo_nws[i].nai[0] = '\0';
				binding->nemo_nws[i].nai_length = 1;
			}
			/* If NAI doesn't end with \0, insert it there */
			if (binding->nemo_nws[i].nai[binding->nemo_nws[i].nai_length] != '\0') 
			{
				DEBUG(DEBUG_FLAG, "\t\tWarning: nw NAI doesn't end with a '\0', add\n");
				binding->nemo_nws[i].nai[binding->nemo_nws[i].nai_length] = '\0';
			}
			/* This version does not support NEMO realms so we put only '\0' in the 
			 * realm field */
			memcpy((unsigned char *) (buf + offset), &str_end, 1);
			ext_length	+= 1;
			offset 		+= 1;
		}
	}
	ro_adv_head->length = htons(ext_length);
	return ext_length;
}
/**
 * Returns 
 *  1 if the R bit was set in the MR RO Capability Extension
 *  0 if it was not set or the entrys are NULL
 */
int ro_capab_r_bit_set(struct msg_extensions *ext) {

	if (ext->ro_capab.s_on != NULL) {
		if (ext->ro_capab.s_on->flags & RO_CAPAB_R) {
			return 1;
		}
	} else if (ext->ro_capab.s_off != NULL) {
		if (ext->ro_capab.s_off->flags & RO_CAPAB_R) {
			return 1;
		}
	}
	return 0;
}
/**
 * send_reg_repl:
 * @s: socket to be used in sending the registration reply
 * @binding: bindingentry for the MN
 * @mn_spi: SPI entry for the MN
 * @ext: parsed registration request or %NULL for teardown reply
 * @forced_iface: forced destination interface or %NULL for no forcing
 * @code: registration reply code
 *
 * Creates and sends a registration reply message to the foreign agent given
 * in the binding. This function can also be used for sending a teardown reply
 * to the old foreign agent by giving %NULL @ext argument.
 *
 * Returns:
 *  0 on success or -1 on failure
 */
static int
send_reg_repl(int s, struct bindingentry *binding, struct spi_entry *mn_spi,
	      struct msg_extensions *ext, struct interface_entry *forced_iface,
	      int code)
{
	unsigned char msg[MAXMSG];
	unsigned char *msgpos = msg;
	struct reg_rep *reply;
	struct msg_key *key;
	
	struct udp_tnl_rep_ext *tnl_rep;
	struct mobile_nw_ack_ext *nemo_ack;
	struct ro_reply_ext *ro_reply;

	struct sockaddr_in dest_addr;
	struct fa_spi_entry *fa_spi;
#ifdef USE_TEARDOWN
	struct registration_ext_dynamics *dyn_ext;
#endif
	int msglen, result, key_to_fa, left = MAXMSG, n;
	struct ha_tunnel_data *t_data;

	ASSERT(s >= 0 && binding != NULL && binding->data != NULL &&
	       mn_spi != NULL);

	t_data = (struct ha_tunnel_data *) binding->data;

	print_trace(traces, "Registration accepted\n");

	/* Create the reply extension */
	DEBUG(DEBUG_FLAG, "Reply\n");
	reply 					= (struct reg_rep *) msg;
	memset(reply, 0, sizeof(struct reg_rep));
	reply->lifetime 		= htons(binding->timeout);
	reply->type 			= REG_REP;
	reply->code 			= code;
	reply->home_addr.s_addr = binding->mn_addr.s_addr;
	reply->ha_addr.s_addr 	= config.sha_addr.s_addr != 0 ?
		config.sha_addr.s_addr : binding->ha_addr.s_addr;
	memcpy(reply->id, binding->id, REG_REQ_ID_LEN);
	if (mn_spi->replay_method == REPLAY_PROTECTION_NONCE)
		t_data->nonce = reply->id[0] = get_rand32();
	
	msgpos 	+= sizeof(struct reg_rep);
	left 	-= sizeof(struct reg_rep);

	n = ha_add_ext_start(msg, msgpos, left, ext, binding->timeout,
			     mn_spi, binding, t_data->auth_type);
	if (n < 0) {
		return -1;	
	}
	msgpos += n;
	left -= n;

	/* Encrypt the session key for the FA if it requested the key with
	 * fa_keyreq (use shared secret and MD5) or fa_pubkey (use RSA) */
	key_to_fa = 0;
	if (binding->fa_spi != 0 && ntohs(reply->lifetime) != 0) {
		fa_spi = get_fa_spi(binding->fa_spi, t_data->lower_saddr);
		if (fa_spi == NULL) {
			LOG2(LOG_ERR, "unknown FA SPI %i\n", binding->fa_spi);
			return -1;
		}
		DEBUG(DEBUG_FLAG, " * fa_keyrep\n");
		if (left < sizeof(struct msg_key) + MAX_SK_LEN){
			return -1;
		}
		key = (struct msg_key *) msgpos;
		n = auth_encrypt(fa_spi->alg, fa_spi->shared_secret,
				 fa_spi->shared_secret_len, binding->key,
				 key, reply, VENDOR_EXT_DYNAMICS_FA_KEYREP,
				 htonl(fa_spi->spi));
		msgpos += n;
		left -= n;
		key_to_fa = 1;
	} else if (binding->fa_pubkey && ntohs(reply->lifetime) != 0 &&
		   binding->last_sent_fa_pubkeyrep) {
		DEBUG(DEBUG_FLAG, " * fa_pubkeyrep\n");
		n = GET_KEY_EXT_LEN(binding->last_sent_fa_pubkeyrep);
		if (left < n)
			return -1;
		/* If the foreign agent included a public key, add an RSA
		 * encrypted session key to the FA */
		key = (struct msg_key *) msgpos;
		memcpy(key, binding->last_sent_fa_pubkeyrep, n);
		msgpos += n;
		left -= n;
		key_to_fa = 1;
	}
	
	/*----------------------------------------------*
 	 | UDP Tunnel Reply								|
	 *----------------------------------------------*/
	if(ext->tnl_req != NULL) {
		DEBUG(DEBUG_FLAG, " * UDP Tunnel Reply Extension\n");
		if (left < sizeof(struct udp_tnl_rep_ext)) {
			DEBUG(DEBUG_FLAG, "\tno space left to add the extension!\n");
			return -1;
		}
		tnl_rep 					= (struct udp_tnl_rep_ext *) msgpos;
		tnl_rep->type 				= 44;
		tnl_rep->length 			= 6;
		tnl_rep->subtype			= 0;
		tnl_rep->reply_code			= 0;			// 0: tunnel / 64: no tunnel
		
		if (ext->tnl_req->reserved2 == 0x80)
			tnl_rep->reserved = htons(0x8000);		// Set forced-flag 
		else 
			tnl_rep->reserved = htons(0x0000);
		
		tnl_rep->keepalive_interval = htons(0);		// 0: MN MUST use its default 
		msgpos 	+= sizeof(struct udp_tnl_rep_ext);
		left 	-= sizeof(struct udp_tnl_rep_ext);		
	}

	/*--------------------------------------------------------------------------*
 	 | Add NEMO Ack if the RegReq had the NEMO Request Extension.				|
	 *--------------------------------------------------------------------------*/
	/* TODO: MOBNET_INVALID_PREFIX_LEN, MOBNET_FWDING_SETUP_FAILED */
	if (ext->nemo_req != NULL && ext->nemo_req != NULL && ext->nemo_nw_amount > 0) {
		int i, k;
		for (i = 0; i < binding->nemo_nw_count; i++) 
		{
			DEBUG(DEBUG_FLAG, " * NEMO SUCCESS Ack Extension number %d\n", i);
			if (left < sizeof(struct mobile_nw_ack_ext)) 
			{
				DEBUG(DEBUG_FLAG, "\tno space left to add the extension!\n", i);
				return -1;
			}
			nemo_ack = (struct mobile_nw_ack_ext *) msgpos;
			
			fill_nemo_ack(nemo_ack, MOBNET_SUCCESS, &binding->nemo_nws[i].prefix,
				binding->nemo_nws[i].prefix_length);
			print_trace(traces, " - NEMO Network %s/%d accepted\n",
				inet_ntoa(binding->nemo_nws[i].prefix), binding->nemo_nws[i].prefix_length);
			
			msgpos 	+= sizeof(struct mobile_nw_ack_ext);
			left 	-= sizeof(struct mobile_nw_ack_ext);		
		}
		for (k = 0; k < binding->rejected_nemo_nw_count; k++) 
		{
			DEBUG(DEBUG_FLAG, " * NEMO UNAUTHORIZED Ack Extension number %d\n", k);
			if (left < sizeof(struct mobile_nw_ack_ext)) 
			{
				DEBUG(DEBUG_FLAG, "\tno space left to add the extension!\n");
				return -1;
			}
			nemo_ack = (struct mobile_nw_ack_ext *) msgpos;
			
			fill_nemo_ack(nemo_ack, MOBNET_UNAUTHORIZED, &binding->rejected_nemo_nws[k].prefix,
				binding->rejected_nemo_nws[k].prefix_length);
			print_trace(traces, " - NEMO Network %s/%d rejected\n",
				inet_ntoa(binding->rejected_nemo_nws[k].prefix), binding->rejected_nemo_nws[k].prefix_length);
			
			msgpos 	+= sizeof(struct mobile_nw_ack_ext);
			left 	-= sizeof(struct mobile_nw_ack_ext);		
		}
	}

	/*--------------------------------------------------------------------------*
 	 | Add RO Reply if the Reg Req had the RO Capability (HAaRO) Ext.			|
	 *--------------------------------------------------------------------------*/
	/* Check if one of the pointers in ro_capab is not NULL */
	if (ext->ro_capab.s_on != NULL || ext->ro_capab.s_off != NULL) {
		DEBUG(DEBUG_FLAG, " * Route Optimization (HAaRO) Reply\n");
		
		if (left < sizeof(struct ro_reply_ext)) {
				DEBUG(DEBUG_FLAG, "\tno space left to add the extension!\n");
				return -1;
		}
		ro_reply 			= (struct ro_reply_ext *) msgpos;
		ro_reply->type 		= TBA_T2;
		ro_reply->length 	= 2;
		ro_reply->subtype 	= TBA_ST2_1;
		ro_reply->code 		= RO_OK; 

		/* If S was set in the RO Capability, set it also in the RO Reply */
		if (ext->ro_capab.s_on != NULL && ext->ro_capab.s_off == NULL) {
			ro_reply->code |= RO_CAPAB_S;
		} else {
			DEBUG(DEBUG_FLAG, "Warning: RO Capability Extension not parsed correctly!\n");		
		}
		msgpos 	+= sizeof(struct ro_reply_ext);
		left 	-= sizeof(struct ro_reply_ext);
		
		DEBUG(DEBUG_FLAG, " * Route Optimization (HAaRO) Prefix Advertisement\n");
			
		/* Check if MR set the R bit and there are more bindings than the current one */
		if (!ro_capab_r_bit_set(ext)) {
			DEBUG(DEBUG_FLAG, "\tMR didn't set R flag -> don't add HAaRO Prefix Adv.\n");
		} else if (bindingcount <= 1) {
			DEBUG(DEBUG_FLAG, "\tNo other bindings -> don't add HAaRO Prefix Adv.\n");
		} else {
			/*----------------------------------------------------------------------*
	 		 | Add the Route Optimization Prefix Advertisement. 					|
			 *----------------------------------------------------------------------*/		
			/* Inform the MR about known MRs that are RO capable and have NEMO networks. */
			/* Put RO Adv data in temp buffer so that it can easily be discarded if there is
			 * too little space to add the extension. */
			unsigned char temp_buf[2000];
			int total_length;
			int ro_adv_length;
			struct in_addr sol_addr;
			/* Load RO Prefix Advertisement into the buffer */
			/*----------------------------------------------*
	 		 | Check the Solicitation bit					|
			 *----------------------------------------------*/		
			if (ext->ro_capab.s_on != NULL && 
				ext->ro_capab.s_off == NULL) 
			{
				/* S = 1 */
				sol_addr.s_addr = ext->ro_capab.s_on->mr_hoa;
				ro_adv_length = ro_prefix_adv_into_buf(binding->mn_addr, 
									temp_buf, sol_addr);
			} else if (ext->ro_capab.s_on == NULL && 
					ext->ro_capab.s_off != NULL) 
			{
				/* S = 0 */
				sol_addr.s_addr = 0;
				ro_adv_length = ro_prefix_adv_into_buf(binding->mn_addr, 
									temp_buf, sol_addr);
			} else
			{
				DEBUG(DEBUG_FLAG, "Could not create RO Prefix Advertisement because "
					"RO Capability Extension not parsed correctly!\n");
			}
			total_length 	= ro_adv_length + 4;
			
			if (total_length >= sizeof(struct ro_prefix_adv_head)) {
				if (total_length <= left) {
					DEBUG(DEBUG_FLAG, "\tAdd extension (adv.length = %d B --> total length %d B)\n", 
						ro_adv_length, total_length);
					/* Extension data from temp_buf into the msgpos */
					memcpy(msgpos, temp_buf, total_length);
					msgpos 	+= total_length;
					left 	-= total_length;
				} else {
					DEBUG(DEBUG_FLAG, "\tno space left to add the extension!\n");
					return -1;
				}
			} else {
				DEBUG(DEBUG_FLAG, "\tError in ro_prefix_adv_into_buf\n");
			}
		}
	}
#ifdef USE_TEARDOWN
	if (teardown) {
		DEBUG(DEBUG_FLAG, " * ext_dynamics (teardown)\n");
		if (left < sizeof(struct registration_ext_dynamics))
			return -1;
		dyn_ext = (struct registration_ext_dynamics *) msgpos;
		dyn_ext->type = VENDOR_EXT_TYPE2;
		dyn_ext->reserved = 0;
		dyn_ext->length = sizeof(struct registration_ext_dynamics) - 2;
		dyn_ext->vendor_id = htonl(VENDOR_ID_DYNAMICS);
		dyn_ext->sub_type = htons(VENDOR_EXT_DYNAMICS_OPTIONS);
		dyn_ext->version = VENDOR_EXT_VERSION;
		dyn_ext->opts = REG_EXT_OWN_TEAR_DOWN;
		dyn_ext->seq = 0;
		msgpos += sizeof(struct registration_ext_dynamics);
		left -= sizeof(struct registration_ext_dynamics);
	}
#endif

	if (key_to_fa) {
		DEBUG(DEBUG_FLAG, " * sk_auth\n");
		if (left < sizeof(struct vendor_msg_auth) + MAX_SK_LEN)
			return -1;
		/* Add session key-based message authentication for FA */
		n = auth_add_vendor(AUTH_ALG_MD5, binding->key,
				    binding->keylen, msg,
				    (struct vendor_msg_auth *) msgpos,
				    VENDOR_EXT_DYNAMICS_SK_AUTH,
				    htonl(mn_spi->spi));
		msgpos += n;
		left -= n;
	}

	n = ha_add_ext_end(msg, msgpos, t_data->lower_saddr, left, ext);
	if (n < 0)
		return -1;
	msgpos += n;
	left -= n;

	msglen = msgpos - msg;

	dest_addr.sin_family = AF_INET;
	dest_addr.sin_addr.s_addr = t_data->lower_saddr.s_addr;
	dest_addr.sin_port = binding->lower_port;

	if (forced_iface != NULL &&
	    setsockopt(s, SOL_SOCKET, SO_BINDTODEVICE, forced_iface->dev,
		       IFNAMSIZ)) {
		DEBUG(DEBUG_FLAG, "setsockopt(SOL_SOCKET, SO_BINDTODEVICE): "
		      "%s\n", strerror(errno));
	}

	DEBUG(DEBUG_FLAG, " * total %i bytes to %s:%i\n", msglen,
	      inet_ntoa(dest_addr.sin_addr), ntohs(dest_addr.sin_port));
	result = sendto(s, msg, msglen, 0,
			(struct sockaddr *) &dest_addr,
			sizeof(struct sockaddr_in));

	if (forced_iface != NULL) {
		/* kernel seems to require optlen >= sizeof(int) */
		char dummy[sizeof(int)];
		memset(dummy, 0, sizeof(dummy));
		if (setsockopt(s, SOL_SOCKET, SO_BINDTODEVICE, dummy,
			       sizeof(dummy)))
			DEBUG(DEBUG_FLAG, "setsockopt(SOL_SOCKET, "
			      "SO_BINDTODEVICE): %s\n", strerror(errno));
	}

	return dynamics_check_sendto(result, msglen, "send_reg_repl");
}


/**
 * send_reg_failure:
 * @s: socket to be used for sending the registration reply
 * @sock_addr: destination address (IP address and UDP port)
 * @mn_spi: pointer to SPI structure for the MN or %NULL if not known
 * @auth_type: authentication data area (%AUTH_RFC2002 / %AUTH_RFC2002BIS)
 * @code: failure code for the registration reply (3 .. 255)
 * @nonce: nonce value to be updated if the MN uses nonce-based replay
 * protection; @nonce can also be %NULL, when the nonce value is not updated
 * @ext: parsed registration request
 * @forced_iface: forced destination interface or %NULL for no forcing
 * @binding: pointer to MN's binding entry if known, or %NULL if not available
 * @from_iface: interface from which the registration request was received
 *
 * Sends a registration reply with a failure code @code to @sock_addr.
 *
 * Returns:
 *   0 on success or -1 on failure
 */
static int
send_reg_failure(int s, struct sockaddr_in *sock_addr,
		 struct spi_entry *mn_spi, int auth_type,
		 int code, __u32 *nonce, struct msg_extensions *ext,
		 struct interface_entry *forced_iface,
		 struct bindingentry *binding,
		 struct interface_entry *from_iface)
{
	unsigned char msg[MAXMSG];
	unsigned char *msgpos = msg;
	struct reg_rep *reply;
	int msglen, result, n, left = MAXMSG;
	struct ha_tunnel_data *t_data = NULL;
	static time_t last_failure_time = 0;
	time_t *timer, now;

	ASSERT(s >= 0 && sock_addr && ext && ext->req && code > 2);

	if (binding) {
		/* known binding - use MN's own timer */
		t_data = (struct ha_tunnel_data *) binding->data;
		timer = &t_data->last_failure_time;
	} else {
		/* no binding - use general timer */
		timer = &last_failure_time;
	}

	/* do not send failure replys too often */
	time(&now);
	if (*timer > now) {
		DEBUG(DEBUG_FLAG, "send_reg_failure: timed failure in the "
		      "future - host time changed?\n");
		*timer = now;
	} else if (now - *timer < config.reg_error_reply_interval) {
		DEBUG(DEBUG_FLAG, "Too frequent error message - skipping\n");
		print_trace(traces, "Not sending Registration Failure because the last one sent so recently\n");		
		return 0;
	}
        *timer = now;

	if (mn_spi == NULL && ext->mh_auth != NULL)
		mn_spi = get_mn_spi(ntohl(ext->mh_auth->spi));
	
	/* Create the reply extension */
	DEBUG(DEBUG_FLAG, "Failure reply\n");
	reply = (struct reg_rep *) msg;
	memset(reply, 0, sizeof(struct reg_rep));
	reply->lifetime = 0;
	reply->type = REG_REP;
	reply->code = code;
	reply->home_addr.s_addr = ext->req->home_addr.s_addr;
	reply->ha_addr.s_addr = ext->req->ha_addr.s_addr;
	if (config.sha_addr.s_addr != 0)
		reply->ha_addr.s_addr = config.sha_addr.s_addr;
	else if (code == REGREP_UNKNOWN_HA_HA) {
		/* use the HA's unicast IP address in case of HA discovery */
		struct node *node;
		int found = 0;
		for (node = list_get_first(&config.interfaces); node != NULL;
		     node = list_get_next(node)) {
			struct interface_entry *iface =
				(struct interface_entry *) node;
			if (iface->ha_disc &&
			    iface->bcaddr.s_addr == ext->req->ha_addr.s_addr) {
				reply->ha_addr.s_addr = iface->addr.s_addr;
				found = 1;
				break;
			}
		}
		if (!found && from_iface != NULL)
			reply->ha_addr.s_addr = from_iface->addr.s_addr;
	}
	memcpy(reply->id, ext->req->id, REG_REQ_ID_LEN);
	if (mn_spi) {
		if (mn_spi->replay_method == REPLAY_PROTECTION_TIMESTAMP &&
		    code == REGREP_ID_MISMATCH_HA) {
			reply->id[0] = htonl(time(NULL) + UNIX_NTP_DIFF);
		} else if (mn_spi->replay_method == REPLAY_PROTECTION_NONCE) {
			reply->id[0] = get_rand32();
			if (nonce != NULL)
				*nonce = reply->id[0];
		}
	}
	msgpos += sizeof(struct reg_rep);
	left -= sizeof(struct reg_rep);

	n = ha_add_ext_start(msg, msgpos, left, ext, 0, mn_spi, NULL,
			     auth_type);
	if (n < 0)
		return -1;
	msgpos += n;
	left -= n;

	n = ha_add_ext_end(msg, msgpos, sock_addr->sin_addr, left, ext);
	if (n < 0)
		return -1;
	msgpos += n;
	left -= n;

	msglen = msgpos - msg;

	if (forced_iface != NULL &&
	    setsockopt(s, SOL_SOCKET, SO_BINDTODEVICE, forced_iface->dev,
		       IFNAMSIZ)) {
		DEBUG(DEBUG_FLAG, "setsockopt(SOL_SOCKET, SO_BINDTODEVICE): "
		      "%s\n", strerror(errno));
	}

	DEBUG(DEBUG_FLAG, "send_reg_failure(code=%i): "
	      "sending %i bytes to %s:%i (forced_iface=%s)\n", code, msglen,
	      inet_ntoa(sock_addr->sin_addr), ntohs(sock_addr->sin_port),
	      forced_iface != NULL ? forced_iface->dev : "N/A");
	result = sendto(s, msg, msglen, 0,
			(struct sockaddr *)sock_addr,
			sizeof(struct sockaddr_in));
	
	print_trace(traces, "Request failed, code %d\n", code);
	
	if (forced_iface != NULL) {
		/* kernel seems to require optlen >= sizeof(int); in addition,
		 * 2.2 kernels seem to report error (Invalid argument) even
		 * when this unbinding success.. */
		char dummy[sizeof(int)];
		memset(dummy, 0, sizeof(dummy));
		if (setsockopt(s, SOL_SOCKET, SO_BINDTODEVICE, dummy,
			       sizeof(dummy)) < 0 &&
		    setsockopt(s, SOL_PACKET, SO_BINDTODEVICE, NULL, 0) < 0)
			DEBUG(DEBUG_FLAG, "setsockopt(SOL_SOCKET, "
			      "SO_BINDTODEVICE): %s (this is probably "
			      "bogus)\n", strerror(errno));
	}

	return dynamics_check_sendto(result, msglen, "send_reg_failure");
}

/* handle_msg_in_tun_fd: Read message from tunnel->tun_fd and send it with MIP 
 * TNL header to the corresponding tunnel->dst_addr:tunnel->dst_port (MR CoA).
 */
int handle_msg_in_tun_fd(TUNNEL *tunnel) {
	
	int fd = tunnel->tun_fd;
	int bufsize = 2000;
	char buffer[bufsize];
	int bytes_read;
	struct udp_tnl_data *tnl_data;
	struct in_addr coa_ip	= tunnel->dst_addr;
	unsigned short coa_port	= tunnel->dst_port;
	
	/*----------------------------------------------------------*
	 | Read the data from the tun_fd into a buffer.				|
	 *----------------------------------------------------------*/
	/* Leave 4 bytes free in the beginning so that the MIP Tunnel 
	 * Data Message header can be fitted there. */
	bytes_read = read(fd, buffer + 4, bufsize - 4);
	DEBUG(DEBUG_FLAG, " ================>>>> \n");
	DEBUG(DEBUG_FLAG, "handle_msg_in_tun_fd: Send %d bytes tun_fd %d to MR CoA %s:%d\n", 
		bytes_read, fd, inet_ntoa(coa_ip), ntohs(coa_port));
  	if (bytes_read < 0) {
    	DEBUG(DEBUG_FLAG, "\thandle_msg_in_tun_fd: Error reading data!\n");
    	return -1;
  	}
  	print_ip_stack(buffer + 4);

  	/*--------------------------------------------------------------*
	 | Add the MIP Tunnel Data Message header.						|
	 *--------------------------------------------------------------*/
	tnl_data 				= (struct udp_tnl_data *) buffer; 	/* point tnl_data to the buffer */
	tnl_data->type 			= 4;								/* fill the buffer */
	tnl_data->next_header 	= 4;
	tnl_data->reserved		= htons(0x00);
  	
  	/*----------------------------------------------------------*
	 | Send as UDP to the end of this tunnel.					|
	 *----------------------------------------------------------*/	
	struct sockaddr_in coa;
 	coa.sin_family	= AF_INET;
  	coa.sin_port	= coa_port; // in nw byte order
  	coa.sin_addr	= coa_ip;
  	bzero(&(coa.sin_zero), 8); 
	
	/* This may block the execution */
	int sendto_ret = sendto(tunnel->tun_socket, buffer, bytes_read + 4, MSG_DONTWAIT, 
		(struct sockaddr *) &coa, sizeof(struct sockaddr_in));
	if (sendto_ret == -1) {
		fprintf(stderr, "sendto - %s", strerror(errno));
	}
	return 0;
}

/**
 * handle_tnl_msg: Handle MIP tnl message which consists of the following headers:
 *  - IP
 *  - MIP (type: TNL Data Msg header)
 *  - UDP (srcPort: MR registration port)
 *  - IP  (srcIP: MR CoA)
 * 
 * Function writes all the data starting from the highest IP packet to the tun_fd 
 * corresponding to the MR whose CoA was the srcIP of the lowest IP packet. 
 */
int handle_tnl_msg(struct sockaddr_in *coa, char *mip_msg_start, int len) {
	struct in_addr *coa_ip 			= &coa->sin_addr;
	short int coa_port				= ntohs(coa->sin_port);
	struct udp_tnl_data *tnl_data 	= (struct udp_tnl_data *) mip_msg_start;
	char *ip_header_start;
	struct iphdr *ip_header;
	struct bindingentry *binding;
	int mn_tun_fd, b;
	TUNNEL *tunnel;
	tunnel = tunnel_fetch2(tunnels, coa_ip);

	binding = binding_fetch_by_coa(bindings, coa_ip);
	if (binding == NULL) {
		DEBUG(DEBUG_FLAG, "MIP TNL Message's CoA %s:%d not associated with any binding!\n",
			inet_ntoa(*coa_ip), ntohs(coa->sin_port));
		return -1;
	}
	DEBUG(DEBUG_FLAG, " <<<<================ \n");
	DEBUG(DEBUG_FLAG, "Received MIP TNL message from CoA %s, ", inet_ntoa(*coa_ip));
	DEBUG(DEBUG_FLAG, "HoA %s\n", inet_ntoa(binding->mn_addr));
	
	mn_tun_fd = tunnel->tun_fd;
	
	/*--------------------------------------------------------------*
	 | MIP Tunnel Data Message header is followed by IP header.		|
	 *--------------------------------------------------------------*/
	/* This IP packet has to be sent to the kernel. */
	ip_header_start = mip_msg_start + sizeof(struct udp_tnl_data);
	ip_header 		= (struct iphdr *) ip_header_start;
	
	print_ip_stack(ip_header_start);
	
	/*------------------------------------------------------------------*
	 | Write the IP packet following the MIP TNL header into the tun_fd.|
	 *------------------------------------------------------------------*/
	DEBUG(DEBUG_FLAG, "\tWrite into tun_fd %d\n", mn_tun_fd);
	
	if(write(mn_tun_fd, ip_header_start, ntohs(ip_header->tot_len)) < 0) {
    	perror("Error writing into TUN fd");
    	return -1;
  	}	
  	return 0;
}

/**
 * handle_reg_msg:
 * @s: registration message socket (UDP)
 *
 * Processes a registration request (or tunnel) message that has arrived on socket @s.
 *
 * Returns:
 *   0 on success or -1 on failure
 */
static int
handle_reg_msg(struct interface_entry *from_iface, int s)
{
	char msg[MAXMSG];				/* MAXMSG == 2048 */
	struct sockaddr_in cli_addr;	/* CoA IP + port */
	struct in_addr dst_addr;
	struct msghdr mh;
	char cdata[256];
	struct iovec iov;
	struct cmsghdr *cmsg;
	struct bindingentry *binding = NULL;
	struct spi_entry *mn_spi;
	int n, res, rsock;
	struct msg_extensions ext;
	int killbinding;
	int code, auth_type;
	char mn_addrstr[17];
	char fa_addrstr[17];
	struct ha_tunnel_data *t_data = NULL;
	struct bindingkey bkey;
	struct node *node;
	struct interface_entry *force_iface = NULL;

	char hoa_str[INET6_ADDRSTRLEN];
	char coa_str[INET6_ADDRSTRLEN];
	char nai_str[MAX_NAI_LEN + 1];
	
	int hoa_was_zero = 0;		/* set if HoA was 0.0.0.0 */
	int nai_was_known = 0;		/* set if HoA was 0.0.0.0 but the NAI was known */
	int tnl_data = 0;			/* set to 1 if message is UDP Tunnel Data Message */
	struct in_addr temp_addr;	/* temp variables */
	struct in_addr zero_addr;
	ext.nemo_nw_amount = 0;
	inet_pton(AF_INET, "0.0.0.0", &zero_addr);
	
	memset(&cli_addr, 0, sizeof(cli_addr));
	dst_addr.s_addr = 0;
	memset(&mh, 0, sizeof(mh));
	iov.iov_base = msg;
	iov.iov_len = MAXMSG;
	mh.msg_name = &cli_addr;			/* Address to send to/receive from. */
	mh.msg_namelen = sizeof(cli_addr);	/* Length of address data. */
	mh.msg_iov = &iov;					/* Vector of data to send/receive into. */
	mh.msg_iovlen = 1;
	mh.msg_control = &cdata;			/* Ancillary data (eg BSD filedesc passing). */
	mh.msg_controllen = sizeof(cdata);

	/*--------------------------------------------------------------*
	 | Get the message into mh.										|
	 *--------------------------------------------------------------*/
	n = recvmsg(s, &mh, 0);

	DEBUG(DEBUG_FLAG, "Received %d bytes from %s:%d, fd %d\n", n,
	      inet_ntoa(cli_addr.sin_addr), ntohs(cli_addr.sin_port), s);

	if (n < 0) {
		LOG2(LOG_ERR, "handle_reg_msg - recvfrom failed - %s\n",
		     strerror(errno));
		return -1;
	}

	for (cmsg = CMSG_FIRSTHDR(&mh); cmsg != NULL;
	     cmsg = DYNAMICS_CMSG_NXTHDR(&mh, cmsg)) {
		if (cmsg->cmsg_level == SOL_IP &&
		    cmsg->cmsg_type == IP_PKTINFO) {
			struct _in_pktinfo *pkt;
			pkt = (struct _in_pktinfo *) CMSG_DATA(cmsg);
			dst_addr = pkt->ipi_addr;
			DEBUG(DEBUG_FLAG, "\tIP_PKTINFO: ipi_ifindex=%i "
			      "ipi_spec_dst=%s ", pkt->ipi_ifindex,
			      inet_ntoa(pkt->ipi_spec_dst));
			DEBUG(DEBUG_FLAG, "ipi_addr=%s\n",
			      inet_ntoa(pkt->ipi_addr));
		}
	}
	
	/* use the matching UDP socket for the registration reply */
	rsock = -1;
	for (node = list_get_first(&config.interfaces); node != NULL;
	     node = list_get_next(node)) {
		struct interface_entry *iface =	(struct interface_entry *) node;
		if (iface->addr.s_addr == dst_addr.s_addr || (iface->ha_disc &&
			 iface->bcaddr.s_addr == dst_addr.s_addr)) {
			rsock = iface->udp_sock;
		}
	}
	if (rsock < 0 && from_iface != NULL &&
	    dst_addr.s_addr == (unsigned int) -1) {
		DEBUG(DEBUG_FLAG, "Broadcast 255.255.255.255 destination - "
		      "using interface[%s] UDP socket[%i]\n", from_iface->dev,
		      from_iface->udp_sock);
		rsock = from_iface->udp_sock;
	}
	if (rsock < 0) {
		LOG2(LOG_WARNING, "Could not find UDP socket for "
		     "registration reply - trying recv. socket\n");
		rsock = s;
	}
	
	/*--------------------------------------------------------------*
	 | Check if it's TNL Data Message								|
	 *--------------------------------------------------------------*/
	if ((unsigned char) msg[0] == TNL) {
		handle_tnl_msg(&cli_addr, msg, n);
		return 0;
	}
	
	/*--------------------------------------------------------------*
	 | Parse the message into ext									|
	 *--------------------------------------------------------------*/
	res = parse_msg(msg, n, &ext);
	
	ext.mr_port 			= cli_addr.sin_port;
	ext.registration_socket	= s;
	
	/* Create address and NAI strings */
	if (ext.req != NULL) {
		inet_ntop(AF_INET, &ext.req->home_addr, hoa_str, INET6_ADDRSTRLEN);	
		inet_ntop(AF_INET, &ext.req->co_addr, coa_str, INET6_ADDRSTRLEN);
	}
	if (ext.mn_nai != NULL) {
		get_nai_str(ext.mn_nai, nai_str);
	}
	
	/*------------------------------------------------------------------*
	 | Check if it's a dereg from home nw (lifetime == 0 ja CoA == HoA)	|
	 *------------------------------------------------------------------*/
	if (ext.req != NULL && ntohs(ext.req->lifetime) == 0 &&
	    ext.req->home_addr.s_addr == cli_addr.sin_addr.s_addr &&
	    ext.req->home_addr.s_addr == ext.req->co_addr.s_addr) 
	{
		/* RFC 2002, Ch. 3.8.3.1:
		 * When HA is replying to deregistration (lifetime == 0 &&
		 * COA == home addr) whose srcIP is MN's home address, the
		 * reply must be sent directly to home network bypassing any
		 * mobility bindings */
		DEBUG(DEBUG_FLAG, "Dereg. from home => bypassing mobility "
		      "bindings in routing\n");
		force_iface = from_iface;
		
		/* Print trace */		
		if (ext.mn_nai != NULL) {
			print_trace(traces, "MR %s deregistering from home network.\n",
				nai_str);
		} else {
			print_trace(traces, "MR %s deregistering from home network %s\n", 
				hoa_str);
		}
	}
	
 	/* check the parse_msg result */	
	if (res == -3 || res == -4) {
		LOG2(LOG_WARNING, "Unknown critical vendor extension in a "
		     "request from %s:%i\n", inet_ntoa(cli_addr.sin_addr),
		     ntohs(cli_addr.sin_port));
		send_reg_failure(rsock, &cli_addr, NULL, 0,
				 res == -3 ? REGREP_UNSUPP_VENDOR_ID_MN_HA :
				 REGREP_UNSUPP_VENDOR_ID_FA_HA, NULL, &ext,
				 force_iface, NULL, from_iface);
		stats.discarded_vendor_ext++;
		report_discarded_msg(msg, n, &cli_addr,
				     "unknown vendor extension");
		return -1;
	} else if ((res != 0 || ext.req == NULL)) {
		char *reason = "N/A";
		if (res == -1) {
			reason = "unknown extension";
			stats.discarded_unknown_ext++;
		} else if (res == -2) {
			reason = "malformed message";
			stats.discarded_malformed_msg++;
		} else if (ext.req == NULL) {
			reason = "not a request";
			stats.discarded_not_request++;
		}
		LOG2(LOG_WARNING, "Message parser failed or not a request "
		     "(from %s:%i): %s\n", inet_ntoa(cli_addr.sin_addr),
		     ntohs(cli_addr.sin_port), reason);
		
		print_trace(traces, "Registration message parsing failed: %s\n", reason);
		
		report_discarded_msg(msg, n, &cli_addr, reason);
		return -1;
	}
	
	if (ext.req != NULL) {
		if (ext.mn_nai != NULL) {
			print_trace(traces, "MR %s requesting registration with HoA %s, CoA %s, id %08x %08x\n", nai_str,
				hoa_str, coa_str, (u32) ntohl(ext.req->id[0]), (u32) ntohl(ext.req->id[1]));
		} else {
			print_trace(traces, "MR requesting registration with HoA %s, CoA %s, id %08x %08x\n", hoa_str,
				coa_str, (u32) ntohl(ext.req->id[0]), (u32) ntohl(ext.req->id[1]));
		}
	}
	
	/*--------------------------------------------------------------*
	 | Check if the HoA is 0.0.0.0		 							|
	 *--------------------------------------------------------------*/
	/* If Request's NAI is found from the binding table 
	 *   - give MR the HoA corresponding to the NAI 
	 * Else 
	 *   - allocate a new HoA from the pool
	 */
	if (ext.req != NULL && ext.req->home_addr.s_addr == zero_addr.s_addr) 
	{
		DEBUG(DEBUG_FLAG, "\tRequest's HoA is 0.0.0.0, NAI %s - check if the NAI is known\n", nai_str);
		hoa_was_zero = 1;

		binding = binding_fetch_by_nai(bindings, nai_str);
		if (binding != NULL) {
			DEBUG(DEBUG_FLAG, "\tRequest's NAI %s found from the binding table\n", nai_str);
			nai_was_known = 1;
			ext.req->home_addr.s_addr = binding->mn_addr.s_addr;
			inet_ntop(AF_INET, &ext.req->home_addr, hoa_str, INET6_ADDRSTRLEN);
			
			print_trace(traces, "Request's NAI %s found from the binding table\n", nai_str);
			print_trace(traces, " -> Reply with binding's HoA: %s\n", inet_ntoa(binding->mn_addr));
		} else {	
			DEBUG(DEBUG_FLAG, "\tRequest's NAI %s not known -> Allocate HoA from address pool\n", nai_str);
			
			/* Can't handle more bindings */
			if (bindingcount > config.hoa_range) {
				DEBUG(DEBUG_FLAG, "\nToo many bindings (%d) for the configured HoA range (%d)!!\n",
					bindingcount, config.hoa_range);
				
				print_trace(traces, "Too many bindings (%d) for the HoA range (%d)!\n",
					bindingcount, config.hoa_range);
				
				send_reg_failure(rsock, &cli_addr, NULL, 0,
					REGREP_TOO_MANY_SIMULTANEOUS, NULL, &ext,
					force_iface, NULL, from_iface);
				
				stats.req_rejected++;
				report_discarded_msg(msg, n, &cli_addr, "Too many bindings for the configured HoA range.");			
				return -1;	
			}
			get_hoa_from_pool(&ext.req->home_addr);
			
			inet_ntop(AF_INET, &ext.req->home_addr, hoa_str, INET6_ADDRSTRLEN);
			
			DEBUG(DEBUG_FLAG, "\tAllocated HoA for this MN: %s\n", hoa_str);
			print_trace(traces, "HoA %s allocated\n", hoa_str);
		}
	}
	
	/*------------------------------------------------------------------------------*
	 | Check if there's already a binding with this bindingkey (HoA + HA addr).		|
	 *------------------------------------------------------------------------------*/
	memcpy(&bkey.mn_addr, &ext.req->home_addr, sizeof(bkey.mn_addr));
	bkey.ha_addr = config.sha_addr.s_addr != 0 ? config.sha_addr :
		ext.req->ha_addr;
	bkey.priv_ha = config.priv_ha;
	if (binding == NULL) {
		binding = binding_fetch(bindings, &bkey);
	}
	if (binding != NULL)
		t_data = (struct ha_tunnel_data *) binding->data; 
	
	/* HoA back to 0.0.0.0 so that the authentication won't fail. */
	if (hoa_was_zero) {
		temp_addr.s_addr			= ext.req->home_addr.s_addr;
		ext.req->home_addr.s_addr 	= zero_addr.s_addr;
	}
	
	/*------------------------------------------------------------------*
	 | Check the authentication.										|
	 *------------------------------------------------------------------*/	
	mn_spi = validate_request(binding, cli_addr.sin_addr, msg, n, &ext,
				  &code, &auth_type);
	
	if (hoa_was_zero) {
		/* Put the new hoa back to ext->req */
		ext.req->home_addr.s_addr = temp_addr.s_addr;
	}
	
	if (code > 2) {
		if (hoa_was_zero) free_hoa(&ext.req->home_addr);
		print_trace(traces, "Invalid message\n");
		send_reg_failure(rsock, &cli_addr, mn_spi, 0, code,
				 (t_data != NULL ? &t_data->nonce : NULL),
				 &ext, force_iface, binding, from_iface);
		stats.req_rejected++;
		report_discarded_msg(msg, n, &cli_addr, "invalid message");
		return -1;
	}
	if (mn_spi == NULL) {
		print_trace(traces, "MN SPI not found\n");
		stats.req_rejected++;
		report_discarded_msg(msg, n, &cli_addr, "MN SPI not found");
		if (hoa_was_zero) free_hoa(&ext.req->home_addr);
		return -1;
	}

	dynamics_strlcpy(mn_addrstr, inet_ntoa(ext.req->home_addr),
			 sizeof(mn_addrstr));
	dynamics_strlcpy(fa_addrstr, inet_ntoa(ext.req->co_addr),
			 sizeof(fa_addrstr));

	if (binding != NULL) 
	{
		t_data->auth_type = (auth_type == AUTH_MAC_RFC2002 ?
				     AUTH_RFC2002BIS : AUTH_RFC2002);
				     				     
		/* If the id field is identical the message was a duplicate.
		 * In order to avoid breaking the tunnel drop this message as
		 * a special case without sending a failure reply if the replay
		 * protection is used
		 */
		if (mn_spi->replay_method != REPLAY_PROTECTION_NONE &&
		    binding->id[0] == ext.req->id[0] &&
		    binding->id[1] == ext.req->id[1]) 
		{
			print_trace(traces, "Received duplicate request - dropping it\n");
			DEBUG(DEBUG_FLAG,
			      "Received duplicate request - dropping it\n");
			stats.req_rejected++;
			if (hoa_was_zero) free_hoa(&ext.req->home_addr);
			return -1;
		}

		/* Make sure that the id field is incremented if the timestamp
		 * protection is used */
		if (mn_spi->replay_method == REPLAY_PROTECTION_TIMESTAMP &&
		    (ntohl(binding->id[0]) > ntohl(ext.req->id[0]) ||
		     (ntohl(binding->id[0]) == ntohl(ext.req->id[0]) &&
		      ntohl(binding->id[1]) >= ntohl(ext.req->id[1])))) {
			print_trace(traces, "Timestamp did not increase (MN %s, CoA %s)\n", 
				mn_addrstr, fa_addrstr);
			LOG2(LOG_NOTICE, "Timestamp did not increase "
			     "(MN=%s, FA=%s)\n", mn_addrstr, fa_addrstr);
			send_reg_failure(rsock, &cli_addr, mn_spi,
					 t_data->auth_type,
					 REGREP_ID_MISMATCH_HA,
					 &t_data->nonce, &ext, force_iface,
					 binding, from_iface);
			stats.req_rejected++;
			if (hoa_was_zero) free_hoa(&ext.req->home_addr);
			return -1;
		}
		
		/* remove binding from list (does not remove tunnels etc.) */
		binding_remove(binding);
		
	} else {
		LOG2(LOG_NOTICE, "MN %s registers, FA is %s\n", 
		     mn_addrstr, fa_addrstr);
	}

	stats.req_accepted++;

	killbinding = 0;
	
	if (binding &&
		(binding->lower_addr.s_addr != (config.sha_addr.s_addr != 0 ?
			config.sha_addr.s_addr :
			ext.req->co_addr.s_addr) || (hoa_was_zero && nai_was_known) || 
		ext.req->lifetime == 0)) 
	{
		DEBUG(DEBUG_FLAG, "Found existing binding for MN %s\n",
		      mn_addrstr);
	
		if (ext.req->co_addr.s_addr == ext.req->home_addr.s_addr &&
		    ext.req->lifetime == 0) {
			DEBUG(DEBUG_FLAG, "MN is deregistering all its "
			      "simultaneous bindings\n");
			/* Note: RFC 2002, 3.6.1.2, special case
			 * This should be taken into account if the
			 * simultaneous bindings are implemented.
			 * At the moment we just deregister the only possible
			 * binding below. */
		}

#ifdef USE_TEARDOWN
		/* if co_addr differs in request, send reply to old
		 * co_addr to purge old tunnel segment */
		if (config.sha_addr.s_addr == 0 &&
		    binding->lower_addr.s_addr != ext.req->co_addr.s_addr) {
			binding->timeout = 0;
			DEBUG(DEBUG_FLAG, "Sending purge msg to %s\n",
			      inet_ntoa(binding->lower_addr));
			if (send_reg_repl(s, binding, mn_spi, NULL, force_iface,
				      REGREP_ACCEPTED) < 0)
			{
				print_trace(traces, "Failed to send Registration Reply!\n");
			}
		}
#endif

		if (ext.req->lifetime > 0) {
			/* it was a location update */
			LOG2(LOG_INFO, "MN %s updating tunnel to FA %s\n", 
			     mn_addrstr, fa_addrstr);
			
			if (ext.mn_nai != NULL) {
				print_trace(traces, "MR %s updating tunnel to CoA %s\n", nai_str, coa_str);
			} else {
				print_trace(traces, "MR %s updating tunnel to CoA %s\n", hoa_str, coa_str);
			}
			if (switch_tunnel(binding, &ext, mn_spi) < 0) {
				print_trace(traces, "Tunnel switching failed!\n");
				if (hoa_was_zero) free_hoa(&ext.req->home_addr);
				return -1;
			}
		} else {
			/* if it is a deregistration request,
			 * destroy the binding after the reply is sent,
			 * since the binding is needed when sending the
			 * reply */
			print_trace(traces, "Deregistration\n");
			
			LOG2(LOG_NOTICE, "MN %s deregisters\n", mn_addrstr);
			remove_tunnel(binding);
			
			if (cli_addr.sin_addr.s_addr == binding->mn_addr.s_addr
			    && ext.req->co_addr.s_addr ==
			    ext.req->home_addr.s_addr) {
				DEBUG(DEBUG_FLAG, "MN returned home and "
				      "deregistered all care-of addresses\n");
				/* FIX: also HA should send gratuitous ARP
				 * telling that MN takes care of its own
				 * packets from now on;
				 * draft-ietf-mobileip-rfc2002-bis-03.txt, 4.6;
				 * This would require that HA would get the
				 * link-layer address of the MN from the
				 * deregistration message and use it in the
				 * gratuitous ARP packet */
			}
			killbinding = 1;
		}
	} 
	
	if (binding == NULL) {
		if (ntohs(ext.req->lifetime) == 0) {
			DEBUG(DEBUG_FLAG, "Deregistration attempt, but binding"
			      " not found - creating temporary binding\n");
			binding = create_binding(&ext, mn_spi, 0);
			killbinding = 1;
		} else {
			/*--------------------------------------------------*
	 		 | Create a new binding for the MN					|
 	 		 *--------------------------------------------------*/
 	 		binding = create_binding(&ext, mn_spi, 1);
		}
		if (binding == NULL) {
			print_trace(traces, "Failed to create binding!\n");
			LOG2(LOG_ERR, "Failed to create binding for MN: %s, "
			     "FA: %s\n", mn_addrstr, fa_addrstr);
			send_reg_failure(rsock, &cli_addr, mn_spi,
					 (auth_type == AUTH_MAC_RFC2002 ?
					  1 : 0),
					 REGREP_NO_RESOURCES_HA, NULL, &ext,
					 force_iface, binding, from_iface);		
			if (hoa_was_zero) free_hoa(&ext.req->home_addr);			 
			return -1;
		}
		t_data = (struct ha_tunnel_data *) binding->data;
		t_data->auth_type = (auth_type == AUTH_MAC_RFC2002 ?
				     AUTH_RFC2002BIS : AUTH_RFC2002);
	}
	ASSERT(binding != NULL);
	t_data = (struct ha_tunnel_data *) binding->data;
	ASSERT(t_data != NULL);

	binding->mod_time = time(NULL);
	t_data->reverse_tunnel = (ext.req->opts & REGREQ_REVERSE_TUNNEL) != 0;
	if (ext.req->opts & REGREQ_MINIMAL_ENCAPS)
		t_data->encapsulation = ENCAPS_MINIMAL;
	else if (ext.req->opts & REGREQ_GRE_ENCAPS)
		t_data->encapsulation = ENCAPS_GRE;
	else
		t_data->encapsulation = ENCAPS_IPIP;

	if (ext.fa_keyreq)
		binding->fa_spi = ntohl(ext.fa_keyreq->spi);
	else
		binding->fa_spi = 0;

	/* If the request has a public key add it to the binding */
	if (ext.fa_pubkey) {
		/* binding->fa_pubkey is NULL if we have generated a
		 * new SK. If the public key sent by the FA is the same 
		 * as in a previous binding, we can reuse the encrypted
		 * SK and save some CPU cycles */
		if (binding->fa_pubkey &&
		    binding->last_sent_fa_pubkeyrep != NULL &&
		    (memcmp(binding->fa_pubkey, 
			    ext.fa_pubkey, 
			    GET_KEY_EXT_LEN(ext.fa_pubkey)) == 0))
		{
			DEBUG(DEBUG_FLAG, "Reusing encrypted SK for FA %s\n",
			      fa_addrstr);
		} else {
			if (binding->fa_pubkey)
				free(binding->fa_pubkey);
			if (binding->last_sent_fa_pubkeyrep)
				free(binding->last_sent_fa_pubkeyrep);
			binding->last_sent_fa_pubkeyrep = NULL;
			binding->fa_pubkey = (struct msg_key *)
				malloc(GET_KEY_EXT_LEN(ext.fa_pubkey));
			if (binding->fa_pubkey != NULL) {
				memcpy(binding->fa_pubkey, ext.fa_pubkey,
				       GET_KEY_EXT_LEN(ext.fa_pubkey));
			} else {
				print_trace(traces, "Not enough memory for fa_pubkey!\n");
				LOG2(LOG_ERR, "Not enough memory "
				     "for fa_pubkey\n");
				if (hoa_was_zero) free_hoa(&ext.req->home_addr);		     
				return -1;
			}
			binding->last_sent_fa_pubkeyrep =
				dynamics_do_rsa_encrypt(binding->key,
							binding->keylen,
							binding->fa_pubkey);
			if (binding->last_sent_fa_pubkeyrep == NULL) {
				print_trace(traces, "RSA encryption failed!\n");
				LOG2(LOG_ERR, "RSA encryption failed (MN=%s, "
				     "FA=%s)\n", mn_addrstr, fa_addrstr);
			}
		}

	} else {
		/* remove the (possible) old FA pubkey from the binding to make
		 * sure that the next FA using RSA will get the correct key */
		if (binding->fa_pubkey)
			free(binding->fa_pubkey);
		binding->fa_pubkey = NULL;

		if (binding->last_sent_fa_pubkeyrep)
			free(binding->last_sent_fa_pubkeyrep);
		binding->last_sent_fa_pubkeyrep = NULL;
	}
	
	memcpy(&binding->id, ext.req->id, REG_REQ_ID_LEN);
	
	/* store where the request came from */
	t_data->lower_saddr.s_addr = cli_addr.sin_addr.s_addr;
	binding->lower_port = cli_addr.sin_port;

	binding->timeout = MIN(ntohs(ext.req->lifetime), mn_spi->max_lifetime);
	binding->exp_time = time(NULL) + binding->timeout;

	/*--------------------------------------------------*
	 | Send Registration Reply							|
	 *--------------------------------------------------*/
	if (send_reg_repl(rsock, binding, mn_spi, &ext, force_iface,
			  code == REGREP_ACCEPTED_NO_SB ?
			  REGREP_ACCEPTED_NO_SB : REGREP_ACCEPTED) != 0) 
	{
		print_trace(traces, "Failed to send registration reply!\n");
		LOG2(LOG_ERR, "sending of registration reply to MN: %s, "
		     "FA: %s failed\n", mn_addrstr, fa_addrstr);
		remove_tunnel(binding);
		destroy_binding(binding);
		return -1;
	}

	if (killbinding) {
		destroy_binding(binding);
	} else {
		/* increase timeout with one second so that the one second
		 * timer resolution doesn't cause too early expirations */
		binding->timeout++;

		/* make sure that the bindings module gets timer update just
		 * before adding the binding */
		check_bindings();
		binding_add(bindings, binding);
	}
	
	return 0;
}


/**
 * ha_send_agent_adv:
 * @iface: interface to be used for sending the advertisement
 * @to: destination IP address for unicast agent advertisement or %NULL for
 * broadcast/multicast agent advertisement
 *
 * Send an agent advertisement to the given interface and destination address.
 *
 * Returns:
 *   0 on success or < 0 on failure (send_agent_advertisement() error code)
 */
static int
ha_send_agent_adv(struct interface_entry *iface, struct sockaddr_ll *to,
		  struct in_addr *dest)
{
	/* Set Agent Advertisement data */
	set_agent_adv_data(
		config.ha_default_tunnel_lifetime,
		iface->addr,
		iface->addr, AGENT_ADV_HOME_AGENT, 0,
		iface->interval > 21845 ?
		65535 : iface->interval * 3, NULL, 0);
	set_agent_adv_nai(ha_nai);

	return send_agent_advertisement(iface->icmp_sock,
					(struct sockaddr *) to, dest,
					iface->if_index);
}


/**
 * send_agent_advs:
 * @next_agentadv: pointer to the next advertisement time
 *
 * Send unsolicited agent advertisements on interfaces for which the configured
 * interval has passed from the previous unsolicited advertisement.
 * This function also fills in @next_agentadv with the time when the next
 * unsolicited advertisement is due.
 *
 * Updates iface.last_adv in the config.interfaces list.
 */
static void
send_agent_advs(struct timeval *next_agentadv)
{
	struct interface_entry *iface;
	struct node *node;
	struct timeval now, next;
	unsigned int diff;
	int first = 1;

	/* Send ICMP Agent Advertisement message and save the timestamp */
	gettimeofday(&now, NULL);
	next.tv_sec = 0;
	next.tv_usec = 0;
	for (node = list_get_first(&config.interfaces); node != NULL;
	     node = list_get_next(node)) {
		iface = (struct interface_entry *) node;
		if (iface->agentadv != INTERFACE_AGENTADV_ALL)
			continue;

		if (now.tv_sec >= iface->last_adv.tv_sec + iface->interval) {
			if (!ha_send_agent_adv(iface, NULL, NULL))
				stats.adv_sent++;
			else {
				DEBUG(DEBUG_FLAG, "send_agent_advs: "
				      "advertisement sending to iface[%s] "
				      "failed\n", iface->dev);
			}
			iface->last_adv.tv_sec = now.tv_sec;
			iface->last_adv.tv_usec = now.tv_usec;
		}

		next.tv_sec = iface->last_adv.tv_sec + iface->interval;
		/* Add random delay */
		next.tv_usec = iface->last_adv.tv_usec +
			(int) (MAX_ADV_DELAY * rand() / (RAND_MAX + 1.0));
		if (next.tv_usec > 999999) {
			next.tv_sec++;
			next.tv_usec -= 1000000;
		}
		if (first || next.tv_sec < next_agentadv->tv_sec) {
			DEBUG(DEBUG_FLAG, "** ");
			*next_agentadv = next;
		}
		first = 0;

		/* for debuging */
		diff = ((next.tv_sec - now.tv_sec) * 1000000 + 
			next.tv_usec - now.tv_usec) / 1000; 
		DEBUG(DEBUG_FLAG, "send_agent_advs: next agentadv: "
		      "%ld.%ld diff = %u msec\n", 
		      next.tv_sec, next.tv_usec, diff);
	}
}


static void
handle_icmp(struct interface_entry *iface)
{
	int r;
	struct sockaddr_ll from;
	struct in_addr to, fromaddr;

	r = check_icmp_sol(iface->icmp_sock, &from, &to, &fromaddr);

	if (r == 0 && iface->agentadv != INTERFACE_AGENTADV_NONE) {
		/* Agent solicitation */
		DEBUG(DEBUG_FLAG,
		      "\tagent solicitation => reply with agent adv\n");
		send_agent_advertisement(iface->icmp_sock,
					 (struct sockaddr *) &from, &fromaddr,
					 from.sll_ifindex);
	}
}


/**
 * set_expr_timer:
 * @tv: pointer to the returned timeval
 * @next_agentadv: timestamp for the next unsolicited agent advertisement
 *
 * Schedule next wake up time for the main loop. set_expr_timer() select the
 * next upcoming event (an unsolicited agent advertisement, expiring binding,
 * queued gratuitous ARP, or delayed tunnel deletion). The wake up time is
 * returned in @tv when an event is available or @tv->sec is set to -1 in case
 * there are no waiting events.
 */
static void
set_expr_timer(struct timeval *tv, struct timeval next_agentadv)
{
	int expire_time;
	struct timeval now;

	gettimeofday(&now, NULL);
	DEBUG(DEBUG_FLAG, "set_expr_timer (now=%li.%06li)\n",
	      now.tv_sec, now.tv_usec);

	if (timerisset(&next_agentadv)) {
		tv->tv_sec = next_agentadv.tv_sec - now.tv_sec;
		tv->tv_usec = next_agentadv.tv_usec - now.tv_usec;
		if (tv->tv_usec < 0) {
			tv->tv_sec--;
			tv->tv_usec += 1000000;
		}
		if (tv->tv_sec < 0) {
			tv->tv_sec = 0;
			tv->tv_usec = 0;
		}
		DEBUG(DEBUG_FLAG, "\tnext_agentadv in %li.%06li sec\n",
		      tv->tv_sec, tv->tv_usec);
	} else {
		tv->tv_sec = -1;
		tv->tv_usec = 0;
	}

	if ((expire_time = binding_nextexpiretime(bindings)) >= 0) {
		DEBUG(DEBUG_FLAG, "\tbinding next expire in %i sec\n",
		      expire_time);
		if (tv->tv_sec < 0 || expire_time < tv->tv_sec) {
			DEBUG(DEBUG_FLAG, "\tsetting expire time to %i sec\n",
			      expire_time);
			tv->tv_sec = (time_t)expire_time;
			tv->tv_usec = 0;
		}
	}

	if (tv->tv_sec == -1 && tunnel_delayed_exists(tunnels)) {
		DEBUG(DEBUG_FLAG, "\tsetting expire time to %i sec (tunnel "
		      "check)\n", TUNNEL_DELAYED_CHECK_INTERVAL);
		tv->tv_sec = TUNNEL_DELAYED_CHECK_INTERVAL;
	}

	if (gratuitous_arp_queue != NULL) {
		struct timeval next;

		next.tv_sec = gratuitous_arp_queue->tv.tv_sec - now.tv_sec;
		next.tv_usec = gratuitous_arp_queue->tv.tv_usec - now.tv_usec;
		if (next.tv_usec < 0) {
			next.tv_sec--;
			next.tv_usec += 1000000;
		}
		if (next.tv_sec < 0) {
			next.tv_sec = 0;
			next.tv_usec = 0;
		}

		if (tv->tv_sec == -1 || cmp_timeval(tv, &next) > 0) {
			tv->tv_sec = next.tv_sec;
			tv->tv_usec = next.tv_usec;
			DEBUG(DEBUG_FLAG, "\tnext gratuitous ARP in %li.%06li "
			      "sec\n", tv->tv_sec, tv->tv_usec);
		}
	}
}


/**
 * init_interfaces:
 *
 * Initialize all the configured interfaces:
 *   fill in the kernel ifindex and IP address (if not forced),
 *   open ICMP socket for agent advertisement
 */
static void
init_interfaces(void)
{
	struct interface_entry *iface;
	struct in_addr bc;

	DEBUG(DEBUG_FLAG, "Initializing interfaces\n");
	inet_aton("255.255.255.255", &bc);
	iface = (struct interface_entry *) list_get_first(&config.interfaces);
	if (iface == NULL) {
		LOG2(LOG_ALERT, "No interfaces defined\n");
		clean_up(-1);
	}
	while (iface != NULL) {
		DEBUG(DEBUG_FLAG, "\t%s: ", iface->dev);
		iface->if_index = dyn_ip_get_ifindex(iface->dev);
		if (iface->if_index < 0) {
			LOG2(LOG_ALERT,
			     "Could not get ifindex for interface %s\n",
			     iface->dev);
			clean_up(-1);
		}
		DEBUG(DEBUG_FLAG, "ifindex=%i ", iface->if_index);

		if (iface->force_addr.s_addr != 0) {
			DEBUG(DEBUG_FLAG, "forcing address ");
			iface->addr = iface->force_addr;
		} else {
			if (dyn_ip_get_ifaddr(iface->dev, &iface->addr) != 0) {
				LOG2(LOG_ALERT, "could not get interface[%s] "
				     "address\n", iface->dev);
				clean_up(-1);
			}
		}

		/* bind unicast UDP socket into local IP address, but not
		 * interface */
		iface->udp_sock = dynamics_open_udp_socket(
			config.socket_priority, 
			iface->force_addr.s_addr != 0 ? iface->force_addr :
			iface->addr, config.udpport, NULL);
		if (iface->udp_sock < 0) {
			LOG2(LOG_ALERT, "UDP socket opening failed\n");
			clean_up(-1);
		}

		if (iface->ha_disc) {
			/* bind broadcast UDP sockets into broadcast IP address
			 * and interface */
			if (dyn_ip_get_bcaddr(iface->dev, &iface->bcaddr) != 0)
			{
				LOG2(LOG_ALERT, "could not get interface[%s] "
				     "broadcast address\n", iface->dev);
				clean_up(-1);
			}
			iface->udp_bc_sock = dynamics_open_udp_socket(
				config.socket_priority, iface->bcaddr,
				config.udpport, iface->dev);
			if (iface->udp_bc_sock < 0) {
				LOG2(LOG_ALERT, "UDP broadcast socket opening "
				     "failed for interface[%s]\n", iface->dev);
				clean_up(-1);
			}

			iface->udp_bc_sock2 = dynamics_open_udp_socket(
				config.socket_priority, bc, config.udpport,
				iface->dev);
			if (iface->udp_bc_sock2 < 0) {
				LOG2(LOG_ALERT, "UDP broadcast "
				     "(255.255.255.255) socket opening "
				     "failed for interface[%s]\n", iface->dev);
				clean_up(-1);
			}
		}

		iface->icmp_sock =
			open_agent_icmp_adv_socket(iface->dev,
						   AGENTADV_FILTER_SOL);
		if (iface->icmp_sock < 0) {
			LOG2(LOG_ALERT, "could not open ICMP socket\n");
			clean_up(-1);
		}

		DEBUG(DEBUG_FLAG, "%s => socket=%i\n", inet_ntoa(iface->addr),
		      iface->icmp_sock);

		iface = (struct interface_entry *) list_get_next(&iface->node);
	}
}


/**
 * ha_parse_command_line:
 * @argc: @argc from main()
 * @argv: @argv from main()
 *
 * Parse command line arguments.
 */
static void
ha_parse_command_line(int argc, char *argv[])
{
	int c, oindex = 0;
	static struct option long_options[] = {
		/* common arguments */
		{"help", no_argument, NULL, 'h'},
		{"version", no_argument, NULL, 'v'},
		{"debug", no_argument, NULL, 0},
		{"fg", no_argument, NULL, 0},
		{"config", required_argument, NULL, 'c'},
		{"traces", no_argument, NULL, 't'},
		/* HA specific arguments */
		/* end of arguments */
		{0, 0, 0, 0}
	};

	DEBUG(DEBUG_FLAG, "HA command line parsing\n");

	while ((c = getopt_long(argc, argv, "+hv", long_options, &oindex)) !=
	       EOF) {
		
		switch (c) {
		case 'h':
			/* show HA specific parameters */
			/* N/A */
			exit(1);
			break;
		
		case 't':
			printf("--traces: Print traces.\n");
			traces = 1;
			break;
		
		case '?':
			printf("Command line parsing failed - aborting\n");
			exit(1);

		case 'v':
		case 'c':
		case 0:
			break;
			
		default:
			printf("?? getopt returned character code 0%o ??\n",
			       c);
			exit(1);
		}
	}
}

/**
 * check_if_tun_fds_set: Called by hashtable_iterator.
 * 
 * Returns 1 to continue iteration.
 */
int check_if_tun_fds_set(struct node *node, void *set) {

	TUNNEL *tunnel_entry 	= (struct tunnel *) node;
	fd_set *fdset 			= (fd_set *) set;
	int fd					= tunnel_entry->tun_fd; 
	//DEBUG(DEBUG_FLAG, "Check if tun fd %d is set\n", fd);
	if (FD_ISSET(fd, fdset)) {
		handle_msg_in_tun_fd(tunnel_entry);
	}
	return 1;
}

/**
 * main:
 * @argc: argument count
 * @argv: argument values
 *
 * Daemon execution start point; does not return.
 */
int
main(int argc, char *argv[])
{
	
	int api_rw_sock; /* unix socket for privileged (may modify data)
			    	  * API calls */
	int api_ro_sock; /* unix socket for unprivileged (read only)
			    	  * API calls */
	struct timeval next_agentadv;
	struct interface_entry *iface;

	DEBUG(DEBUG_FLAG, "Note that the system will not work if there is a default route with metric 0.\n");

	program_name = parse_long_options(argc, argv, "home agent", PACKAGE,
					  VERSION, dynamics_usage);

	ha_parse_command_line(argc, argv);

	memset(&config, 0, sizeof(struct ha_config));

	if (load_config(&config, program_name,
		opt_config == NULL ? HA_GLOBAL_CONF_FILE : opt_config)
		== FALSE) 
	{
		clean_up(1);
	}
	set_ha_nai();
	
	/*----------------------------------------------*
 	 | IP in IP and GRE support check disabled.		|
	 *----------------------------------------------*/
	/* ... because this version uses only UDP in IP tunneling. */
	check_kernel_support(CHECK_KERNEL_IPIP | CHECK_KERNEL_NETLINK);
	
	if (getuid() || geteuid()) {
		fprintf(stderr, "%s: This program must be run by root.\n",
			program_name);
		exit(1);
	}
 	
 	if (config.hoa_pool_first.s_addr <= 0 || config.hoa_range <= 0) {
 		fprintf(stderr, "HoA pool configured incorrectly!\n");
 	}
 	init_hoa_pool();
 	
	iface = (struct interface_entry *) list_get_first(&config.interfaces);
	srand(time(NULL) ^ (iface ? iface->addr.s_addr : 0));

	allow_ipv4_forwarding();

	/* init tables */
	bindings = binding_init(config.max_bindings,
				config.ha_default_tunnel_lifetime);
	if (bindings == NULL) {
		LOG2(LOG_ERR, "binding_init failed\n");
		clean_up(1);
	}

	init_interfaces();

	tunnels = tunnel_init("TUNL", 1, 253);
	if (tunnels == NULL) {
		LOG2(LOG_ERR, "tunnel_init failed\n");
		clean_up(1);
	}

	api_rw_sock = api_open_socket(config.ha_api_admin_socket_path,
				      config.ha_api_admin_socket_group,
				      config.ha_api_admin_socket_owner,
				      config.ha_api_admin_socket_permissions);
	if (api_rw_sock < 0) {
		LOG2(LOG_ERR, "Could not create API rw-socket\n");
		clean_up(1);
	}

	api_ro_sock = api_open_socket(config.ha_api_read_socket_path,
				      config.ha_api_read_socket_group,
				      config.ha_api_read_socket_owner,
				      config.ha_api_read_socket_permissions);
	if (api_ro_sock < 0) {
		LOG2(LOG_ERR, "Could not create API ro-socket\n");
		clean_up(1);
	}

	if (!opt_foreground && dynamics_fork_daemon() == -1)
		clean_up(1);
	dynamics_write_pid_file(HA_PID_FILE);

	memset(&stats, 0, sizeof(stats));
	dynamics_strlcpy(stats.version, VERSION, sizeof(stats.version));

	/* set up signal actions */
	signal(SIGTERM, sig_handler);            /* clean death */
	signal(SIGINT, sig_handler);             /* clean death */
	signal(SIGHUP, reload_config);

	syslog(LOG_INFO, "%s home agent daemon version %s started\n",
	       PACKAGE, VERSION);

	next_agentadv.tv_sec = 0;
	next_agentadv.tv_usec = 0;
	send_agent_advs(&next_agentadv);

	DEBUG(DEBUG_FLAG, "HoA pool starts from %s\n", inet_ntoa(config.hoa_pool_first));
	DEBUG(DEBUG_FLAG, "HoA range %d\n", config.hoa_range);
	
	/* loop forever waiting for messages */
	for (;;) {
		fd_set set;
		int oldmask;
		struct node *node;
		struct timeval tv;
		
		int i = 0;
		
		oldmask = sigblock(sigmask(SIGHUP));
		FD_ZERO(&set);
		
		hashtable_iterator(tunnels, tun_fds_into_fdset_iter, &set);
		FD_SET(api_rw_sock, &set);
		FD_SET(api_ro_sock, &set);
		
		for (node = list_get_first(&config.interfaces); node != NULL;
		     node = list_get_next(node)) {
			iface = (struct interface_entry *) node;
			FD_SET(iface->udp_sock, &set);
			if (iface->ha_disc && iface->udp_bc_sock > -1)
				FD_SET(iface->udp_bc_sock, &set);
			if (iface->ha_disc && iface->udp_bc_sock2 > -1)
				FD_SET(iface->udp_bc_sock2, &set);
			FD_SET(iface->icmp_sock, &set);
		}

		set_expr_timer(&tv, next_agentadv);
		sigsetmask(oldmask);

		if (select(FD_SETSIZE, &set, NULL, NULL,
			   tv.tv_sec == -1 ? NULL : &tv) < 0) {
			/* don't exit on e.g. SIGHUP */
			if (errno != EINTR) {
				LOG2(LOG_ERR, "select - %s", strerror(errno));
				LOG2(LOG_ERR, "!!!\n");
				clean_up(1);
			}
			continue;
		}
		       
		/* we don't want to reread the configuration file
		 * while we are handling requests; hold the signal
		 * until we are ready */
		oldmask = sigblock(sigmask(SIGHUP));

		/*--------------------------------------------------*
 	 	 | See if there are messages the from the kernel.	|
	 	 *--------------------------------------------------*/
		hashtable_iterator(tunnels, check_if_tun_fds_set, &set);
		
		if (FD_ISSET(api_rw_sock, &set)) {
			/* privileged API call */
			handle_api(api_rw_sock, 1);
		}

		if (FD_ISSET(api_ro_sock, &set)) {
			/* unprivileged API call */
			handle_api(api_ro_sock, 0);
		}
		
		for (node = list_get_first(&config.interfaces); node != NULL;
		     node = list_get_next(node)) {
			iface = (struct interface_entry *) node;
			
			/* UDP message */
			if (FD_ISSET(iface->udp_sock, &set)) {
				DEBUG(DEBUG_FLAG, "Got UDP message(%s)\n",
				      iface->dev);
				handle_reg_msg(iface, iface->udp_sock);
			}
			
			/* UDP broadcast message */
			if (iface->ha_disc && iface->udp_bc_sock > -1 &&
			    FD_ISSET(iface->udp_bc_sock, &set)) {
				DEBUG(DEBUG_FLAG, "Got UDP (BC) message(%s)\n",
				      iface->dev);
				handle_reg_msg(iface, iface->udp_bc_sock);
			}
			
			/* UDP broadcast 255.255.255.255 message */
			if (iface->ha_disc && iface->udp_bc_sock2 > -1 &&
			    FD_ISSET(iface->udp_bc_sock2, &set)) {
				DEBUG(DEBUG_FLAG, "Got UDP (BC "
				      "255.255.255.255) message(%s)\n",
				      iface->dev);
				handle_reg_msg(iface, iface->udp_bc_sock2);
			}
			
			/* ICMP message */
			if (FD_ISSET(iface->icmp_sock, &set)) {
				handle_icmp(iface);
			}
		}

		/* Send AA messages */
		if (timerisset(&next_agentadv))
			send_agent_advs(&next_agentadv);
			
		check_bindings();
		tunnel_check_delayed(tunnels, 0);
		check_queued_gratuitous_arp();

		sigsetmask(oldmask);
		
	}

	/* not reached */
	return 0;
}
